import javax.swing.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.*;


public class Curve {

	ArrayList<Point> curve = new ArrayList<Point>();
	ArrayList<Point> points  = new ArrayList<Point>();
	ArrayList<Point> intPoints  = new ArrayList<Point>();
	ArrayList<Edge> edges  = new ArrayList<Edge>();
	ArrayList<Face> faces = new ArrayList<Face>();
	
	ArrayList<Point> startingIntPoints = new ArrayList<Point>();
	int idealSplittingPoints = 0;
	ArrayList<Point> faceReps = new ArrayList<Point>();
	double areaSwept = 0; 
	double windingArea = 0;
	double curWindingArea = 0;
	boolean overlapping; 
	Point anchor;
	
	Color C;
	    
	public Curve( ArrayList<Point> C  )  {
		ArrayList<Point> used = new ArrayList<Point>();
		
		for (Point P: C) { 
    		if ( !contains(used, P) ) {
    			used.add(P);
    			Point Q = new Point(P.getX(), P.getY() );
        		Q.setNum( P.getNum()); // keep the #, for debugging purposes
        		Q.setIndex( P.getIndex(0) ); Q.setIndex( P.getIndex(0) );
        		Q.setIntPointIndex( P.getIntPointIndex() ); // in case we have an int point passed down
        		Q.name( P.toString());
        		this.curve.add( Q );  
    		}
    		else {
    			Point add = null;
    			for (Point Q: this.curve) {  // P was copied already, so find the copy Q in the array and add it again
    				if (Q.equals(P)) { 
    					add = Q; 
    					break; 
    				}
    			}
    			try { if (add == null) { throw new Exception("Point was used but could not be found"); }     		}
    			catch ( Exception EE) { System.err.println(EE); }
    			this.curve.add( add );
    		}
    	}
		
		for (Point P: this.curve) { P.resetUses(); } 
    	for (Point P: this.curve) { P.use(); }
    	for (Point P: this.curve) {
    		if (P.getUses() == 1 ) { P.notIntersection(); points.add(P); }
    		else if (P.getUses() == 2) {
    			P.setIntersection(); 
    			if (!contains(startingIntPoints, P)) {startingIntPoints.add(P); }
    		}
    		else {
    			try { throw new Exception("Error. Point used more than twice on a curve.");	}
    			catch ( Exception EE) { System.err.println(EE);  }
    		}
    	}
    	for (Point P: startingIntPoints) { // delete int points and recompute them!!
    		P.resetIndices();
    		while ( contains(this.curve, P) ) {
    			int index = this.curve.indexOf(P);
    			this.curve.remove(index);
    		}
    	}
    	for (int i=0; i < points.size(); i++) { 
    		Point cur = points.get(i);
    		Point next = points.get( (i+1) % points.size() ) ;
    		edgeAdd( cur, next );
    	}
    	for (Point P: intPoints) {
    		if ( P == null) {
    			try { 
    				System.out.println("null point in int points");
    			}
    			catch (Exception E) {
    				System.err.println(E);
    			}
    		}
    	}
    	numberPoints(false);
	}
	
	public Point getAnchor() {
		for (Point P: curve) {
    		if (P.toString().contains("i")) {
    			return P;
    		}
    	}
		return curve.get(0);
	}
	
	public void clear() {
		curve.clear();
		faces.clear();
		edges.clear();
		points.clear();
		intPoints.clear();
		return;
	}
	public double getAreaSwept() { 
		return this.areaSwept;
	}
	public ArrayList<Edge> getEdges() {
		return this.edges;
	}
	
	public ArrayList<Point> getCurve()  {
		return this.curve;
	}
	
	public Point get(int i) {
		return curve.get(i);
	}
	
	public void setColor(Color C) {
		this.C = C;
	}
	
	public Color getColor() {
		return this.C;
	}
	
	public Curve() {
		curve = new ArrayList<Point>();
		points = new ArrayList<Point>();
		intPoints = new ArrayList<Point>();
	}
	
	public boolean contains(ArrayList<Edge> A, Edge e) {  // the arraylist.contains(Edge) method doesn't work
    	for (Edge E: A) {
    		if (E.equals(e)) {
    			return true;
    		}
    	}
    	return false;
    }
    
    public boolean contains(ArrayList<Face> A, Face f) {  // the arraylist.contains(Edge) method doesn't work
    	for (Face F: A) {
    		if (F.equals(f)) {
    			return true;
    		}
    	}
    	return false;
    }
    
    public boolean contains(ArrayList<Point> A, Point p) { // arraylist.contains(Point) method doesn't work
		for (Point P: A) {
			if (P.equals(p)) {
				return true;
			}
		}
		return false;
    }
    
    public void addEdge(Edge e) { // take an edge and add it to the edges of each point on the edge
    	ArrayList<Point> edgePoints = new ArrayList<Point>(); 
    	edgePoints.add( e.getP1() ); edgePoints.add( e.getP2() ) ;
    	edgePoints.addAll( e.getIntPoints() ); 
    	for (Point P: edgePoints ) { // edgePoints has all intpoints on the edge + endpoints
    		if ( !contains(P.getEdges() , e ) ) {
    			P.addEdge(e); // ensure each point has the edge!
    		}
    	}
    }
    
    
    public void classifyFaces() {
    	
    	ArrayList<Face> classifieds = new ArrayList<Face>(); 
    	
    	for (Point P: curve) { // find all faces with depth 0 
    		Face F = faces.get(P.getFaces().get(0).intValue());
    		if ( P.faceMembership() == 1 && !contains(classifieds,F) ) {
    			classifieds.add( F);
    			F.setDepth(1);
    		}
    	}
    	
    	int i =0;
    	ArrayList<Face> remaining = new ArrayList<Face>();
    	remaining.addAll(faces);
    	remaining.removeAll( classifieds);
    	
    	while (i < 1000 && remaining.size() > 0 ) { // don't allow loop to run forever (go till faces are classified)
    		
    		ArrayList<Face> remove = new ArrayList<Face>(); // all faces classified in the latest run through
    		for (Face newF: remaining) {
    			boolean done = false;
    			for ( Edge E: newF.getEdges()) {
    				if (!done) {
    					for (Face oldF: classifieds) {
    						if ( contains(oldF.getEdges(), E)) {
        	    				done = true; // if so, compute winding # for the new face f
        	    				remove.add(newF);
        	    				newF.setDepth( oldF.getDepth() + 1 );
        	    				break;
        	    			}
        				}
    				}
    			}
    		}
    		
    		for (Face F: remove) {
    			remaining.remove(F); // remove face from unclassifieds
    			classifieds.add(F); // add to classifieds
    		}
    		i++;
    	}
    	
    	for (Face F: faces) { 
    		if ( F.getDepth() == Math.abs( F.getWindingNum() ) ) {
    			F.setGood(true);
    		}
    		else {
    			F.setGood(false);
    		}
    	}
    	
    }
    
    public boolean colinear(Edge e1, Edge e2) {
    	Point P = e1.getP1(); Point Q = e1.getP2(); Point R = e2.getP2();
    	ArrayList<Point> candidates = new ArrayList<Point>();
    	candidates.add(P); candidates.add(Q); candidates.add(R); // add all 3 points to arraylist 
    	for (Edge E: edges) { 
    		ArrayList<Point> pointsOnE = new ArrayList<Point>();
    		
    		pointsOnE.add( E.getP1() ); pointsOnE.add( E.getP2() ); // add endpoints 
    		for ( Point u: E.getIntPoints() ) {  pointsOnE.add(u); } // add all intersection points 
    		if ( pointsOnE.containsAll(candidates) ) {
    			return true;
    		}
    	}
    	return false; // false if no edge contained all the points
    }
		
    public Face getFace(Edge E) {
    	Edge curEdge = E;
    	Point p = curEdge.getP1();
    	ArrayList<Point> pointsUsed = new ArrayList<Point>();
    	ArrayList<Point> pointsUsedTwice = new ArrayList<Point>();
    	ArrayList<Edge> faceEdges = new ArrayList<Edge>();
    	faceEdges.add(curEdge);
    	Point prev = curEdge.getP1();
    	Point curPoint = curEdge.getP2();
    	Point startPoint = curEdge.getP1();
    	pointsUsed.add(startPoint);
    	pointsUsed.add(curPoint);
    	boolean stop = false;
    	
    	while (true && !stop) {
    		Edge best = null;
    		if ( curPoint.isIntersection() ) { // if at an intersection point, turn left
    			int s = curPoint.getShortEdges().size();
    			for (Edge e: curPoint.getShortEdges()) { // ensure a left turn
        			double val = curEdge.turn(e);
        			if ( colinear( curEdge, e) ) {
        				//System.out.println("Points: " + curEdge.getP1().toString() +","+ curEdge.getP2().toString() +","+ e.getP2().toString() + " are colinear");
        			}
        			else if ( val < 0) {
        				best = e;
        				break;
        			}
        		}
    			if ( best == null && s >= 4 ) {
    				System.out.println("Current point " + curPoint.toString() + " has " + s + " edges");
    				System.out.println("NO LEFT TURN FROM THE EDGE: " + curEdge.toString() + ". The turns are: ");
    				for (Edge e: curPoint.getShortEdges()) { // ensure a left turn
    					double val = curEdge.turn(e);
            			System.out.println("The turn: "+  curEdge.getP1().toString() +"->"+ curEdge.getP2().toString() +"->"+ e.getP2().toString() + " is " + val);
            			if ( val < 0 && !colinear(curEdge, e)) {
            				best = e;
            				break;
            			}
            		}
        			return null; // nowhere to go !
        		}
    			else if ( s < 4) {
    				System.out.println("Edges connected to point: " + curPoint.getNum() );
    				for (Edge e: curPoint.getShortEdges()) {
            			double angle = curEdge.turn(e); // show info since weird stuff
            			System.out.println("first edge: " + curEdge.toString() +  
            					"and  second edge: " + e.toString() + " with angle " + angle );
            		}
    				return null;
    			}
    		}
    		else { // otherwise not at an intersection point, so take the non- backwards edge
	    		Edge e1 = curPoint.getShortEdge(0);
	    		Edge e2 = curPoint.getShortEdge(1);
	    		if ( e1.getP2().equals(prev) ) {
	    			best = e2;
	    		}
	    		else { 
	    			best = e1;
	    		}
    		}
    		curEdge = best; // update current edge, save it 
    		faceEdges.add(curEdge);
    		prev = curPoint; // current point becomes previous point 
    		if ( best == null || curEdge.getP2() == null) {
    			return null;
    		}
    		curPoint = curEdge.getP2(); // update current point
    		
    		// if starting point is an int point, and we reached it a 2nd time, check for lightbulb shape
    		if ( contains(pointsUsed, curPoint) && curPoint.equals(startPoint) && curPoint.isIntersection() && !contains(pointsUsedTwice, curPoint))  { // special check for lightbulb curve
    			pointsUsedTwice.add(curPoint);
    			for (Edge e: curPoint.getShortEdges()) { // ensure a left turn
        			double val = curEdge.turn(e);
        			if ( !colinear(curEdge, e) && val < 0) {
        				best = e;
        				break;
        			}
        		}
    				if ( contains(pointsUsed, best.getP2() ) ) { // check if we are at a new point
    					break;
    			}
    			else { // we have a lightbulb shape!
    				curEdge = best; // update current edge, save it 
            		faceEdges.add(curEdge);
            		prev = curPoint; // current point becomes previous point 
            		curPoint = curEdge.getP2(); // update current point
    			}
    		}  
    			
    		if ( curPoint.equals(p) ) { 
    			break; // quit once starting point is reached ( if point is non-intersection )
    		}
    		else { // continue if not back at starting point yet
    			
    			if ( pointsUsed.contains(curPoint) ) { // check if point is used 
    				if (pointsUsedTwice.contains(curPoint) ) {// if used twice, then face is degenerate
    	    			return null; // nothing found since  
    				}
    				else {
    					pointsUsedTwice.add(curPoint);
    				}
    			}
    			else {
    				pointsUsed.add(curPoint); // keep going if point only hit once 
    			}
    		}
    	}
    	try {
	    	if ( faceEdges.size() <= 2) { // NOT A FACE!!
	    		throw new Exception("Incomplete Face!! <=2 Points on the Face!");
	    	}
    	}
    	catch (Exception EE) {
    		System.err.println(EE);
    		return null;
    	}
    	return new Face( faceEdges);
    }
    
    public boolean adjacent(Point P, Point Q) {  // determine if points are adjacent on the curve
    	int p1 = -1; int p2 = -1;
    	for (int i = 0; i< curve.size(); i++) {
    		Point R = curve.get(i);
    		if ( R.equals(P) && p1 == -1 ) {
    			p1 = i;
    		}
    		else if ( R.equals(P) && p2 == -1) {
    			p2 = i;
    			break;
    		}
    	}
    	if ( curve.get( (p1+1) % curve.size() ).equals(Q) )  {
    		return true;
    	}
    	if ( p2 != -1 && curve.get( ( p2 + 1 ) % curve.size() ).equals(Q) ) {
    		return true;
    	}
    	return false; 
    } 
    
 // input the curve in order of traversal, with an interior point z
  	public double windingNum(Point z){
  	    double wNum;
  	    ArrayList<Point> A = new ArrayList<Point>();
  	    for (int i = 0; i < this.curve.size(); i++) {
  	    	A.add( new Point(this.curve.get(i).getX() - z.getX(), this.curve.get(i).getY() - z.getY()) );
  	  	}
  	 	wNum = 0;
  	 	   
  	 	for (int a = 0; a < A.size(); a++) {
  	 		
  	 		double sign = 0.0;  
  	 		sign = (A.get(a).getY()) * (A.get((a+1) % A.size()).getY());
  	 		  
  	 		  if (sign < 0) {
  	 			  
  	 			  double r = A.get(a).getX() + (((A.get(a).getY()) * (A.get((a+1)%A.size()).getX() - A.get(a).getX()))/ (A.get(a).getY() - A.get((a+1)%A.size()).getY()));
  	 			  
  	 			  if (r > 0) {
  	 				  if (A.get(a).getY() < 0) {
  	 					  wNum++;
  	 				  }
  	 				  else {
  	 					  wNum--;
  	 				  }
  	 			  }
  	 			  
  	 			  else if ((A.get(a).getY() == 0) && (A.get(a).getX() > 0)) {
  	 				  if (A.get((a+1)%A.size()).getY() > 0) {
  	 					  wNum = wNum + .5;
  	 				  }
  	 				  else {
  	 					  wNum = wNum - .5;
  	 				  }
  	 			  }
  	 			  else if ((A.get((a+1)%A.size()).getY() == 0) && (A.get((a+1)%A.size()).getX() > 0)) {
  	 				  if (A.get(a).getY() < 0) {
  	 					 wNum = wNum + .5;
  	 				  }
  	 				  else {
  						  wNum = wNum - .5;
  	 				  }
  	 			  }
  	 		  }
  	 	  }
  	 		return -1*wNum; // winding numbers off by factor of -1
  	 }
    
  	public void removeIntPoints(Point p) {//remove all intersection points of edges connected to p
  		while ( curve.contains(p)  ) {
  			int index = curve.indexOf(p);
  			curve.remove(index);
  		}
    	ArrayList<Integer> badIndices = new ArrayList<Integer>(); // all indices of edges to be deleted
		for (int i=0; i< edges.size(); i++) { 
			// check if either point on the edge is the point to be removed
			Edge curEdge = edges.get(i);
			if ( curEdge.p1 == null) { 
				System.out.println("error: edge has null 1st point!!");
			}
			if ( p == null ) {
				System.out.println("ERROR: Point input is null!");
			}
			if (curEdge.p1.equals(p) ) {
				badIndices.add( new Integer(i));
			}
			else if (curEdge.p2.equals(p) ) {
				 badIndices.add( new Integer(i));
			}
		} 
		// reverse order of indices, so deleting elements) doesn't affect indices
		Collections.sort(badIndices, Collections.reverseOrder());
		
		for (int j:badIndices) { // delete all edges connected to the newly deleted point
			Edge e = edges.get(j);
		    
		    int r = e.getIntPoints().size();
		    for (int k=r-1; k>=0; k--) {
				removeIntPoint(e.getIntPoint(k)); // delete intersection points on the edge b4 deleting edge
		    }
		}
    }
  	
  	 public void removeIntPoint(Point p) { // delete intersection point 
     	for (Edge E: p.getShortEdges() ) {
     		Point P;
     		if ( E.getP1().equals(p) ) {
     			P = E.getP2();
     		}
     		else {
     			P = E.getP1();
     		}
     		P.getShortEdges().remove(E);
     	}
     	for (Edge e: p.getEdges()) {
     		e.removeIntPoint(p);
     	}
     	intPoints.remove(p);
     	curve.remove(p);
     	curve.remove(p);
     }
  	 
  	 public void deletePoint(Point p) {
  		 int index = this.points.indexOf(p); 
  		 if ( index == -1) {
  			 try { 
  				 throw new Exception("Cannot remove the point. It is not on the curve " + this.toString());
  			 }
  			 catch (Exception E) {
  				 System.out.println(E);
  			 }
  			 return;
  		 }
  		 removePoint(p);
  		 if (this.points.size() == 2) {
  			 this.edges.clear();
  			 this.points.get(0).getEdges().clear();
  			 this.points.get(1).getEdges().clear();
  			 edgeAdd( this.points.get(0), this.points.get(1) );
  		 }
  		 else if (this.points.size() > 2) {
  			 if (index > 0) { // not the first point that was deleted
  				 if (index == this.points.size() ) { // last point was deleted
					// connect new last point to first point
					Edge e1 = edgeAdd(this.points.get(index-1), this.points.get(0) );
					findIntersections(e1);
				}
				else { // otherwise an internal (not first or last) edge was removed
					Edge e2 = edgeAdd( this.points.get(index), this.points.get(index-1) );
					findIntersections(e2);
				}
			}
			else { // first point was deleted
				Edge e3 = edgeAdd(this.points.get(0), this.points.get(this.points.size()-1) );
				findIntersections(e3);
			}
		}
  	 }
  	 
  	 public void removePoint(Point p) {   	
    	ArrayList<Integer> badIndices = new ArrayList<Integer>(); // all indices of edges to be deleted
		for (int i=0; i< edges.size(); i++) { 
			// check if either point on the edge is the point to be removed
			Edge curEdge = edges.get(i);
			if (curEdge.p1.equals(p) ) {
				badIndices.add( new Integer(i));
				Point v = curEdge.p2; //  vertex connected to removed vertex
				v.getEdges().remove(curEdge); // remove edge from other vertex's edge list
			}
			else if (curEdge.p2.equals(p) ) {
				 badIndices.add( new Integer(i));
				 Point v = curEdge.p1; //  vertex connected to removed vertex
			//	 System.out.println("Point " + v.toString() + " had " + v.getEdges().size() + " edges" );
				 v.getEdges().remove(curEdge); // remove edge from other vertex's edge list
			//	 System.out.println(" Now it has : " + v.getEdges().size());
			}
		} 
		
		// reverse order of indices, so deleting elements doesn't affect indices
		Collections.sort(badIndices, Collections.reverseOrder());
		
		for (int j:badIndices) { // delete all edges connected to the newly deleted point
			Edge e = edges.get(j);
		    
		    int r = e.getIntPoints().size();
		    for (int k=r-1; k>=0; k--) {
				removeIntPoint(e.getIntPoint(k)); // delete intersection points on the edge b4 deleting edge
		    }
		 
		    edges.remove(j); // remove the edge
		}
		points.remove(p);
		curve.remove(p);
    }
  	 
  	 public void findIntersections(Edge e) { // find and add all int points associated to e that are currently not found
 	   for (int r = 0; r < this.edges.size(); r++) {
 	       Edge current = (Edge)this.edges.get(r);
 	       // ensure edges don't share points
 	       if ( !e.getP1().equals(current.getP1() ) && !e.getP1().equals(current.getP2() ) &&
 	    		   ( !e.getP2().equals(current.getP1() ) && !e.getP2().equals( current.getP1() ) ) ) {  
 	       
 		       if (e.doesIntersect(current)) { // check if edges e and current intersect
 		    	   Point intersectionPoint = e.getIntersection(current);
 		           intersectionPoint.setIntersection(); // the point is an intersection point
 		           if ((intersectionPoint.getX() == 1000.0) && (intersectionPoint.getY() == 1000.0)) {
 		               System.out.println(" Edges of false intersection: " + e.toString() + " and " + current.toString());
 		               return;
 		            }
 		           
 		            Point Q = intersectionPoint.closestEqAllowed(this.startingIntPoints);
 		            Point p = intersectionPoint.closestEqAllowed(this.intPoints);
 		            double d = intersectionPoint.distance(p);
 		            
 		            if (d > 0 )  {
 		            	if ( Q.equals(intersectionPoint) ) {
 		            		intersectionPoint = Q;
 		            	}
 		            	current.addIntPoint(intersectionPoint);
	 		            e.addIntPoint(intersectionPoint); // update int point arrays of both edges 
	 		            this.intPoints.add(intersectionPoint); // add it to global array of int points
	 		            intersectionPoint.addEdge(e);
	 		            intersectionPoint.addEdge(current); // update edges of the intersection point
 		             }
 		             else { // degenerate case of multiple shared intersection points!!
 		            	 
 		            	 return;
 		             }
 		       	}
 	       }   
 	   }
    }
  	 
  	public Edge edgeAdd(Point p1, Point p2) {
    	// add edge to list, to arrayList of edges for each point
    	Edge e = new Edge(p1, p2);
    	edges.add(e);
        p1.addEdge(e);
    	p2.addEdge(e);
    	findIntersections(e);
    	return e;
    }
  	
  	 public Edge getEdge(Point u, Point v) { // get the edge connecting points u and v
     	for (Edge e: u.getEdges() ) {
     		if ( e.getP1().equals(v) || e.getP2().equals(v) ) {
     			return e;
     		}
     	} 
     	return null; // return null if the edge does not exist 
     } 
  	 
  	 public void process() {
  		 int runs = 0;
  		 while ( true ) {
  			 if ( runs > 91) {
  				 try{ 
  					 throw new Exception("Process of the curve not working.");
  				 }
  				 catch (Exception E ){ 
  					System.err.print(E);
  				 }
  			 }
  			 numberPoints(true);
  			 ArrayList<Face> isolatedMonogons  = new ArrayList<Face>();
  			 outerLoop: for (Face F: this.faces) {
  				 if ( F.getType() == 1 ) { // look for monogons 
  					 for (Point P: F.getPoints() ){
  						 if (P.faceMembership() == 1) { //see if the face is isolated 
  							 isolatedMonogons.add(F);
  							 break outerLoop;
  						 }
  					 }
  				 }
  			 }
  			 runs++;
  			 if ( isolatedMonogons.size() == 0 ) {
  				 return;
  			 }
  			 for (Face empty: isolatedMonogons) { 
				 this.TitusI(empty); // delete the empty monogons
  			 }
  		 }
  	 }
  	 
  	 public void numberPoints(boolean calcFaces) {
    	
		ArrayList<Edge> shortEdges = new ArrayList<Edge>();
    	
    	boolean findFaces = true;
    	for (Point p : curve) { 
    		if ( p.getEdges().size() >= 3) {
    			System.out.println("Point " + p.toString() + " is connected to " + p.getEdges().size() + " edges");
    			for (Edge E: p.getEdges() ) {
    				System.out.print(E.toString() + " ");
    			}
    			System.out.println("");
    			findFaces = false;
    		}
    		p.getFaces().clear();
    		p.removeConnections();
    		p.setNum(-1); // ensure all points are renumbered
    		p.getShortEdges().clear();
    	}
    	curve.clear();
    	faces.clear();
    	faceReps.clear();
    	int intPointCount = 0;
    	
    	if (points.size() == 0) { return; } // do nothing
    	
    	if ( intPoints.size() == 0 ) { //  NO intersection points yet, so we know the intersection sequence
    		for (int j=0; j< points.size() ; j++) {
    			Point cur = points.get(j);
    			Point next = points.get( (j+1) % points.size() );
    			points.get(j).setNum(j); // points have predictable numbers 
    			Edge e = new Edge(cur, next);
    			
    			Edge E = this.getEdge(cur, next);
    			if ( E.getP1()!= cur) { // check orientation of the edge
    				E.reverse();
    			}
 
    			Edge f = new Edge(next, cur);
    			cur.addShortEdge(e); // add correct short edges to both points' arraylists of short edges
    			next.addShortEdge(f);
    			shortEdges.add(e); // add edge to short edge 
    			shortEdges.add(f);
    			curve.add(cur);// order all points in allPoints array
    			
    		} 
    	}
    	else {   // there are MAYBE intersection points on the current edge
    		
    		int count = 0; // the number of points encountered so far 
    		boolean done = false;
        	for (int i=0; i< points.size() && !done; i++) {
        		Point cur = points.get(i);
        		Point next = points.get((i+1)%points.size());
        		
        		
        		Edge e = getEdge(cur, next);
        		
        		if ( e.getP1()!= cur) { // check orientation of the edge 
        			e.reverse();  // if edge is backwards, flip it
        		}
        		// if there are no intersection points on the current edge
        		if ( e.getIntPoints().size() == 0 ) { 
        			if ( cur.getNum() == -1 ) {// current point is NOT numbered
        				cur.setNum(count); 
            			count++; // number the point and update the count 
        			}
        			Edge g = new Edge(cur, next); // create directed edges between points 
        			Edge h = new Edge(next, cur);
        			cur.addShortEdge(g);
        			next.addShortEdge(h);
        			shortEdges.add(g); // add to list of all edges
        			shortEdges.add(h);  
        			//intSeq += cur.getNum() + "."; // add the # of cur to the int seq.
        			curve.add(cur);
        			if ( !next.equals(points.get(0)) && next.getNum() == -1 ) {
        				next.setNum(count); // if we aren't at the start again, next has no #, then # it
        				count++; //update count
        			}
        			
        		}
        		else { //THERE ARE INTERSECTION POINTS!
        			
        			ArrayList<Point> A = new ArrayList<Point>();
            		for (Point u: e.getIntPoints() ) { // all AVAIALABLE points to use, (don't re-use)
            			A.add(u);
            		}
            		A.add(e.p1); A.add(e.p2); // add endpoints of the edge
            		while (true) {        	
            			if (cur.getNum() == -1) { // if point is currently unreached
            				cur.setNum(count); // number the point, update count
            				count++;
            			}
            			//intSeq += cur.getNum() + ".";  // update intersection sequence 
            			if ( !contains(curve, cur) ) {
            				if ( cur.isIntersection() ) { 
            					cur.setIntPointIndex(intPointCount);
            					intPointCount++;
            				}
            				curve.add(cur);
            			}		
            			else {
            				curve.add(cur); // always add the current point to "curve" array
            			}
            			
            			if (cur.isIntersection()) {
            				int index = cur.getIntPointIndex();
     
            			}
            			A.remove(cur); // don't allow reuses 
            			Point upNext = cur.closest(A);
            			
            			Edge newEdge1 = new Edge(cur, upNext);
            			Edge newEdge2 = new Edge(upNext, cur);
            			// if either point is an intersection point, then we have an interior edge 
            			
            			shortEdges.add(newEdge1); 
        				shortEdges.add(newEdge2);
            			cur.addShortEdge(newEdge1); // add directed edges for cur and upNext 
            			upNext.addShortEdge(newEdge2);
            			cur = upNext;
            			if (cur.isIntersection() ) {
            				if ( cur.getVisits() == 0 || cur.getVisits() == 1) {
            					cur.addVector(new Edge(cur, next));
            					cur.setIndex(count-1); // remove one b/c count was already incremented
            				}
                			cur.visit();
                			if ( cur.getVisits() == 2 ) {
                				cur.calcSign();
                			}
                		}
            			if ( cur.equals( points.get(0) ) ) { // finished traversal of curve
            				done = true;
            				break;
            			}
            			else if ( cur.equals(next) ) { // the whole edge [p(i)->p(i+1)] has been traversed 
            				cur.setNum(count);
            				count++;
            				break; // exit the while loop 
            			}
            		}
        		}
        	}
    	}
    	
    	
    	if ( points.size()>=3 && findFaces && calcFaces) { // CALCULATE FACES!!
    		faceReps.clear();
    			
	    	for ( Edge E: shortEdges) {  // generate faces from interior edges 
	    		Face f = this.getFace(E);
	    	
		    	if ( f != null &&  !contains(faces, f) ) {  // ensure face is nontrivial and new
	    			faces.add(f);
	    			f.getNewRep();
	    			faceReps.add(f.getRep()); // add face rep
		    	}
	    	}
	    }
    	
		if ( faces.size() > 1) { 
			double maxArea = 0;
			Face badFace = null;
			for (Face f: faces) { // delete degenerate face of exterior of the graph
				if ( f.getArea() > maxArea) { // it is the face with maximal area
					maxArea = f.getArea();
					badFace = f;
				}
			}
			//System.out.println("Bad face was: " + badFace.toString());
			faceReps.remove(badFace.getRep());
			faces.remove(badFace);
		}
		
		windingArea = 0;
		
		for (Point P: curve) {
			P.getFaces().clear();
    		P.removeConnections();
		} 
		
		int count = 0;
		ArrayList<Face> facesCopy = new ArrayList<Face>();
		int size = faces.size();
		ArrayList<Color> colors = new ArrayList<Color>();
		for (int j=0; j<faces.size(); j++) {
			Face F = faces.get(j);
			Color c = new Color( 20 + (int)(110/size*j), 40 + (int) (210/size * (j+1) ), (int)(250 - 180/(size) *(size-j)) );
			F.setColor(c);
			colors.add(c);
			//F.setColor(c); 
			facesCopy.add(F); // add to the copy
			F.setNum(count); // number the face
			count++;
			F.incrementPoints(); // increase face count for each point on the face
			if ( F.isCW() ) {
				F.reverse();
			}
		}

		if ( calcFaces)  {
			windingArea = 0;
			for ( Face f: faces ){
				try {
					if ( f.getPolygon().contains(f.getRep()) ){
						double wNum = this.windingNum( f.getRep() );
						int wNumInt = (int)wNum;
						if ( Math.round(wNum) !=  wNumInt) {
							try { throw new Exception("winding number not an integer");}
							catch (Exception EEE) { System.err.println(EEE); }
						}
						f.setWindingNum( wNumInt );
					}
					else {
						throw new Exception("Face rep not inside the face for " + f.toString());
					}
				}
				catch (  Exception EE ) { System.err.println(EE); } 
				windingArea += Math.abs(f.getWindingNum()) * f.getArea();
			}
			windingArea = round(windingArea, 2);
		}

    } 
  
  	
  	public void TitusI(Face F) {
    	
    	try {
    		if ( F.getType() != 1) {
    			throw new Exception("Cannot apply Ia move. Face is not a monogon.");
        	}
    	}
    	catch ( Exception EE) {
    		System.err.println(EE);
    		return;
    	}
    	
    	boolean zeroRemoved = false; //remember if the 0th point was deleted or not
    	Point intPoint = null; // the intersection point from the face 
    	 // indices of the 
    	ArrayList<Integer> indices = new ArrayList<Integer>();
    	for (int i=F.getSize()-1; i>=0 ; i--) { // look through p
    		Point P = F.getPoint(i);
    		if ( !P.isIntersection() ) { // save indices of non-int points 
    			indices.add( new Integer( points.indexOf(P) ) );
    		}
    		else {
    			intPoint = P;
    		}
    	}
    	try {
	    	if ( intPoint == null) {
	    		throw new Exception("NULL INT POINT DURING TITUS Ia MOVE");
	    	}
    	}
    	catch ( Exception EE) {
    		System.err.println(EE);
    	}
    	
    	Collections.sort(indices); // sort indices of all non-int points
    	Collections.reverse(indices); // then reverse so they are in decreasing order 
    	// points connected to the one int point of the face
    	ArrayList<Point> connections = new ArrayList<Point>();
    	// edges of the intersection point
    	ArrayList<Edge> edgeC = new ArrayList<Edge>();
    	for ( Edge e: intPoint.getShortEdges() ) {
    		connections.add(e.getP2());
    		edgeC.add(e);
    	}
    	for (Integer k: indices) {
    		int j = k.intValue();
    		if ( j==0) { zeroRemoved = true; } 
    		
    		if (connections.contains(points.get(j))) {
    			int n = connections.indexOf(points.get(j));
    			Edge E = edgeC.get(n);
    			intPoint.removeShortEdge( E );
    			connections.remove(n);
    			edgeC.remove(n);
    		}
    		
    		removePoint(points.get(j)); // remove the point 
    	}
    	
    	ArrayList<Edge> localEdges = intPoint.getShortEdges();
    	if (localEdges.size()!= 2) {
    		System.out.println("Error! Titus move not working. Only "+localEdges.size() +" edges");
    		return;
    	}
    	else {
    		// get the points connected to the int point, make new edges,  
    		// 					add the int point to the points array 
    		intPoint.notIntersection();
    		Edge e1 = intPoint.getEdge(0);
    		Edge e2 = intPoint.getEdge(1);
    		Point P1;
    		Point P2;
    		// get the points that need to be connected to the int point of F
    		if ( points.contains(e1.getP1() ) ) { P1 =  e1.getP1(); } else { P1 = e1.getP2(); }
    		if ( points.contains(e2.getP1() ) ) { P2 =  e2.getP1(); } else { P2 = e2.getP2(); }
    		int j;
    		// get the larger index of the two non-int points P1,P2 
    		// if we have intpoint -> 0, then treat 0 as max value ( 0 ≈ points.size() (modular) )
    		if ( intPoint.getVector(1).getP2().equals(points.get(0)) ) {
    			j = 0;
    		}
    		else {
    			j = Math.max( points.indexOf(P1), points.indexOf(P2) );
    		}
    		// reset the edges of the int point
    		intPoint.getEdges().clear();
    		intPoint.getShortEdges().clear();
    		
    		if ( zeroRemoved ) { points.add(0, intPoint); }
    		// insert the int point between both points, then
    		// add edges both ways
    		else { points.add(j, intPoint); }
    		edgeAdd(intPoint, P1);
    		edgeAdd(intPoint, P2);
    		intPoint.addShortEdge(new Edge(intPoint, P1));
    		intPoint.addShortEdge(new Edge(intPoint, P2));
    	}
    	boolean g1 = faceReps.remove(F.getRep());
    	boolean g2 = faces.remove(F);
    	if ( !g1 || !g2 ) { // ensure the face F and its rep were removed 
    		try {
	    		if ( ! g1 ) {  throw new Exception("Face rep not removed during Titus Ia move!"); }
	    		else if ( ! g2 ){ throw new Exception("Face F not removed during Titus Ia move!"); }
	    		else {
	    			throw new Exception("Neither face nor face rep removed during Titus Ia move!");
	    		}
    		}
    		catch (Exception EE) {
    			System.err.println(EE);
    		}
    	}
    	areaSwept += F.getArea();  // keep track of the area of the face 
    	//umberPoints(true);
    }
  	
    public double windingArea() {
    	
    	ArrayList<Point> points = new ArrayList<Point>(); 
    	ArrayList<Point> intPoints = new ArrayList<Point>();
    	ArrayList<Point> allPoints = new ArrayList<Point>();
    	ArrayList<Edge> edges = new ArrayList<Edge>();
    	
    	for (Point P: curve) {
    		if ( !contains( (P.isIntersection()  ? intPoints : points), P) ) {
    			(P.isIntersection()  ? intPoints : points).add(P);
    		}
    		allPoints.add(P);
    		P.resetIndices();
    	}
    	int outOfOrder = 0;
    	for (int j= 0; j< curve.size(); j++) {
    		Point now = curve.get(j);
    		Point next = curve.get( (j+1)%curve.size() );
    		if ( ! adjacent(now, next) ) {
    			outOfOrder ++;
    		}
    	} 
    	if ( outOfOrder > 0 ) { 
    		try { 
    			throw new Exception("There were "  + outOfOrder +  " points out of order" );
    		}
    		catch (Exception EE) { 
    			System.out.println(EE);
    			return -1;
    		}
    	}
    	for (Point p : curve) { 
    		if ( p.getEdges().size() >= 3) {
    			System.out.println("Point " + p.toString() + " is connected to " + p.getEdges().size() + " edges");
    			for (Edge E: p.getEdges() ) {
    				System.out.print(E.toString() + " ");
    			}
    			try { 
    				throw new Exception("Point has more than 3 edges.");
    			}
    			catch (Exception E) { System.err.println(E); return -1; }
    		}
    		
    	}
    	
    	if (intPoints.size() == 0) {
    		ArrayList<Edge> E = new ArrayList<Edge>();
    		for (int i=0; i<points.size(); i++) {
    			Point now = points.get(i);
    			Point next = points.get( (i+1)% points.size() );
    			E.add(new Edge(now, next));
    		}
    		Face F = new Face(E);
    		return F.getArea();
    	}
    	ArrayList<Face> faces = new ArrayList<Face>();
    	ArrayList<Edge> shortEdges = new ArrayList<Edge>(); 
 
    	if (points.size() <= 2) { return 0; } // no area
    	
    	if ( intPoints.size() == 0 ) { //  NO intersection points yet, so we know the intersection sequence
    		for (int j=0; j< points.size() ; j++) {
    			Point cur = points.get(j);
    			Point next = points.get( (j+1) % points.size() );
    			Edge e = new Edge(cur, next);
    			edges.add(e); // keep track of all global edges 
    			addEdge(e);  // ensure each point on the edge also has the edge in its edges array
    			Edge f = new Edge(next, cur);
    			// add correct short edges to both points' arraylists of short
    			if ( !contains(cur.getShortEdges(), e) ) { cur.addShortEdge(e); }
    			if ( !contains(next.getShortEdges(), f) ) { next.addShortEdge(f); }
    			shortEdges.add(e); // add edge to short edge 
    			shortEdges.add(f);
    		} 
    	}
    	else {   // there are MAYBE intersection points on the current edge
    		
    		int count = 0; // the number of points encountered so far 
    		boolean done = false;
        	for (int i=0; i< points.size() && !done; i++) {
        		Point cur = points.get(i);
        		Point next = points.get((i+1)%points.size());
        		
        		int l = curve.indexOf(cur);
        		int m = curve.indexOf(next);
        		if ( m == 0) { m = curve.size(); } // we are doing math mod curve.size()
        		Edge e = new Edge(cur, next);
        		edges.add(e); // store the edge in the local edges 
        		
        		for (int n = l+1; n < m ; n++) {// all points btwn cur + next in allPoints are int points
        			e.addIntPoint( curve.get(n) );
        		}
        		
        		addEdge(e); // ensure all points (end + int ) points have the edge stored as well
        		
        		// if there are no intersection points on the current edge
        		if ( e.getIntPoints().size() == 0 ) { 
        			if ( cur.getNum() == -1 ) {// current point is NOT numbered
        				//cur.setNum(count); 
            			count++; // number the point and update the count 
        			}
        			Edge g = new Edge(cur, next); // create directed edges between points 
        			Edge h = new Edge(next, cur);
        			if ( !contains(cur.getShortEdges(), g) ) { cur.addShortEdge(g); }
        			if ( !contains(next.getShortEdges(), h) ) { next.addShortEdge(h); }
        			shortEdges.add(g); // add to list of all edges
        			shortEdges.add(h);  
        			
        			if ( !next.equals(points.get(0)) && next.getNum() == -1 ) {
        				//next.setNum(count); // if we aren't at the start again, next has no #, then # it
        				count++; //update count
        			}
        			
        		}
        		else { //THERE ARE INTERSECTION POINTS!
        			
        			ArrayList<Point> A = new ArrayList<Point>();
            		for (Point u: e.getIntPoints() ) { // all AVAIALABLE points to use, (don't re-use)
            			A.add(u);
            		}
            		A.add( e.getP1() ); A.add( e.getP2() ); // add endpoints of the edge
            		while (true) {        	
            			if (cur.getNum() == -1) { // if point is currently unreached
            			//	cur.setNum(count); // number the point, update count
            				count++;
            			}
            			
            			A.remove(cur); // don't allow reuses 
            			Point upNext = cur.closest(A);
            			
            			Edge newEdge1 = new Edge(cur, upNext);
            			Edge newEdge2 = new Edge(upNext, cur);
            			// if either point is an intersection point, then we have an interior edge 
            			
            			shortEdges.add(newEdge1); 
        				shortEdges.add(newEdge2);
        				// add directed edges for cur and upNext 
        				if ( !contains(cur.getShortEdges(), newEdge1) ) { cur.addShortEdge(newEdge1); }
            			if ( !contains(upNext.getShortEdges(), newEdge2) ) { upNext.addShortEdge(newEdge2); }
            			cur = upNext;
            			if (cur.isIntersection() ) {
            				if ( cur.getVisits() == 0 || cur.getVisits() == 1) {
            					cur.addVector(new Edge(cur, next));
            					cur.setIndex(count-1); // remove one b/c count was already incremented
            				}
                			cur.visit();
                			if ( cur.getVisits() == 2 ) {
                				cur.calcSign();
                			}
                		}
            			if ( cur.equals( points.get(0) ) ) { // finished traversal of curve
            				//done = true;
            				break;
            			}
            			else if ( cur.equals(next) ) { // the whole edge [p(i)->p(i+1)] has been traversed 
            			//	cur.setNum(count);
            				count++;
            				break; // exit the while loop 
            			}
            		}
        		}
        	}
    	}
    	
    	for (Point P: allPoints){
    		P.getShortEdges().clear();
    	}
    	for (int i=0; i< curve.size(); i++) {
    		Point cur = curve.get(i);
    		Point next = curve.get( (i+1) % curve.size() );
    		Edge forward = new Edge(cur, next);
    		Edge backward = new Edge(next, cur);
    		if ( !contains(cur.getShortEdges(), forward) ) { 
    			cur.addShortEdge(forward);
    		}
    		if ( !contains(next.getShortEdges(), backward) ) { 
    			next.addShortEdge(backward);
    		}
    		
    	}
    	
    	
    	Point last = allPoints.get(allPoints.size()- 1); 
    	
    	if ( last.isIntersection() ) {
    		Point prev = new Point(); Point next = new Point();
    		for (int j = allPoints.size()-2; j >=0; j--) { // find the first non-intersection point BEFORE the last point 
    			Point Z = allPoints.get(j);
    			if ( !Z.isIntersection() ) { prev = Z; break; } 
    		}
    		for (int k=0; k < allPoints.size()- 2; k++) { // find the first non-intersection point AFTER the last point
    			Point Z = allPoints.get(k);
    			if ( !Z.isIntersection() ) { next = Z; break; }
    		}
    		Edge E = new Edge(prev, next);
    		if ( !contains(last.getEdges(), E) ) { last.addEdge( E); }
    		
    	}
    	int a =0;
    	for (Point P: allPoints) {
    		if ( (P.isIntersection() && P.getShortEdges().size() != 4) ||
    				( !P.isIntersection() && P.getEdges().size() !=2 ) ) {
    			a++;
    		}
    	}
    	try { 
    		if ( a > 0 ) { 
    			throw new Exception("Error. Short Edges not correctly added.");
    		}
    	}
    	catch ( Exception EE) { System.err.println(EE);  }
    	for (Point P: allPoints) {
    		if ( P.getEdges().size() < 2) { 
    			try {
    				throw new Exception("Edges weren't fully added.");
    			}
    			catch (Exception EE) { return -1; }
    		}
    	}
    	for ( Edge E: shortEdges) {  // generate faces from interior edges 
    		Face f = getFace(E);
    	
	    	if ( f != null &&  !contains(faces, f) ) {  // ensure face is nontrivial and new
    			faces.add(f);
	    	}
    	}
    	
    	
		if ( faces.size() > 1) { 
			double maxArea = 0; 
			Face badFace = null;
			for (Face f: faces) { // delete degenerate face of exterior of the graph
				if ( f.getArea() > maxArea) { // it is the face with maximal area
					maxArea = f.getArea();
					badFace = f;
				}
			}
			faces.remove(badFace);
		}

		double wArea = 0;
		
		for (Point P: allPoints) {
			P.getFaces().clear();
    		P.removeConnections();
		} 
		
		int count = 0;
		ArrayList<Face> facesCopy = new ArrayList<Face>();
		for (Face F: faces) {
			F.getNewRep();
			facesCopy.add(F); // add to the copy
			F.setNum(count); // number the face
			count++;
			F.incrementPoints(); // increase face count for each point on the face
			if ( F.isCW() ) {
				F.reverse();
			}
		}
		
		/**
		ArrayList<Face> wDone = new ArrayList<Face>();
		
		int outerPoints = 0;
		
		for ( Point p: points ) {
			// look for edges on an outer face 
			if ( p.faceMembership() == 1) {
				Face f = faces.get((int)p.getFaces().get(0));
				if ( facesCopy.contains(f) ) {
					outerPoints++;
					Edge e = ( contains(edges, p.getShortEdge(0) ) )? p.getShortEdge(0) : p.getShortEdge(1);
					f.setWindingNum( winding(f, e, edges) );
					facesCopy.remove(f);
					wDone.add(f);
				}
			}
		}
		if ( outerPoints == 0) {
			try { 
				throw new Exception("No outer points, so no faces to start with.");
			}
			catch ( Exception EE) { System.err.print(EE); return -1; }
		}
		count = 0;
		
		while ( facesCopy.size() > 0 ) {
			
			if ( facesCopy.size() == 1 ) { // check for quadruple triangle curve ( like quadT)
				Face F = facesCopy.get(0);
				if (F.getType() == 3 && F.getPoints().size() == 3) {
					Edge E = F.getEdges().get(0);
					F.setWindingNum( winding(F, E, edges ) ); //set the winding number
					facesCopy.remove(F);
				}
			}
			
			if (  count > 10000 ) {
				try {
					throw new newException("ERROR 10000 loops through faces and didn't find windingNums");
				}
				catch (newException EE) {
					for ( Face F: faces) {
						F.setWindingNum(99);
					}
					System.err.println(EE);
					for (Face F: facesCopy) {
						System.out.println("Could not find the windingNum for " + F.toString());
					}
					System.exit(0);
					return -1;
					
				}
			}
			count++;
			ArrayList<Face> remove = new ArrayList<Face>();
			for (Face f: facesCopy) { // look through all faces that don't have winding #'s
				boolean done = false;
				for (Edge e: f.getEdges()) { // look through each edge of that face
					if ( !done) { 
						for (Face F: wDone) { // see if the edge is common to a face with a winding #
							if ( contains(F.getEdges(), e)) {
								done = true; // if so, compute winding # for the new face f
								remove.add(f);
								int newW = F.getWindingNum() + winding(f, e, edges );
								
								f.setWindingNum( newW );
								break;
							}
						}
					}
				}
			}
			for ( Face face: remove) { //remove the faces from incomplete array, add them to done array
				facesCopy.remove(face);
				wDone.add(face);
			}
		}	
		wArea  = 0;
		for ( Face f: faces ){
			wArea += Math.abs(f.getWindingNum()) * f.getArea();
		}
		double w = wArea * 10000;
		w = Math.round(w);
		w = w/10000;
		wArea = w;
		**/
			
		wArea = 0;
		for ( Face f: faces ){
			try {
				if ( f.getPolygon().contains(f.getRep()) ){
					double wNum = this.windingNum( f.getRep());
					int wNumInt = (int)wNum;
					if ( Math.round(wNum) !=  wNumInt) {
						try { throw new Exception("Winding number not an integer");}
						catch (Exception EEE) { System.err.println(EEE); }
					}
					f.setWindingNum( wNumInt );
				}
				else {
					throw new Exception("Face rep not inside the face for " + f.toString());
				}
			}
			catch (  Exception EE ) { System.err.println(EE); } 
			wArea += Math.abs(f.getWindingNum()) * f.getArea();
		}
		wArea = round(wArea, 3);
		windingArea = wArea;
		return wArea;
    }
    
    public Curve reverse() {
    	ArrayList<Point> reverseCurve = new ArrayList<Point>();
    	for (int j = curve.size()-1; j >= 0; j--) {
    		reverseCurve.add( curve.get(j) );
    	}
    	return new Curve(reverseCurve);
    }
    public static double round(double value, int places) {
        if (places < 0) throw new IllegalArgumentException();

        BigDecimal bd = new BigDecimal(value);
        bd = bd.setScale(places, RoundingMode.HALF_UP);
        return bd.doubleValue();
    }
    public String toString() {
    	String s = "";
    	for (Point P: curve){ 
    		if ( P.toString().contains("i")) {
    			s += P.toString() + "-";
    		}
    	}
    	s = s.substring(0, s.length());
    	return s;
    }

	
} 
