import javax.swing.*;

import java.math.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;


public class DrawingPanel extends JPanel{
	static final int RADIUS = 10; // radius of points for drawing them 
	static final int THRESHOLD = 10; // distance threshold for deleting point from right-click
	//public ArrayList<Point> points = new ArrayList<Point>(); // non intersection points 
	//public ArrayList<Point> intPoints = new ArrayList<Point>(); // intersection points
	
	//public ArrayList<Edge> edges = new ArrayList<Edge>(); // full edges between non intersection points
	public ArrayList<Edge> shortEdges = new ArrayList<Edge>();
	public ArrayList<Edge> directedEdges = new ArrayList<Edge>(); // local directed edges that traverse whole curve
	//public ArrayList<Face> faces = new ArrayList<Face>(); 
	public ArrayList<Point> faceReps = new ArrayList<Point>();
	public ArrayList<Point> allPoints = new ArrayList<Point>(); // all points in order of intersection #
	public Point curPoint; // point being moved around while the mouse is dragged 
	public String intSeq = "";
	public boolean print = false;
	public boolean built = false; // assume curve isn't input yet 
	public ArrayList<Point> pointsSelected = new ArrayList<Point>();
	public boolean justFinished = false;
	public double areaSwept = 0; // area swept by Titus moves
	public boolean complete = false; 
	public boolean overlapping; // if the curve is self-overlapping
	
	static final int diameter = 9; // for drawing edges 
	static final double arrowWidth = 2.5;
	static final int arrowLength = 10; 
	public double windingArea;  // winding area shown through the process of Titus moves 
	public double startingWindingArea = -1; // winding area of the completed curve 
	public ArrayList<String> titusSeq = new ArrayList<String>();
	public double bestArea = Integer.MAX_VALUE;
	public ArrayList<String> bestTitus = new ArrayList<String>(); // string for showing best Titus moves
	
	public boolean addedBox = false;
	public boolean allowMovement = false;
	public boolean showPointNums = false;
	public boolean colorFaces = false;
	public boolean moving;
	public EastPanel east;
	public DrawingPanelFrame frame;
	
	public Curve Curve = new Curve();
	
	// the intersection sequence (just the int point indices)
	public ArrayList<Integer> intSeqIntegers = new ArrayList<Integer>();
	// the WHOLE curve sequence, both int and non-int points
	public ArrayList<Integer> intersectionSeq = new ArrayList<Integer>();
	public int faceSide = -1;
	public int highlightFaces = -1;
	public Color highlightColor = new Color(255, 255, 255);
	public Face selectedFace;
	public boolean pointNew = false;
	public double minimumHomotopyArea = -1;
	public Color RED = new Color(233, 0, 0);
	public Color PURPLE = new Color(159, 56, 255);
	public int baseCases = 0;
	public Point smoothPoint;
	public boolean smoothDraw;
	public boolean go = false;
	
	public double falseWindingArea = -1;
	
	
	Node root;
	ArrayList< Curve  > backupOverlappingCurves  = new ArrayList< Curve >(); // self-overlapping curves from the minimum homotopy
	ArrayList< Curve  > selfOverlappingCurves  = new ArrayList< Curve >(); // self-overlapping curves from the minimum homotopy
	
	
	ArrayList<Integer> anchorIndices = new ArrayList<Integer>(); // depths of anchor points in the tree
	ArrayList<Point> anchorPoints = new ArrayList<Point>(); // store them in order (L to R)
	ArrayList<Point> originalAnchors = new ArrayList<Point>();
	int homotopyCalls = 0;
	int extraSplits = -1;
	public double avgMaxSplits = -1;
	public double avgSplits = -1;
	int splits = -1;
	
	public static final int anchorMovement = 5;
	
	public DrawingPanel(DrawingPanelFrame frame){
		this.frame = frame;
		MyMouseAdapter mouseAdapter = new MyMouseAdapter();
		addMouseListener (mouseAdapter);
		addMouseMotionListener(mouseAdapter);
		this.setBackground(new Color(235, 235, 235) );
	}
	
	public void drawEdge(Graphics graphics, Edge E, int weight){
		 // directed edge from v -> u
		Point v = E.getP1();
		Point u = E.getP2();
	
		double l = Math.sqrt((u.getX()-v.getX())*(u.getX()-v.getX())+(u.getY()-v.getY())*(u.getY()-v.getY()));
	      graphics.setColor(Color.BLACK);  	       
	      // arrow
	      Polygon arrowhead = new Polygon();
	      int a = (int) (v.getX()-u.getX());
	      int b = (int) (v.getY()-u.getY());
	
	      arrowhead.addPoint((int)(u.getX() + a/5) , (int) (u.getY() + b/5 )); 
	      arrowhead.addPoint((int)(v.getX()+(l-diameter/2-arrowLength)*(u.getX()-v.getX())/l-2*arrowWidth*(u.getY()-v.getY())/l + a/5), 
	    		               (int)(v.getY()+(l-diameter/2-arrowLength)*(u.getY()-v.getY())/l+2*arrowWidth*(u.getX()-v.getX())/l + b/5)); 
	      arrowhead.addPoint((int)(v.getX()+(l-diameter/2-arrowLength)*(u.getX()-v.getX())/l+2*arrowWidth*(u.getY()-v.getY())/l + a/5), 
	               (int)(v.getY()+(l-diameter/2-arrowLength)*(u.getY()-v.getY())/l-2*arrowWidth*(u.getX()-v.getX())/l + b/5)); 
	      
	      graphics.fillPolygon(arrowhead);
	  }  

    protected void paintComponent(Graphics graphics){
    	
    	ArrayList<Point> curve = Curve.curve; 
    	
    	super.paintComponent(graphics);
    	Graphics2D g = (Graphics2D) graphics;
    	
    	if ( colorFaces )  {
	    	for (int l=0; l< Curve.faces.size(); l++)  {
				Face f = Curve.faces.get(l);
				if ( f!= null) {
					Color c = f.getColor();
	
					if ( highlightFaces == f.getType() ) {
						c = highlightColor;
					}
					if ( selectedFace !=null && f.equals(selectedFace) ) {
						c = Color.black;
					}
					graphics.setColor(c);
					graphics.fillPolygon(f.getPolygon());
					
				}
	    	}	
		} 

    	// ONLY DRAW EDGES IF THEY ARE NOT FROM THE SELECTED FACE

    	for(int k=0; k< Curve.edges.size(); k++){ 
			g.setPaint(Color.BLACK);
			Edge e1 = Curve.edges.get(k);
			//Edge e1 = (selectedFace == null || !selectedFace.getEdges().contains(Curve.edges.get(k)) ) ?
		//			Curve.edges.get(k) : null;
			if ( e1 != null) {
	   		  	int x1 = (int)e1.getP1().getX();
	    		int y1 = (int)e1.getP1().getY();
	   		  	int x2 = (int)e1.getP2().getX();
	    		int y2 = (int)e1.getP2().getY(); 
	    		g.setStroke( new BasicStroke(2) );
	    		g.drawLine(x1,y1,x2,y2); 
    		}
			
    	}
    	
    	if (curPoint != null) { // draw point being moved around
			if ( !built && Curve.points.size() == 0 || curPoint.equals(Curve.points.get( Curve.points.size()-1) ) ) {
				g.setColor(Color.YELLOW); // color yellow if moving the newest point
			}
			else {
				g.setColor(Color.BLACK);
			}
			
			int x = (int)curPoint.getX()-(RADIUS/2);
    		int y = (int)curPoint.getY()-(RADIUS/2);
    		if ( Curve.points.contains(curPoint) ) {
    			g.fillOval(x,y,RADIUS,RADIUS);
    		}
		}
    	
    	if ( selectedFace != null && !contains(Curve.faces, selectedFace)  ) {
    		selectedFace = null;
    	}

    	//color the edges of the highlighted face differently 
    	
    	if ( ! showPointNums && Curve.points.size() >= 1) { 
    		g.setColor(Color.BLACK);
    		graphics.drawString("0", (int)Curve.points.get(0).getX()-4, (int)Curve.points.get(0).getY()-RADIUS/2 );
    	}
    	
    	if (built) {
    		g.setColor( Color.YELLOW);
    		g.fillOval( (int)Curve.points.get(0).getX()-RADIUS/2, (int)Curve.points.get(0).getY()-RADIUS/2, RADIUS, RADIUS );
    	}
    	
    	for (Curve C: selfOverlappingCurves ) {
    		if ( C !=null) {
				ArrayList<Edge> edges = C.getEdges();
				Color c = C.getColor();
				
				if ( c.getRed() < 60 && c.getBlue() < 60 && c.getGreen() < 60 ) {
					Color Co = new Color( 150 + (int)(Math.random()*100), 200, 0);
					C.anchor.color = Co;
					C.setColor(Co);
				} 
				c = C.getColor(); // get the (possibly updated) color 
				g.setPaint(c);
				//g.setPaint(new Color((int)(Math.random()*255), (int)(Math.random()*255), (int)(80+150*Math.random()) ));
				for (Edge E: edges) {
					int x1 = (int)E.getP1().getX();
		    		int y1 = (int)E.getP1().getY();
		   		  	int x2 = (int)E.getP2().getX();
		    		int y2 = (int)E.getP2().getY(); 
		    		g.setStroke( new BasicStroke(3) );
		    		g.drawLine(x1,y1,x2,y2); 
				}
    		}
		}
    	
		if ( selectedFace != null ) { // color the edges correctly for the selected face
			ArrayList<Point> first,second,third;
			first = second = third = new ArrayList<Point>();
			
			if ( selectedFace.getType() == 1 ) {
				first = selectedFace.getPoints();
			}
			else if ( selectedFace.getType() == 2 ) {
				first = selectedFace.splitFirst();
				second = selectedFace.splitSecond();
			}
			else if ( selectedFace.getType() == 3 ) {
				first = selectedFace.split(0);
				second = selectedFace.split(1);
				third = selectedFace.split(2);
			}
			
			ButtonPanel B = frame.east.buttonPanel;
			ArrayList< ArrayList<Point> > sections = new ArrayList< ArrayList<Point> >();
			sections.add(first); sections.add(second); sections.add(third);
			ArrayList<Color> colors = new ArrayList<Color>(); 
			// colors in order corresponding to color of each section
			colors.add(B.PINK); colors.add(B.GREEN); colors.add(B.ORANGE);
			g.setStroke( new BasicStroke(4) );
			
			for (int k=0; k<3; k++) {
				g.setColor(colors.get(k)); // get the current color
				ArrayList<Point> section = sections.get(k); // get the current section 
				for (int i=0; i< ( (selectedFace.getType()==1) ? section.size(): section.size()-1); i++) {
					int j = (i+1) % section.size(); // get the point after the i^th point
					int x1 = (int)section.get(i).getX();
		    		int y1 = (int)section.get(i).getY();
		   		  	int x2 = (int)section.get(j).getX();
		    		int y2 = (int)section.get(j).getY();
		    		g.drawLine(x1,y1,x2,y2); // draw the edge between them in the appropriate color 
				}
			}
		}

    	g.setPaint (Color.BLACK);
    	
    	Point curr = null;
    	Point prev = null;
    	int prevIndex = -1;
    	int curIndex = -1;
    	g.setPaint(Color.BLACK);
    	if (built && !moving) {
    		int startIndex  = -1;
    		Point P = null;
	    	for (int i=0; i< Curve.curve.size(); i++) {
	    		if ( curve.get(i).isIntersection() ) {
	    			P = curve.get(i);
	    			if ( curr == null) {
	    				curr = P;
	    				curIndex = i;
	    				startIndex = i;
	    			}
	    			else {
	    				prev = curr; // update current and previous int points
	    				prevIndex = curIndex; // move index to
	    				curr = P;
	    				curIndex = i;
	    				int mid1 = Math.floorDiv(  curIndex + prevIndex, 2);
	    				int mid2 = mid1+1;
	    				Edge e1 = new Edge(curve.get(mid1), curve.get(mid2) ) ;
	    				e1 = (selectedFace == null || !selectedFace.getEdges().contains(e1) ) ? e1 : null;
	    				if ( e1 != null) {
	    		   		  	drawEdge(g, e1, 3);
	    	    		}
	    			}
	    		}
	    	}
	    	if ( Curve.intPoints.size() > 0 ) {
	    		startIndex = startIndex + curve.size(); 
	    		int stopIndex = curIndex; // use index of last int point
	    		int mid1 = Math.floorDiv(  startIndex + stopIndex, 2);
	    		mid1 = mid1 % (curve.size());
				int mid2 = mid1+1;
				mid2 = mid2 % curve.size();
				Edge e1 = new Edge(curve.get(mid1), curve.get(mid2) ) ;
				e1 = (selectedFace == null || !selectedFace.getEdges().contains(e1) ) ? e1 : null;
				if ( e1 != null) {
		   		  	drawEdge(g, e1, 3);
	    		}
	    		
	    	}
    	}
    	
    	
		for(int i=0; i<Curve.points.size(); i++){
			boolean selected;
			
			g.setPaint(Color.BLACK);
			if ( !built && i == Curve.points.size()-1 ) {
				g.setPaint(Color.green);
			}
			if ( i ==0 ) {
				g.setPaint(Color.YELLOW);
			}
			Point point = Curve.points.get(i);
			if ( point.isSelected() ) { 
				g.setPaint(Color.CYAN);
				selected = true;
			} 
			
			if ( anchorPoints.size() > 0 && i == 0) {
				g.setPaint(PURPLE);
			}
			selected = point.isSelected();
			
   		  	int x = (int)point.getX()-(RADIUS/2);
    		int y = (int)point.getY()-(RADIUS/2);
    		if ( showPointNums || i==0 || (i == Curve.points.size() -1 && !built) || selected ) {
    			g.fillOval(x,y,RADIUS,RADIUS);
    			String num = ""+ point.getNum();
    			if ( showPointNums) {
	    			g.setColor(Color.BLACK);
	        		graphics.drawString( num , x, y); 
        		}
    		}
    		else if ( !complete || allowMovement ) {
    			g.setColor(Color.BLACK);
    			g.fillOval(x,y,RADIUS-1,RADIUS-1);
    		}
    		
    	}
		
		for (int i=0; i< Curve.intPoints.size(); i++){
			Point cur = Curve.intPoints.get(i); // get the intersection point
			g.setPaint( (cur.getSign() ? new Color(65, 215, 30) : RED ));
			
			
			if ( close(anchorPoints, cur) )  {
				g.setPaint( PURPLE);
			}
			
   		  	int x = (int)cur.getX()-(RADIUS/2);
    		int y = (int)cur.getY()-(RADIUS/2);
    		g.fillOval(x,y,RADIUS,RADIUS);
    		if ( built ) {
	    		String sign = null;
	    		if ( cur.getSign() ) { sign = "+"; } else if (!cur.getSign()) { sign = "_";} 
	    		g.setPaint(Color.BLACK);
	    		Font F = graphics.getFont(); // store current font   
	    		g.setFont(new Font("TimesRoman", Font.BOLD, 17)) ; // change font 
	    													// for int point signs 
	    		if ( cur.getSign() ) {
	    			//g.setPaint(Color.);
	    			graphics.drawString( sign, (int)x, (int)(y+RADIUS)); 
	    		}
	    		else {
	    			g.setPaint(new Color(240, 240, 240));
	    			graphics.drawString( sign, (int)x+1, (int)(y+RADIUS-7)); 
	    			graphics.drawString( sign, (int)x+1, (int)(y+RADIUS-6));
	    		}
	    		
	    		g.setFont(F); // return to the old font
    		}
    		
    		
    		g.setPaint(Color.black);
    		String num = (showPointNums) ? new Integer( cur.getIntPointIndex() ).toString()+"i" :
    			 "" + (cur.getIntPointIndex());
    		
    		graphics.drawString( num, ( (showPointNums)? x+7 : x ), y);
    		if ( showPointNums) {
	    		g.setPaint(Color.BLACK);
	    		int space = ( cur.getNum() > 9) ? 10 : 0;
	    		graphics.drawString( new Integer( cur.getNum() ).toString(), x-6-space, y);
	    		graphics.drawString( "-", x-2, y);
	    		graphics.drawString( "-", x, y);
    		}
    		
    	}

		for (int m=0; m<Curve.faces.size(); m++ ) {
			Face f = Curve.faces.get(m);
			
			System.out.println("Face " + m + " -- " + f.toString() + " with area " + f.getArea());
    		g.setColor(Color.BLACK);
    		
    		Point rep = f.getRep();
    		int k = 4;
    		int x = (int)rep.getX()-(RADIUS/2)-k;
    		int y = (int)rep.getY()-(RADIUS/2)-k;
    		// BRIGHT = 0.299 R + 0.587 G + 0.114 B formula for perceived brightness of RGB color
    		Color C = f.getColor();
    		if ( moving) {
    			C = new Color(0, 250, 200);
    		}
    		double brightness = C.getRed()*0.299 + C.getGreen()*0.587+C.getBlue()*0.144;
    		String s = "" +  f.getWindingNum();
			int n = s.length();
			if ( f.getWindingNum() == 0) {
				n += -1;
			}
			int nonNeg;
			if ( f.getWindingNum() > 0) {
				nonNeg = 2;
			}
			else if (f.getWindingNum() == 0 ) {
				nonNeg = 2;
			}
			else { nonNeg = 0; }
			
			boolean dark = brightness < 135;
			if ( colorFaces ) {
				//ensure there is room around the rep inside the face 
				if ( f.getPolygon().contains(rep.getX()-8, rep.getY()-8, 16, 16) ){
					if ( selectedFace != null && selectedFace.equals(f) ) {
						g.setColor(Color.GRAY);
						g.fillOval(x, y, RADIUS+2*k, RADIUS+2*k);
		    			g.setColor(Color.WHITE);
		    			graphics.drawString("" +f.getWindingNum(), x+k-n-1+nonNeg, y+8+k+(2-n));
		    			g.setColor(Color.WHITE);
					}
					else {
						if ( highlightFaces == f.getType() ) {
							g.setColor(Color.BLACK);
							g.fillOval(x, y, RADIUS+2*k, RADIUS+2*k);
			    			g.setColor(Color.WHITE);
			    			graphics.drawString("" +f.getWindingNum(), x+k-n-1+nonNeg, y+8+k+(2-n));
			    			g.setColor(Color.BLACK);
						}
						
						else if ( dark)  {
							//g.setColor(new Color(255, 255, 255) );
			    			g.setColor(new Color( Math.min(C.getRed()+40, 255),
			    					Math.min(255,C.getGreen()+40),Math.min(255,C.getBlue()+40)) );	
			    			g.fillOval(x, y, RADIUS+2*k, RADIUS+2*k);
			    			//g.fillOval(x,y,RADIUS,RADIUS); // draw rep
			    			g.setColor(Color.LIGHT_GRAY);
			    			graphics.drawString("" +f.getWindingNum(), x+k-n-1+nonNeg, y+8+k+(2-n));
			    			g.setColor(Color.LIGHT_GRAY);
			    			//g.drawOval(x, y, RADIUS+2*k+2, RADIUS+2*k+2);
			    			g.setColor(Color.WHITE);
			    		}
			    		else {
			    			//g.setColor( new Color(0, 0, 0));
			    			g.setColor(new Color(Math.max(C.getRed()-40,0), Math.max(C.getGreen()-40, 0), Math.max(C.getBlue()-40, 0)));
			    			g.fillOval(x, y, RADIUS+2*k, RADIUS+2*k);
			    			g.setColor(Color.WHITE);
			    			graphics.drawString("" +f.getWindingNum(), x+k-n-1+nonNeg, y+8+k+(2-n));
			    			g.setColor(Color.LIGHT_GRAY);
			    			//g.drawOval(x, y, RADIUS+2*k+2, RADIUS+2*k+2);
			    			g.setColor(Color.BLACK);
			    		}
					}
					graphics.drawString("F"+new Integer(m).toString() ,x,y);
				}
				else {
					/** if face is too small, don't draw anything inside
					if ( highlightFaces == f.getType() ) {
						g.setColor(Color.BLACK);
						if ( dark ) { g.setColor(Color.WHITE); }
						g.fillOval(x, y, RADIUS-1, RADIUS-1);
						g.setColor(RED);
						g.drawString(new Integer (m).toString(), x, y);
					}
					else {
						g.setColor(Color.BLACK);
						if ( dark ) { g.setColor(Color.WHITE); }
						g.fillOval(x, y, RADIUS-1, RADIUS-1);
						g.setColor(RED);
						g.drawString(new Integer (m).toString(), x, y);
						
					} **/
				}
			}
			else {
				g.setColor(Color.BLACK);
				// don't show info on whether faces are good/bad 
				/**
				String bad = f.isGood() ? "" : "B:"; // write whether face is good
				int b = f.isGood() ? -2 : 0; **/
				graphics.drawString("" + f.getWindingNum(), x+k-n-1+nonNeg, y+8+k+(2-n));
			}
		} // end of faces loop
		
		String builtYet = "";
		
		if ( built ) { 
			if (complete) { builtYet = "completed!";	}
			else { builtYet = "built but not completed!";  }
		}  
		else { builtYet = "Not built yet!"; }
		if (built) {
			overlapping = isSelfOverlapping(Curve);
			Color C = ( overlapping ) ? new Color(0, 200, 0) : new Color(220, 0 ,0 );
			g.setColor( C );
			String overlap = "";
			if ( overlapping) { overlap = "Curve is self-overlapping!"; }
			else if ( !overlapping) { overlap = "Curve is not self-overlapping!"; }
			else { overlap = "not determined if curve is self-overlapping"; }
			graphics.drawString(overlap , 20, 560);
		}
		g.setColor(Color.black);
		graphics.drawString(builtYet, 20,580);
		graphics.drawString("Area swept: " + areaSwept, 20,600);
		graphics.drawString("Current winding area: " + windingArea, 20, 640);
		if ( startingWindingArea != -1 ) {
			graphics.drawString("Starting winding area: " + startingWindingArea, 20, 660);
		}
		graphics.setColor(Color.blue);
		if (Curve.intPoints.size() <= 17) {
			graphics.drawString("Intersection Sequence: " + intSeq.toString(), 20, 620);
		}
		graphics.setColor(Color.black);
		
		if ( minimumHomotopyArea != -1) {
			minimumHomotopyArea = round(minimumHomotopyArea);
			graphics.setColor(new Color(140, 0, 200));
			graphics.drawString("Minimum Homotopy Area: " + minimumHomotopyArea, 413, 640);
			graphics.setColor(Color.BLACK);
			graphics.drawString("Splitting Sequence", 460 +anchorPoints.size(), 660);
		
			if (anchorPoints != null) {
				for (int j=0; j< anchorPoints.size(); j++) {
					Point P = anchorPoints.get(j  );
					g.setPaint( P.color); 
	    			String s = P.toString();
	    			String Sindex = ( !s.contains("i")) ? s : s.substring(0, s.length()-1);
					
					graphics.drawString(Sindex, 527 - 16*anchorPoints.size() + 32*j, 680);
					g.setPaint( Color.black);
					if ( j != anchorPoints.size()-1) 
						graphics.drawString("->", 527 - 16*anchorPoints.size() + 32*j +15, 680);
					
					if ( !P.equals(new Point(0,0) ) ) {
						if ( close(Curve.intPoints, P ) ) {
							for (Point Q: Curve.intPoints) {  
								if ( Q.distance(P) < anchorMovement ) { 
									P.setLocation( Q.getX(), Q.getY() );// update anchor location
									P.setIntPointIndex( Q.getIntPointIndex() );
								} 
							}  
							g.setPaint( Color.BLACK);
						}
						else {
							g.setPaint(new Color(125, 125, 125));
						}
					}
				}
			}
			else {
				System.out.println("Null anchors");
			}
		}
		
		// no need to print average max splits info and other superfluous details
		/** 
		BigInteger f = factorial(Curve.intPoints.size());
		g.setPaint(new Color(200, 0, 250));
		if ( splits != -1) {
			int size = Curve.intPoints.size();
			if ( size < 10 ) {
				graphics.drawString("TNS: " + addCommas(splits) + ". Worst case: " + 
				addCommas(f) , 20, 680);
			}
			else {
				graphics.drawString("TNS: " + addCommas(splits), 20, 680);
			}
		}
		
		g.setPaint(new Color(250, 0, 200));
		
		if (avgMaxSplits != -1) {
			int s = Curve.intPoints.size();
			graphics.drawString("Average max splits:  " + avgMaxSplits + " /" + (s-1), 400, 680);
		} 
		
		g.setPaint(new Color(0,0, 150));
		 
		if ( baseCases != 0 ) {
			graphics.drawString("Times split more than once: " + extraSplits + ". Base case calls: " +
								baseCases, 300, 640);
		}
		if ( extraSplits != -1 ) {
			//g.setPaint( new Color(120, 0, 0));
		} 
		if ( avgSplits != -1 ) {
			g.setPaint( new Color(200, 120, 0));
			graphics.drawString("Average complexity after a split: " + avgSplits + "/ " + 
			Math.ceil(Curve.intPoints.size()/3), 600, 680);
		} **/
		frame.east.repaint();
    }
    
    public boolean close(ArrayList<Point> A, Point P) {
    	for (Point Q: A) {
    		if ( Q.distance(P) < anchorMovement)  {
    			return true;
    		}
    	}
    	return false; 
    }
    public String addCommas(BigInteger B) {
    	String s = B.toString();
    	return addCommas(s);
    }
    
    public String addCommas(int a) {
    	String s = "" + a;
    	return addCommas(s);
    }
    
    public String addCommas(String s) {
    	String newString = "";
    	int count = 1;
    	for (int i= s.length()-1; i>=0; i--) {
    		newString = s.charAt(i) + newString;
    		if ( count > 0 && count % 3 ==0 && i > 0 ) {
    			newString = "," + newString;
    		}
    		count++;
    	}
    	return newString;
    }
    
    public BigInteger factorial(int a) {
    	if ( a < 0) {
    		return BigInteger.valueOf(-1);
    	}
    	BigInteger answer = BigInteger.valueOf(1);
    	for (int i=2; i <= a; i++) {
    		answer = answer.multiply(BigInteger.valueOf(i));
    	}
    	return answer;
    }
    
    public static double round(double value, int places) {
        if (places < 0) throw new IllegalArgumentException();

        BigDecimal bd = new BigDecimal(value);
        bd = bd.setScale(places, RoundingMode.HALF_UP);
        return bd.doubleValue();
    }
    
    public double round(double a) {
    	double m = a; 
		m = m * 1000;
		m = Math.round(m);
		m = m/1000;
		return m;
    }
    public void distanceCheck() { 
    	ArrayList<Point> badPoints = new ArrayList<Point>();
    	for (int i=0; i< allPoints.size(); i++ ) {
    		int j = (i+1) % allPoints.size();
    		Point P = allPoints.get(i); Point Q = allPoints.get(j);
    		if ( (!P.isIntersection() || !Q.isIntersection()) && P.distance(Q) < THRESHOLD) { 
    			Point remove = ( !P.isIntersection() ) ? P: Q; // point to be removed 
    			// either one, whichever is not an int point (if both aren't int points, take P) 
    			badPoints.add( remove );
    		}
    	}
    	for (Point bad: badPoints){ 
    		removePoint(bad);
    	}
    }
    
    public int minDistanceIndex(Point p) { // return min distance from p to any other point
    	double minDistance = Integer.MAX_VALUE-2;
    	int closestIndex = 0;
    	if ( Curve.points.size() == 0) {
    		return -1; // shouldn't happen
    	}
    	for (Point point: Curve.points) {
    		double curDistance = point.distance(p);
    		if ( !point.equals(p) && curDistance < minDistance) {
    			minDistance = curDistance;
    			closestIndex = Curve.points.indexOf(point);
    		}
    	}
    	return closestIndex; 
    }
    
    public double minDistance(Point p) { // return min distance from p to any other non-intersection point
    	
    	if (Curve.points.size() == 0) {
    		return THRESHOLD+1;
    	}
    	double minimumDistance = Integer.MAX_VALUE-2;
    	for (Point point: Curve.points) {
    		double curDistance = point.distance(p);
    		if ( !point.equals(p) && curDistance < minimumDistance) {
    			minimumDistance = curDistance;
    		}
    	}
    	return minimumDistance;
    }
  
    public void unUse() {
    	for (Edge e: shortEdges) {
    		e.unUse();
    	}
    } 
    
    public boolean colinear(Edge e1, Edge e2) {
    	Point P = e1.getP1(); Point Q = e1.getP2(); Point R = e2.getP2();
    	ArrayList<Point> candidates = new ArrayList<Point>();
    	candidates.add(P); candidates.add(Q); candidates.add(R); // add all 3 points to arraylist 
    	for (Edge E: Curve.edges) { 
    		ArrayList<Point> pointsOnE = new ArrayList<Point>();
    		
    		pointsOnE.add( E.getP1() ); pointsOnE.add( E.getP2() ); // add endpoints 
    		for ( Point u: E.getIntPoints() ) {  pointsOnE.add(u); } // add all intersection points 
    		if ( pointsOnE.containsAll(candidates) ) {
    			return true;
    		}
    	}
    	return false; // false if no edge contained all the points
    }
    
    
  
    public void checkShortEdges() {
    	for ( int k = 0; k< allPoints.size(); k++) {
    		Point p = allPoints.get(k);
    		System.out.println("Point: " + k  + " has edges:");
    		for ( Edge e: p.getShortEdges() ) {
    			System.out.print(e.toString() + " " );
    		}
    		System.out.println("");
    	}
    }  
    
    public void remove(ArrayList<Edge> A, Edge E ) {
    	Edge bad = null;
    	for ( Edge e : A ) {
    		if ( e.equals(E) ) {
    			bad = e;
    			break;
    		}
    	}
    	if ( bad != null) {
    		A.remove(bad);
    	}
    }
    public void removeIntPoint(Point p) { // delete intersection point 
    	for (Edge E: p.getShortEdges() ) {
    		Point P;
    		if ( E.getP1().equals(p) ) {
    			P = E.getP2();
    		}
    		else {
    			P = E.getP1();
    		}
    		remove(P.getShortEdges(), E);
    	}
    	for (Edge e: p.getEdges()) {
    		e.removeIntPoint(p);
    	}
    	Curve.removeIntPoint(p);
    	repaint();
    }
    
    public void removeIntPoints(Point p) {//remove all intersection points of edges connected to p
    	ArrayList<Integer> badIndices = new ArrayList<Integer>(); // all indices of edges to be deleted
		for (int i=0; i< Curve.edges.size(); i++) { 
			// check if either point on the edge is the point to be removed
			Edge curEdge = Curve.edges.get(i);
			if ( curEdge.p1 == null) { 
				System.out.println("error: edge has null 1st point!!");
			}
			if ( p == null ) {
				System.out.println("ERROR: Point input is null!");
			}
			if (curEdge.p1.equals(p) ) {
				badIndices.add( new Integer(i));
			}
			else if (curEdge.p2.equals(p) ) {
				 badIndices.add( new Integer(i));
			}
		} 
		// reverse order of indices, so deleting elements) doesn't affect indices
		Collections.sort(badIndices, Collections.reverseOrder());
		
		for (int j:badIndices) { // delete all edges connected to the newly deleted point
			Edge e = Curve.edges.get(j);
		    
		    int r = e.getIntPoints().size();
		    for (int k=r-1; k>=0; k--) {
				removeIntPoint(e.getIntPoint(k)); // delete intersection points on the edge b4 deleting edge
		    }
		}
		repaint();
    }
    
    // remove all edges and intersection points connected to p
    public void removePoint(Point p) {   	
    	ArrayList<Integer> badIndices = new ArrayList<Integer>(); // all indices of edges to be deleted
		for (int i=0; i< Curve.edges.size(); i++) { 
			// check if either point on the edge is the point to be removed
			Edge curEdge = Curve.edges.get(i);
			if (curEdge.p1.equals(p) ) {
				badIndices.add( new Integer(i));
				Point v = curEdge.p2; //  vertex connected to removed vertex
				v.getEdges().remove(curEdge); // remove edge from other vertex's edge list
			}
			else if (curEdge.p2.equals(p) ) {
				 badIndices.add( new Integer(i));
				 Point v = curEdge.p1; //  vertex connected to removed vertex
			//	 System.out.println("Point " + v.toString() + " had " + v.getEdges().size() + " edges" );
				 v.getEdges().remove(curEdge); // remove edge from other vertex's edge list
			//	 System.out.println(" Now it has : " + v.getEdges().size());
			}
		} 
		
		// reverse order of indices, so deleting elements doesn't affect indices
		Collections.sort(badIndices, Collections.reverseOrder());
		
		for (int j:badIndices) { // delete all edges connected to the newly deleted point
			Edge e = Curve.edges.get(j);
		    
		    int r = e.getIntPoints().size();
		    for (int k=r-1; k>=0; k--) {
				removeIntPoint(e.getIntPoint(k)); // delete intersection points on the edge b4 deleting edge
		    }
		 
		    Curve.edges.remove(j); // remove the edge
		}
		Curve.points.remove(p);
		repaint();
    }
    
    public Edge edgeAdd(Point p1, Point p2) {
    	// add edge to list, to arrayList of edges for each point
    	Edge e = new Edge(p1, p2);
    	Curve.edges.add(e);
        p1.addEdge(e);
    	p2.addEdge(e);
    	findIntersections(e);
    	return e;
    }
    
    public void edgeRemove(Point P1, Point P2) { // remove the edge connecting P1 and P2
    	for (Edge edge: P1.getEdges() ) {
    		// check if current edge has the other point
			if ( edge.getP1().equals(P2) || edge.getP2().equals(P2) ) {
				P1.getEdges().remove(edge); // if so remove the edge from P1's edge list
				P2.getEdges().remove(edge); // also remove edge from P2's edge list
				int r = edge.getIntPoints().size();
				for (int i=r-1; i>=0; i--) {
					removeIntPoint( edge.getIntPoint(i)); 
				}
				Curve.edges.remove(edge);// also remove edge from global edge list
				break;
			}
		}
    	repaint();
    }
    
    /**
   public boolean checkIntersections(Edge e) {
	   for (int r=0; r<Curve.edges.size(); r++) {
			Edge current = Curve.edges.get(r);
			if ( e.getP1().equals(current.getP1()) || e.getP1().equals(current.getP2() ) ) {
				return false;
				// do nothing
			}
			else if ( e.getP2().equals( current.getP1() ) || e.getP2().equals( current.getP1() ) ) {
				return false;
				// do nothing 
			}
			else if (e.doesIntersect( current ) ) {
					// get intersection point
					Point intersectionPoint = e.getIntersection( current );
					intersectionPoint.setIntersection();
					Point p = intersectionPoint.closest(Curve.intPoints);	
					if ( intersectionPoint.distance(p) > THRESHOLD/2 ) {
						 // create a new intersection point
						if (intersectionPoint.getX() == 1000 && intersectionPoint.getY() == 1000) {
							System.out.println("ERROR: Edges of false intersection: " + e.toString() + " and " + current.toString());
							return false; //NOT A REAL INTERSECTION
						}
					}
					else {
						return true;
					}
			}
		}
   		return false;
   } **/
   
    public boolean pointShared(Edge E, Edge e1) {
    	return  E.getP1() == e1.getP1() || E.getP1() == e1.getP2() ||
	    		   E.getP2() == e1.getP1() || E.getP2() == e1.getP1();
    	
    }
    
   public void findIntersections(Edge e) { // find and add all int points associated to e that are currently not found
	   for (int r = 0; r < this.Curve.edges.size(); r++) {
	       Edge current = (Edge)this.Curve.edges.get(r);
	       // ensure edges don't share points
	       if ( !( e.getP1() == current.getP1() ) && !(e.getP1() == current.getP2() ) &&
	    		   ( !(e.getP2() == current.getP1() ) && !( e.getP2() == current.getP1() ) ) ) {  
	       
		       if ( e.doesIntersect(current)) { // check if edges e and current intersect
		    	   Point intersectionPoint = e.getIntersection(current);
		           intersectionPoint.setIntersection(); // the point is an intersection point
		           if ((intersectionPoint.getX() == 1000.0) && (intersectionPoint.getY() == 1000.0)) {
		               System.out.println("ERROR: Edges of false intersection: " + e.toString() + " and " + current.toString());
		               return;
		            }
		           
		            Point p = intersectionPoint.closestEqAllowed(Curve.intPoints);
		            
		            if (intersectionPoint.distance(p) > 0 )  {
		               
		               current.addIntPoint(intersectionPoint);
		               e.addIntPoint(intersectionPoint); // update int point arrays of both edges 
		               
		               Curve.intPoints.add(intersectionPoint); // add it to global array of int points
		               intersectionPoint.addEdge(e);
		               intersectionPoint.addEdge(current); // update edges of the intersection point
		             }
		             else { // degenerate case of multiple shared intersection points!!
		            	 
		            	 return;
		             }
		       	}
	       }   
	   }
   }
   
   	
    public Edge getEdge(Point u, Point v) { // get the edge connecting points u and v
    	for (Edge e: u.getEdges() ) {
    		if ( e.getP1().equals(v) || e.getP2().equals(v) ) {
    			return e;
    		}
    	} 
    	return null; // return null if the edge does not exist 
    } 

    public void clear() { // clear everything!! - start over!
    	Curve.points.clear();
    	allPoints.clear();
    	Curve.edges.clear();
    	shortEdges.clear();
    	Curve.faces.clear();
    	intSeq = "";
    	Curve.intPoints.clear();
    	faceReps.clear();
    	built = false;
    	complete = false;
    	areaSwept = 0; 
    	windingArea = 0;
    	splits = -1;
    	startingWindingArea = 0;
    	bestArea = Integer.MAX_VALUE;
    	titusSeq.clear();
    	bestTitus.clear();
    	Curve.idealSplittingPoints = 0;
    	for (Point P: Curve.curve) 
    		P.unSelect();
    	pointsSelected.clear();
    	curPoint = null;
    	selectedFace = null;
    	avgSplits = -1;
    	avgMaxSplits = -1;
    	directedEdges.clear();
    	highlightFaces = -1;
    	intSeqIntegers.clear();
    	Curve = new Curve(); // reset the curve 
    	minimumHomotopyArea = -1;
    	anchorPoints.clear();
    	go = false;
    	selfOverlappingCurves.clear();
    	baseCases = 0;
    	extraSplits = -1;
    	repaint();
    }
    
 // determine if e1 and e2 intersect at all ( even if on the endpoints ) 
    public static boolean intersects(Edge e1, Edge e2) {
    	return Line2D.linesIntersect(e1.getP1().getX(),e1.getP1().getY(),e1.getP2().getX(),
    			e1.getP2().getY(), e2.getP1().getX(), e2.getP1().getY(), e2.getP2().getX(), e2.getP2().getY());
    }
    
	public static boolean insideTriangle(Point p, Point a, Point b, Point c){

	    double ap_x = p.getX() - a.getX();
	    double ap_y = p.getY() - a.getY();

	    boolean p_ab = (b.getX()-a.getX())*ap_y-(b.getY()-a.getY())*ap_x > 0;

	    if((c.getX()-a.getX())*ap_y-(c.getY()-a.getY())*ap_x > 0 == p_ab) return false;

	    if((c.getX()-b.getX())*(p.getY()-b.getY())-(c.getY()-b.getY())*(p.getX()-b.getX()) > 0 != p_ab) return false;

	    return true;
	}

    
 // determine if edge intersects interior of triangle T
    public static boolean triIntersect(Edge e, Point p1, Point p2, Point p3){
    try {
		/*

    	if ( T.npoints != 3) {
    		 //throw new Exception("Polygon is not a triangle. Cannot perform method triIntersects");
    		}
    	
    	// get all points from the triangle 
    	Point p1 = new Point( T.xpoints[0], T.ypoints[0]);
    	Point p2 = new Point( T.xpoints[1], T.ypoints[1]);
    	Point p3 = new Point( T.xpoints[2], T.ypoints[2]);
    	*/
    	// form edges, which are the boundary of the triangle T 
    
    	Edge e1 = new Edge(p1, p2);
    	Edge e2 = new Edge(p1, p3);
    	Edge e3 = new Edge(p2, p3);
    
    	int vertexIntersections = 0;
    	int intersectionCount = 0; // count # of boundary edges of T the edge e intersects
    	if ( intersects(e, e1) ) {
    		//if (! e.doesIntersect(e1)) { // check if one of the edge's endpoints shares with the triangle's edge 
    		// Check if one of the edge's endpoints are too close to the endpoints of a triangle edge
    		if(e.getP1().distance(p1)<6 || e.getP2().distance(p1)<6 || e.getP1().distance(p2)<6 || e.getP2().distance(p2)<6){ 
    			vertexIntersections++;
    		}
    		intersectionCount++;
    	}
    
    	if ( intersects(e, e2) ) { 
    		//if (! e.doesIntersect(e2)) {
        	// Check if one of the edge's endpoints are too close to the endpoints of a triangle edge
        	if(e.getP1().distance(p1)<6 || e.getP2().distance(p1)<6 || e.getP1().distance(p3)<6 || e.getP2().distance(p3)<6){ 
    			vertexIntersections++;
    		}
    		intersectionCount++;
    	}
    
    	if ( intersects(e, e3) ) { 
    		//if (! e.doesIntersect(e3)) {
           	// Check if one of the edge's endpoints are too close to the endpoints of a triangle edge
           	if(e.getP1().distance(p2)<6 || e.getP2().distance(p2)<6 || e.getP1().distance(p3)<6 || e.getP2().distance(p3)<6){ 
    			vertexIntersections++;
    		}
    		intersectionCount++;
    	}
    	
    	// check if one or both of the edge's endpoints are inside the triangle 
    	//boolean contains1 = T.contains(e.getP1());
    	//boolean contains2 = T.contains(e.getP2());
    	boolean contains1 = insideTriangle(e.getP1(), p1, p2, p3);
    	boolean contains2 = insideTriangle(e.getP2(), p1, p2, p3);
    	if(contains1 || contains2){
    	//if ( T.contains(e.getP1()) || T.contains(e.getP2()) ) {
    	
    		return true;
    	}
    	
    	else if (vertexIntersections == 2 ) {
    		
        	// if neither point is inside, and one endpoint of the edge is in a triangle endpoint 
        	// 
    		if(intersectionCount==3)
    			return true;

    		
        	// if neither point is inside, and there are two fake intersections 
        	// at the triangle's endpoints, then it is a false case
    		else			
    			return false; 
    	}
    	
    	// otherwise the two intersections are legitimate and it does hit the interior 
    	else if ( intersectionCount == 2 ) { 
    		return true;
    	}
 

    	else { // no other intersection cases to consider
    		return false;
    	} 
    }
    	catch (Exception E) {
    		System.err.print(E);
    		return false;
    	}
    }
    
    public static boolean isSelfOverlapping(Curve C) {
    	ArrayList<Point> allPoints = C.getCurve();
    	
		ArrayList<Point> goodPoints = new ArrayList<Point>();
		for (int a = 0; a < allPoints.size(); a++) {
			Point current = allPoints.get(a);
			if( ! current.isIntersection()){
				goodPoints.add(current);
			}
		}
			int[][] Q = new int[goodPoints.size()][goodPoints.size()];
			for (int l = 0; l < goodPoints.size(); l++) {
				for (int i = 0; i < goodPoints.size(); i++) {
					int j;
					j = ((l + i) % goodPoints.size());
					if (l ==0) {
						Q[i][j] = 1;
					} else if (l == 1) {
						Q[i][j] = 1;
					} else if (l == 2) {
						
						
						double detForConvexity;
						detForConvexity = (goodPoints.get((i+2)%goodPoints.size()).getX()*goodPoints.get((i+1)%goodPoints.size()).getY()) 
								- (goodPoints.get((i+2)%goodPoints.size()).getY()*goodPoints.get((i+1)%goodPoints.size()).getX()) 
								- (goodPoints.get(i).getX()*goodPoints.get((i+1)%goodPoints.size()).getY()) 
								+ (goodPoints.get(i).getY()*goodPoints.get((i+1)%goodPoints.size()).getX())
								+ (goodPoints.get(i).getX()*goodPoints.get((i+2)%goodPoints.size()).getY())
								- (goodPoints.get(i).getY()*goodPoints.get((i+2)%goodPoints.size()).getX());

						if (detForConvexity > 0) {
							Q[i][j] = 1;
						}
						
						else {
							Q[i][j] = 0;
						}
					} else{ // l>=3
						for (int k = (i + 1)%goodPoints.size(); k != j; k = (k+1)%goodPoints.size()) {
							if (Q[i][k] == 1 && Q[k][j] == 1) {
								/*first requirement met */
								double det;

								det = (goodPoints.get(j).getX()*goodPoints.get(k).getY()) 
										- (goodPoints.get(j).getY()*goodPoints.get(k).getX()) 
										- (goodPoints.get(i).getX()*goodPoints.get(k).getY()) 
										+ (goodPoints.get(i).getY()*goodPoints.get(k).getX())
										+ (goodPoints.get(i).getX()*goodPoints.get(j).getY())
										- (goodPoints.get(i).getY()*goodPoints.get(j).getX());

								if (det > 0) {
									ArrayList<Point> relVertices = new ArrayList<Point>(); //make an arrayList of points now relative to k as origin
									
									if (k == 0){ //deals with index exceptions when k is 0
										Point ri = new Point(goodPoints.get(i).getX() - goodPoints.get(k).getX(),goodPoints.get(i).getY()  - goodPoints.get(k).getY());
										Point rj = new Point(goodPoints.get(j).getX() - goodPoints.get(k).getX(),goodPoints.get(j).getY()  - goodPoints.get(k).getY());
										Point rkplus = new Point(goodPoints.get((k+1)%goodPoints.size()).getX() - goodPoints.get(k).getX(),goodPoints.get((k+1)%goodPoints.size()).getY()  - goodPoints.get(k).getY());
										Point rkminus = new Point(goodPoints.get(goodPoints.size()-1).getX() - goodPoints.get(k).getX(),goodPoints.get(goodPoints.size()-1).getY()  - goodPoints.get(k).getY());
										
										relVertices.add(ri);
										relVertices.add(rj);
										relVertices.add(rkplus);
										relVertices.add(rkminus);
									}
									else{
										Point ri = new Point(goodPoints.get(i).getX() - goodPoints.get(k).getX(),goodPoints.get(i).getY()  - goodPoints.get(k).getY());
										Point rj = new Point(goodPoints.get(j).getX() - goodPoints.get(k).getX(),goodPoints.get(j).getY()  - goodPoints.get(k).getY());
										Point rkplus = new Point(goodPoints.get((k+1)%goodPoints.size()).getX() - goodPoints.get(k).getX(),goodPoints.get((k+1)%goodPoints.size()).getY()  - goodPoints.get(k).getY());
										Point rkminus = new Point(goodPoints.get((k-1)%goodPoints.size()).getX() - goodPoints.get(k).getX(),goodPoints.get((k-1)%goodPoints.size()).getY()  - goodPoints.get(k).getY());
									
									
										relVertices.add(ri);
										relVertices.add(rj);
										relVertices.add(rkplus);
										relVertices.add(rkminus);
									}
									
									double [] angles = new double[4];//place to store angles from "x axis" from k point
									for (int c = 0; c < 4; c++) {
										
										//now need to deal with 4 quadrant cases to get unique angle 
										if(relVertices.get(c).getX() >= 0 && relVertices.get(c).getY() >= 0){//1st quadrant
											angles[c] = Math.toDegrees(Math.atan(relVertices.get(c).getY()/relVertices.get(c).getX()));
										}
										else if(relVertices.get(c).getX() <= 0 && relVertices.get(c).getY() >= 0){//2nd quadrant
											angles[c] = Math.toDegrees(Math.PI - Math.atan(relVertices.get(c).getY()/(-1*relVertices.get(c).getX())));
										}
										
										else if(relVertices.get(c).getX() <= 0 && relVertices.get(c).getY() <= 0){//3rd quadrant
											angles[c] = Math.toDegrees(Math.PI + Math.atan((-1*relVertices.get(c).getY())/(-1*relVertices.get(c).getX())));
										}
										
										else{//4th quadrant
											angles[c] = Math.toDegrees(2*Math.PI - Math.atan((-1*relVertices.get(c).getY())/relVertices.get(c).getX()));
										}
									}

									int indexOfmin = 0;
									double min = angles[0];
									
									for (int d = 0; d < 4; d++) {
										if (angles[d] < min) {
											min = angles[d];
											indexOfmin = d;
										}
									}
									
									/*third condition is that they appear in i, j, k+1, k-1 order CW in drawing, CCW in java flipped
									 * if third statement is met: */
									if (i == (k-1)%goodPoints.size() || (k ==  0 && i == goodPoints.size()-1)){
										if (angles[indexOfmin] <= angles[(indexOfmin+1)%4] && 
												angles[(indexOfmin+1)%4]<= angles[(indexOfmin+2)%4]){
											
											/*Polygon Triangle = new Polygon();

											Triangle.xpoints[0] = (int)goodPoints.get(i).getX();
											Triangle.xpoints[1] = (int)goodPoints.get(j).getX();
											Triangle.xpoints[2] = (int)goodPoints.get(k).getX();

											Triangle.ypoints[0] = (int)goodPoints.get(i).getY();
											Triangle.ypoints[1] = (int)goodPoints.get(j).getY();
											Triangle.ypoints[2] = (int)goodPoints.get(k).getY();*/
											
											Point tpoint0 = goodPoints.get(i);
											Point tpoint1 = goodPoints.get(j);
											Point tpoint2 = goodPoints.get(k);
											
											Edge r1 = new Edge(goodPoints.get(i), goodPoints.get((i+1)%goodPoints.size()));
											Edge r2;
											Edge r3 = new Edge(goodPoints.get(k), goodPoints.get((k+1)%goodPoints.size()));
											Edge r4;
											
											if (k == 0) {
												r2 = new Edge(goodPoints.get(goodPoints.size()-1), goodPoints.get(k));
											}
											
											else{
												r2 = new Edge(goodPoints.get((k-1)%goodPoints.size()), goodPoints.get(k));
											}
											
											if (j == 0) {
												r4 = new Edge(goodPoints.get(goodPoints.size()-1), goodPoints.get(j));
											}
											else{
												r4 = new Edge(goodPoints.get((j-1)%goodPoints.size()), goodPoints.get(j));
											}	

											Edge[] otherEdges = new Edge[4];
											otherEdges[0] = r1;
											otherEdges[1] = r2;
											otherEdges[2] = r3;
											otherEdges[3] = r4;


											/*
											boolean intersect0 = triIntersect(otherEdges[0], Triangle);
											boolean intersect1 = triIntersect(otherEdges[1], Triangle) ;
											boolean intersect2 = triIntersect(otherEdges[2], Triangle);
											boolean intersect3 = triIntersect(otherEdges[3], Triangle);*/
											boolean intersect0 = triIntersect(otherEdges[0], tpoint0, tpoint1, tpoint2);
											boolean intersect1 = triIntersect(otherEdges[1], tpoint0, tpoint1, tpoint2);
											boolean intersect2 = triIntersect(otherEdges[2], tpoint0, tpoint1, tpoint2);
											boolean intersect3 = triIntersect(otherEdges[3], tpoint0, tpoint1, tpoint2);
											
											if(intersect0 == false && intersect1 == false && intersect2 == false && intersect3 == false){

											//if (triIntersect(otherEdges[0], Triangle) == false && triIntersect(otherEdges[1], Triangle) == false 
											//		&& triIntersect(otherEdges[2], Triangle) == false && triIntersect(otherEdges[3], Triangle) == false) {
												Q[i][j] = 1;
											}
										}
									}
								
									else if (angles[indexOfmin] <= angles[(indexOfmin+1)%4] && 
											angles[(indexOfmin+1)%4]<= angles[(indexOfmin+2)%4]
													&&  angles[(indexOfmin+2)%4]<=angles[(indexOfmin+3)%4]) {
										
										Polygon Triangle = new Polygon();

										/*iangle.xpoints[0] = (int)goodPoints.get(i).getX();
										Triangle.xpoints[1] = (int)goodPoints.get(j).getX();
										Triangle.xpoints[2] = (int)goodPoints.get(k).getX();

										Triangle.ypoints[0] = (int)goodPoints.get(i).getY();
										Triangle.ypoints[1] = (int)goodPoints.get(j).getY();
										Triangle.ypoints[2] = (int)goodPoints.get(k).getY();*/
										Point tpoint0 = goodPoints.get(i);
										Point tpoint1 = goodPoints.get(j);
										Point tpoint2 = goodPoints.get(k);

										Edge r1 = new Edge(goodPoints.get(i), goodPoints.get((i+1)%goodPoints.size()));
										Edge r2;
										Edge r3 = new Edge(goodPoints.get(k), goodPoints.get((k+1)%goodPoints.size()));
										Edge r4;
										
										if (k == 0) {
											r2 = new Edge(goodPoints.get(goodPoints.size()-1), goodPoints.get(k));
										}
										
										else{
											r2 = new Edge(goodPoints.get((k-1)%goodPoints.size()), goodPoints.get(k));
										}
										
										if (j == 0) {
											r4 = new Edge(goodPoints.get(goodPoints.size()-1), goodPoints.get(j));
										}
										else{
											r4 = new Edge(goodPoints.get((j-1)%goodPoints.size()), goodPoints.get(j));
										}	
										
										Edge[] otherEdges = new Edge[4];
										otherEdges[0] = r1;
										otherEdges[1] = r2;
										otherEdges[2] = r3;
										otherEdges[3] = r4;

										/*
										boolean intersect0 = triIntersect(otherEdges[0], Triangle);
										boolean intersect1 = triIntersect(otherEdges[1], Triangle) ;
										boolean intersect2 = triIntersect(otherEdges[2], Triangle);
										boolean intersect3 = triIntersect(otherEdges[3], Triangle);*/
										boolean intersect0 = triIntersect(otherEdges[0], tpoint0, tpoint1, tpoint2);
										boolean intersect1 = triIntersect(otherEdges[1], tpoint0, tpoint1, tpoint2);
										boolean intersect2 = triIntersect(otherEdges[2], tpoint0, tpoint1, tpoint2);
										boolean intersect3 = triIntersect(otherEdges[3], tpoint0, tpoint1, tpoint2);
										
										
										if(intersect0 == false && intersect1 == false && intersect2 == false && intersect3 == false){

										//if (triIntersect(otherEdges[0], Triangle) == false && triIntersect(otherEdges[1], Triangle) == false 
										//		&& triIntersect(otherEdges[2], Triangle) == false && triIntersect(otherEdges[3], Triangle) == false) {
											Q[i][j] = 1;
										}
									}

								} // det<=0
								
							}
						}
					} // l>=3
				} // for i
			} // for l
			
			// no need to print anymore
			
			/** 
			System.out.println("Printing Q for " + C.toString());
			for (int a = 0; a < goodPoints.size(); a++) {
				for (int b = 0; b < goodPoints.size(); b++) {
					//System.out.println("here");
					if (b == goodPoints.size()-1) {
						System.out.println(Q[a][b]);
					}
					else {
					System.out.print(Q[a][b]);
					}
				}
			} 
			System.out.flush(); **/
		
			
			if(Q[0][goodPoints.size()-1] == 1)
				return true;
			for (int z = 1; z < goodPoints.size(); z++) {
				//System.out.println(Q[z][z-1]);
				if (Q[z][(z-1) % goodPoints.size()] == 1) {
					//System.out.println("True");
					return true;
				}
			}
			//System.out.println("False");
			return false;
	    }
	
    
    	public void classifyFaces() {
    	
	    	ArrayList<Face> classifieds = new ArrayList<Face>(); 
	    	
	    	for (Point P: Curve.curve) { // find all faces with depth 0 
	    		Face F = Curve.faces.get(P.getFaces().get(0).intValue());
	    		if ( P.faceMembership() == 1 && !contains(classifieds,F) ) {
	    			classifieds.add( F);
	    			F.setDepth(1);
	    		}
	    	}
	    	
	    	int i =0;
	    	ArrayList<Face> remaining = new ArrayList<Face>();
	    	remaining.addAll(Curve.faces);
	    	remaining.removeAll( classifieds);
	    	
	    	while (i < 1000 && remaining.size() > 0 ) { // don't allow loop to run forever (go till faces are classified)
	    		
	    		ArrayList<Face> remove = new ArrayList<Face>(); // all faces classified in the latest run through
	    		for (Face newF: remaining) {
	    			boolean done = false;
	    			for ( Edge E: newF.getEdges()) {
	    				if (!done) {
	    					for (Face oldF: classifieds) {
	    						if ( contains(oldF.getEdges(), E)) {
	        	    				done = true; // if so, compute winding # for the new face f
	        	    				remove.add(newF);
	        	    				newF.setDepth( oldF.getDepth() + 1 );
	        	    				break;
	        	    			}
	        				}
	    				}
	    			}
	    		}
	    		
	    		for (Face F: remove) {
	    			remaining.remove(F); // remove face from unclassifieds
	    			classifieds.add(F); // add to classifieds
	    		}
	    		i++;
	    	}
	    	
	    	for (Face F: Curve.faces) { 
	    		if ( F.getDepth() == Math.abs(F.getWindingNum()) ) {
	    			F.setGood(true);
	    		}
	    		else {
	    			F.setGood(false);
	    		}
	    	}
    }
    	
    public void numberPoints(boolean calcFaces) {
    	
    	ArrayList<Point> points = Curve.points;
    	
    	ArrayList<Point> curve = Curve.getCurve();
    	
    	if ( !built) {
    		allPoints.clear();
    		for (int i=0; i< points.size(); i++) {
    			points.get(i).setNum(i);
    			points.get(i).name("" + i);
    			allPoints.add( points.get(i) );
    		}
    		return;
    	}
    	
    	boolean findFaces = true;
    	for (Point p : allPoints) { 
    		if ( p.getEdges().size() >= 3) {
    			System.out.println("Point " + p.toString() + " is connected to " + p.getEdges().size() + " edges");
    			for (Edge E: p.getEdges() ) {
    				System.out.print(E.toString() + " ");
    			}
    			System.out.println("");
    			findFaces = false;
    		}
    		p.getFaces().clear();
    		p.removeConnections();
    		p.setNum(-1); // ensure all points are renumbered
    		p.getShortEdges().clear();
    	}
    	curve.clear();
    	directedEdges.clear();
    	Curve.faces.clear();
    	faceReps.clear();
    	shortEdges.clear(); // recompute short edges, faces, and ordering of points
    	allPoints.clear();
    	intSeq = "0,"; // recompute the intersection sequence 
    	intersectionSeq.clear();
    	intSeqIntegers.clear();
    	int intPointCount = 1;
    	
    	if (points.size() == 0) { return; } // do nothing
    	
    	if ( Curve.intPoints.size() == 0 ) { //  NO intersection points yet, so we know the intersection sequence
    		for (int j=0; j< points.size() ; j++) {
    			Point cur = points.get(j);
    			Point next = points.get( (j+1) % points.size() );
    			points.get(j).setNum(j); // points have predictable numbers 
    			points.get(j).name("" + j);
    			Edge e = new Edge(cur, next);
    			
    			Edge E = getEdge(cur, next);
    			if ( E.getP1()!= cur) { // check orientation of the edge
    				E.reverse();
    			}
 
    			Edge f = new Edge(next, cur);
    			directedEdges.add(e);
    			cur.addShortEdge(e); // add correct short edges to both points' arraylists of short edges
    			next.addShortEdge(f);
    			shortEdges.add(e); // add edge to short edge 
    			shortEdges.add(f);
    			allPoints.add(cur); curve.add(cur);// order all points in allPoints array
    			//intSeq += j + "."; // update intersection sequence
    			intersectionSeq.add(new Integer(j));
    		} 
    	}
    	else {   // there are MAYBE intersection points on the current edge
    		
    		int count = 0; // the number of points encountered so far 
    		boolean done = false;
        	for (int i=0; i< points.size() && !done; i++) {
        		Point cur = points.get(i);
        		Point next = points.get((i+1)%points.size());
        		
        		
        		Edge e = getEdge(cur, next);
        		if ( e == null)  {
        			try {
        				throw new Exception("Could not find edge between adjacent points.");
        			}
        			catch (Exception E) {
        				System.err.println(E);
        			}
        		}
        		if ( e.getP1()!= cur) { // check orientation of the edge 
        			e.reverse();  // if edge is backwards, flip it
        		}
        		// if there are no intersection points on the current edge
        		if ( e.getIntPoints().size() == 0 ) { 
        			if ( cur.getNum() == -1 ) {// current point is NOT numbered
        				cur.setNum(count); 
        				cur.name("" + count);
            			count++; // number the point and update the count 
        			}
        			Edge g = new Edge(cur, next); // create directed edges between points 
        			Edge h = new Edge(next, cur);
        			cur.addShortEdge(g);
        			next.addShortEdge(h);
        			directedEdges.add(g); // add correct directed edge 
        			shortEdges.add(g); // add to list of all edges
        			shortEdges.add(h);  
        			//intSeq += cur.getNum() + "."; // add the # of cur to the int seq.
        			intersectionSeq.add(new Integer(cur.getNum()) );
        			allPoints.add(cur); curve.add(cur);
        			if ( !next.equals(points.get(0)) && next.getNum() == -1 ) {
        				next.setNum(count); // if we aren't at the start again, next has no #, then # it
        				next.name("" + count);
        				count++; //update count
        			}
        			
        		}
        		else { //THERE ARE INTERSECTION POINTS!
        			
        			ArrayList<Point> A = new ArrayList<Point>();
            		for (Point u: e.getIntPoints() ) { // all AVAIALABLE points to use, (don't re-use)
            			A.add(u);
            		}
            		A.add(e.p1); A.add(e.p2); // add endpoints of the edge
            		while (true) {        	
            			if (cur.getNum() == -1) { // if point is currently unreached
            				cur.setNum(count); // number the point, update count
            				cur.name("" + count);
            				count++;
            			}
            			//intSeq += cur.getNum() + ".";  // update intersection sequence 
            			intersectionSeq.add(new Integer(cur.getNum())); 
            			if ( !contains(allPoints, cur) ) {
            				if ( cur.isIntersection() ) { 
            					cur.name("" + intPointCount + "i");
            					cur.setIntPointIndex(intPointCount);
            					intPointCount++;
            				}
            				allPoints.add(cur); // keep track of all points 
            				curve.add(cur);
            			}		
            			else {
            				curve.add(cur); // always add the current point to "curve" array
            			}
            			
            			if (cur.isIntersection()) {
            				int index = cur.getIntPointIndex();
            				//intSeq += "" + ( index < 10 ? index : "(" + index + ")" );
            				intSeqIntegers.add(new Integer(index));
            				intSeq += index + ",";
            			}
            			A.remove(cur); // don't allow reuses 
            			Point upNext = cur.closest(A);
            			
            			Edge newEdge1 = new Edge(cur, upNext);
            			Edge newEdge2 = new Edge(upNext, cur);
            			// if either point is an intersection point, then we have an interior edge 
            			
            			directedEdges.add(newEdge1); // add correct directed edge from curve 
            			shortEdges.add(newEdge1); 
        				shortEdges.add(newEdge2);
            			cur.addShortEdge(newEdge1); // add directed edges for cur and upNext 
            			upNext.addShortEdge(newEdge2);
            			cur = upNext;
            			if (cur.isIntersection() ) {
            				if ( cur.getVisits() == 0 || cur.getVisits() == 1) {
            					cur.addVector(new Edge(cur, next));
            					cur.setIndex(count-1); // remove one b/c count was already incremented
            				}
                			cur.visit();
                			if ( cur.getVisits() == 2 ) {
                				cur.calcSign();
                			}
                		}
            			if ( cur.equals( points.get(0) ) ) { // finished traversal of curve
            				done = true;
            				break;
            			}
            			else if ( cur.equals(next) ) { // the whole edge [p(i)->p(i+1)] has been traversed 
            				cur.setNum(count);
            				cur.name("" + count);
            				count++;
            				break; // exit the while loop 
            			}
            		}
        		}
        	}
    	}
    	intSeq += "0"; // return to the starting point
    	intersectionSeq.add(new Integer(0));
    	
    	if ( points.size()>=3 && findFaces && calcFaces) { // CALCULATE FACES!!
    		faceReps.clear();
    			
	    	for ( Edge E: shortEdges) {  // generate faces from interior edges 
	    		Face f = Curve.getFace(E);
	    	
		    	if ( f != null &&  !contains(Curve.faces, f) ) {  // ensure face is nontrivial and new
	    			Curve.faces.add(f);
	    			f.getNewRep();
	    			faceReps.add(f.getRep()); // add face rep
		    	}
	    	}
	    }
    	
		if ( Curve.faces.size() > 1 ) { 
			if ( Curve.faces.size() != Curve.intPoints.size()+2 ) { // faces were not properly found
				Curve.faces.clear();
				Curve.faceReps.clear();
				return;
			}
			double maxArea = 0;
			Face badFace = null;
			for (Face f: Curve.faces) { // delete degenerate face of exterior of the graph
				if ( f.getArea() > maxArea) { // it is the face with maximal area
					maxArea = f.getArea();
					badFace = f;
				}
			}
			//System.out.println("Bad face was: " + badFace.toString());
			faceReps.remove(badFace.getRep());
			Curve.faces.remove(badFace);
		}
		
		windingArea = 0;
		
		for (Point P: allPoints) {
			P.getFaces().clear();
    		P.removeConnections();
		} 
		
		int count = 0;
		ArrayList<Face> facesCopy = new ArrayList<Face>();
		int size = Curve.faces.size();
		ArrayList<Color> colors = new ArrayList<Color>();
		for (int j=0; j<Curve.faces.size(); j++) {
			Face F = Curve.faces.get(j);
			Color c = new Color( 20 + (int)(110/size*j), 40 + (int) (210/size * (j+1) ), (int)(250 - 180/(size) *(size-j)) );
			F.setColor(c);
			colors.add(c);
			//F.setColor(c); 
			facesCopy.add(F); // add to the copy
			F.setNum(count); // number the face
			count++;
			F.incrementPoints(); // increase face count for each point on the face
			if ( F.isCW() ) {
				F.reverse();
			}
		}
		
		if ( calcFaces)  {
			windingArea = 0;
			for ( Face f: Curve.faces ){
				try {
					if ( f.getPolygon().contains(f.getRep()) ){
						double wNum = Curve.windingNum( f.getRep() );
						int wNumInt = (int)wNum;
						if ( Math.round(wNum) !=  wNumInt) {
							try { throw new Exception("winding number not an integer");}
							catch (Exception EEE) { System.err.println(EEE); }
						}
						f.setWindingNum( wNumInt );
					}
					else {
						throw new Exception("Face rep not inside the face for " + f.toString());
					}
				}
				catch (  Exception EE ) { System.err.println(EE); } 
				windingArea += Math.abs(f.getWindingNum()) * f.getArea();
			}
			windingArea = round(windingArea);
		}
		if ( Curve.faces.size()  > 0) {
			
			classifyFaces(); // classify faces as good or bad 
			
			for (Face F: Curve.faces ) {
				double w = F.getWindingNum();
				int k = 250-(int)(250*Math.abs(w)/10);
				if ( w > 0 ) {
					F.setColor( new Color( Math.max(0, k-25), k, 255));
				}
				else if (w < 0){
					F.setColor( new Color(Math.max(0, k-25), 255, k));
				}
				else {
					F.setColor(new Color(255, 255, 255));
				}
			}
			
		}
		
    	ArrayList<Point> reversePoints = new ArrayList<Point>();
    	for (int j = allPoints.size()-1; j >= 0; j--) {
    		reversePoints.add( allPoints.get(j) );
    	}
    	
		overlapping = isSelfOverlapping(Curve) || isSelfOverlapping(Curve.reverse());
		/** 
		 * Counter C = new Counter();
		avgSplits = C.avgSplits(intSeqIntegers);
		avgSplits = round(avgSplits);
		avgMaxSplits = C.avgMaxSplits(intSeqIntegers);
		avgMaxSplits = round(avgMaxSplits);	
		if ( Curve.intPoints.size() < 9) {
			splits = C.TNS( C.convert(intSeqIntegers) );		
		}
		else {
			splits = -1;
		} **/
		
    }
    
    
    public Edge curveEdge( Edge e, ArrayList<Edge> edges) { // return the edge e, with the orientation of the curve 
   		for (Edge E: edges){ 
   			if ( e.equals(E) ) {
   				return E; 
   			}
   		}
   		try {
   			throw new newException("Input edge to curveEdge is not an edge from the input edges.");
   		}
   		catch (newException EE) {
   			System.err.println(EE);
   			return null; 
   		}
   	}
    
    public Edge edgeGrab(Point P, Point Q, ArrayList<Edge> edges) {  //return edge from edges array that has both points
    	if ( !P.isIntersection() && !Q.isIntersection() ) {
    		return curveEdge(new Edge(P, Q), edges);
    	}
    	if ( !Q.isIntersection() ) {
    		Point R = P;
    		P = Q; // swap points so Q is an intersection point
    		Q = R;
    	}
    	// either way, we now have Q == int
    	if ( P.getEdge(0).getIntPoints().contains(Q) )  {
    		return curveEdge(P.getEdge(0), edges);
		}
		else if ( P.getEdge(1)!= null && P.getEdge(1).getIntPoints().contains(Q) ) {
			return curveEdge(P.getEdge(1), edges);
		}
		else {
			try {
				throw new newException("Input points to edgeGrab not on the same edge!!");
			}
			catch (newException EE){ 
				System.err.println(EE);
				return null;
			}
		}
    }
    // take an edge from the face and determine whether the face is +1 or -1 windingNum
    // as traversed on the curve ( the face is input in CCW order) 
    public int winding(Face F, Edge e, ArrayList<Edge> edges) {
    	Edge E = edgeGrab(e.getP1(), e.getP2(), edges); // get the FULL edge that is a part of ( maybe just e)
    	E = curveEdge(E, edges); // determine the correct orientation of the edge E
    	if ( E == null) {
    		try {
    			throw new newException("Could not find edge when computing winding #!");
    		}
    		catch (newException EE) {
    			System.err.println(EE);
    			return Integer.MAX_VALUE;
    		}
    	}
    	// to see the actual order of traversal
    	if ( e.getP1().getX() == e.getP2().getX())  { // check for verticality
    		boolean up = E.getP2().getY() < E.getP1().getY();
    		if ( upIsInside(F.getPolygon(), e) ) {
    			if ( up ) { // edge is crossed in CCW orientation
    				return +1;
    			}
    			else { // edge is crossed in CW orientation
    				return -1; 
    			}
    		}
    		else {
    			if (up) { return -1; }
    			else { return +1; }
    		}
    	}
    	boolean left = E.getP2().getX() <  E.getP1().getX();
    	if ( upIsInside(F.getPolygon(), e) ) {
    		if ( left ) { return -1; }
    		else { return +1; }
    	}
    	else {
    		if ( left) { return +1; }
    		else { return -1; }
    	}
    }
 
    
    public void sortEdges() {
    	for (Point u: Curve.points) { // sort edges of all non-intersection points
    		u.sortShortEdges();
    	}
    	for (Point v: Curve.intPoints) { // sort edges of all intersection points
    		v.sortShortEdges();
    	}
    }
    
    public void addPointBtwn(Point p) { // add point between two existing points
    	Point p1 = pointsSelected.get(0);
    	Point p2 = pointsSelected.get(1);
    	if ( p1.getNum() > p2.getNum() ) { // swap points so #(p1) < #(p2) 
    		Point r = p1;
    		p1 = p2;
    		p2 = r;
    	}
    	if ( Curve.points.indexOf(p2) == Curve.points.size()-1 && Curve.points.indexOf(p1) == 0) {
    		Curve.points.add( p);
    		//Curve.curve.add(p);
    	}
    	else {
    		Curve.points.add(Curve.points.indexOf(p2), p); // insert point in correct place in the points array
    		//Curve.curve.add(e)
    	}
    	edgeRemove( p1, p2 ) ;
    	Edge e1 = edgeAdd(p1, p);
    	Edge e2 = edgeAdd( p, p2);
    	findIntersections(e1);
    	findIntersections(e2);
		p1.unSelect(); p2.unSelect(); // de-select both points now that the new point was added
		pointsSelected.clear();
		p.made();
		pointNew = true;
    }
    
    public void addPoint(Point newPoint) {

    	if ( minDistance(newPoint) < THRESHOLD && Curve.points.size()!= 0) {
    		// check if new point is close to an existing point
			System.out.println("Point attempted to be added at "+ newPoint.closest(Curve.points).getNum() + " but was too close ");
			curPoint = null;
			return;
		}
    	else if ( newPoint.distance(newPoint.closest(Curve.intPoints)) < THRESHOLD ) {
    		// don't add a new point very close to an intersection point
    		System.out.println("Point attempted to be added at "+ newPoint.closest(Curve.intPoints).getNum() + " but was too close ");
			curPoint = null;
			return;
    	}
		else {
			newPoint.made();
			Curve.curve.add( newPoint) ;
			Curve.points.add( newPoint );
			if (Curve.points.size() >= 2) {
				Point prev = Curve.points.get(Curve.points.size()-2);
				Edge e1 = edgeAdd( prev, newPoint ); 
				if (Curve.points.size() > 2) {  // detect intersections between new edge and old edges
					findIntersections(e1);
				}
			}
		}
    }
    
    public class curveDoneException extends Exception {
        public curveDoneException(String message) {
            super(message);
        }
    }
    
    // move point and determine according new intersection points 
    public void movePoint(Point p, double x, double y) {    	
    	
    	if ( selfOverlappingCurves.size() > 0) {
    		selfOverlappingCurves.clear();
    		anchorPoints.clear();
    		anchorIndices.clear();
    		minimumHomotopyArea = -1;
    	}
    	try {
    		if ( !complete || allowMovement) {
		    	removeIntPoints(p); // get rid of old intersection points
		    	p.setLocation(x, y);
				for (Edge e: p.getEdges()) {
					findIntersections(e);
				}
				repaint();
		    	}
    		else {
    			throw new curveDoneException("Allow free movement to move points.");
    		}
    	}	
    	catch (curveDoneException e) {
    		System.err.println(e);
    	}
	}
    
    public double rand() { // random number between -1 and 1
    	int sign;
		if ( Math.random() < 0.5) { sign = 1; } else { sign = -1; }
		return Math.random() * sign;
    }
    
    public Point randomMove(Point p) {
    	double x = p.getStartX();
		double y = p.getStartY();
		
		x += rand()*10; y += rand()*10; // perturb x and y a bit 
		return new Point(x, y);
    }
    
    public void perturb(Point p) {// perturb until not too close to existing points 
    	
    	if ( p.closestVal(allPoints) < THRESHOLD ) { // check if too close to an existing point
			Point Q = randomMove(p);
			
			for (int i=0; i<=100; i++)  { // perturb point until not too close to an existing point
				
				if ( Q.closestVal(allPoints) > THRESHOLD ) { // if new test point not too close
					movePoint(p, Q.getX(), Q.getY()); // update location
					return;
				}
				Q = randomMove(Q); // move again 
			}
		}
    	System.out.println("ERROR: 100 PERTURBATIONS WERE UNSUCCESSFUL!!");
    }
    
    public boolean contains(ArrayList<Edge> A, Edge e) {  // the arraylist.contains(Edge) method doesn't work
    	for (Edge E: A) {
    		if (E.equals(e)) {
    			return true;
    		}
    	}
    	return false;
    }
    
    public boolean contains(ArrayList<Face> A, Face f) {  // the arraylist.contains(Edge) method doesn't work
    	for (Face F: A) {
    		if (F.equals(f)) {
    			return true;
    		}
    	}
    	return false;
    }
    
    public boolean contains(ArrayList<Point> A, Point p) { // arraylist.contains(Point) method doesn't work
    	for (Point P: A) {
    		if (P.equals(p)) {
    			return true;
    		}
    	}
    	return false;
    }
	
    public int countIntersections(Curve C, Edge E) {
    	int total = 0;
    	for (Edge e: C.edges) {
    		if ( E.doesIntersect(e) ) {
    			total++;
    		}
    	}
    	return total;
    }
    
    
    public void continuousDrawProcess() {
    	Point first = Curve.curve.get(0);
    	Point last = Curve.get(Curve.points.size() -1 );
    	built = true;
    	smoothDraw = false;
    	Edge e = edgeAdd(first, last);
    	findIntersections(e);
    	int check = 0;
    	int startIndex  = -1;
		Point curr = null; // needed to check when first pair of int points was found
    	int prevIndex = -1;
    	int curIndex = -1;
    	ArrayList<Point> curve = Curve.getCurve(); 
    	
    	ArrayList<Point> badPoints = new ArrayList<Point>();
    	numberPoints(false);
    	repaint();
    	
    	for (int i=0; i< Curve.curve.size(); i++) {
    		Point P = Curve.get(i);
    		if ( Curve.curve.get(i).isIntersection()) {
    			if ( curr == null) {
    				curr = P;
    				curIndex = i;
    				startIndex = i;
    				for (int k = 1; k < curIndex; k += 2 ){ // between the int points
						// take every other point and remove it 
    					badPoints.add( Curve.curve.get(k) ); 
    				}
    			}
    			else {
    				prevIndex = curIndex; // move index to
    				curr = P;
    				curIndex = i;
    				for (int k = prevIndex+1; k < curIndex; k += 2 ){ // between the int points
    												// take every other point and remove it 
    					badPoints.add( Curve.curve.get(k) ); 
    				}
    			}
    		}
    	}
    	// final deletion process at the end for the edge connecting n -> 0 
    	if ( curIndex != -1) { 
			for (int k = curIndex+1; k < Curve.curve.size(); k++ ){ // between the int points
				// take every other point and remove it 
				if ( (k - curIndex) % 2 == 0) {
					badPoints.add( Curve.curve.get(k) ); 
				}
			}
		}
    	else  { 
			for (int k = 0; k < Curve.curve.size(); k++ ){ // between the int points
				// take every other point and remove it 
				if ( k % 2 == 0) {
					badPoints.add( Curve.curve.get(k) ); 
				}
			}
		}
    	check = 1;
    	for (Point Q: badPoints) { // delete the inbetween-er points
    		Curve.deletePoint(Q);
    	}
    	
    	check = 2;
    	for (Edge E: Curve.edges) { // ensure all intersection points are found 
    		findIntersections(E);
    	}
    	
    	check = 3;
    	/**
    	for (int i =0; i <  Curve.edges.size(); i++) { // find all int points that weren't found 
    		for (int j=i+1; j < Curve.edges.size(); j++ ) {
    			Edge E1 = Curve.edges.get(i);
    			Edge E2 = Curve.edges.get(j);
    			if ( E1.getP1() != E2.getP1() && E1.getP1() != E2.getP2() && E1.getP2() != E2.getP1() && E1.getP2() != E2.getP2() ) {
	    			Point P1 = E2.getP1(); Point P2 = E2.getP2();
	    			Point Q = ( E1.distance(P1) < E1.distance(P2) ) ? P1 : P2; // get point closer to the first edge
	    			boolean intersects = E1.intersects(E2);
	    			double epsilon = 2;
	    			if ( !intersects ) { // confirm that we do NOT have an intersection
	    				Point S = E2.getP1();
	    				Edge Eup = new Edge(S, new Point(Q.getX(), Q.getY()-epsilon) );
	    				Edge Edown = new Edge(S, new Point(Q.getX(), Q.getY()+epsilon) );
	    				Edge Eleft = new Edge(S, new Point(Q.getX()-epsilon, Q.getY()) );
	    				Edge Eright = new Edge(S, new Point(Q.getX()+epsilon, Q.getY()) );
	    				Point candidate = null;
	    				if ( E1.intersects(Eup) ) 
	    					candidate = E1.getIntersection(Eup);
	    				if ( E1.intersects(Edown) ) 
	    					candidate = E1.getIntersection(Edown);
	    				if ( E1.intersects(Eleft) ) 
	    					candidate = E1.getIntersection(Eleft);
	    				if ( E1.intersects(Eright) ) 
	    					candidate = E1.getIntersection(Eright);
	    				if ( Curve.intPoints.size() > 0 && candidate != null && candidate.closestVal(Curve.intPoints) > 3 ) {
	    					E1.addIntPoint(candidate);
	    					E2.addIntPoint(candidate);
	    					Curve.intPoints.add(candidate);
	    					candidate.addEdge(E1);
	    					candidate.addEdge(E2);
	    				}
	    						
	    			}
    			}
    		}
    	}
    	**/
    	
 
    	for ( int u=0; u < 100; u++) {
	    	ArrayList<Point> awkward = new ArrayList<Point>(); // awkward co-linear points 
	    	
	    	for (int j= 0; j< Curve.points.size(); j++) {
	    		Point Q = Curve.points.get(j);
	    		for (Edge E: Curve.edges) {
	    			boolean contains = E.getP1().equals(Q) || E.getP2().equals(Q);
	    			if (! contains) {
	    				if ( E.distance(Q) < 6 ) {
	    					awkward.add( Q) ; 
	    				}
	    			}
	    		}
	    	}
	    	
	    	if ( awkward.size() == 0 ) // terminate when no awkward points are found
	    		break;
	    	
	    	//PROCESS weird new ones
	    	for (Point Q: awkward) {
	    		int j = Curve.points.indexOf(Q);
	    		int k = (j+1)%(Curve.points.size());
	    		Point R = Curve.points.get(k);
	    		movePoint(Q, 0.5*(Q.getX() + R.getX()), 0.5*(Q.getY() + R.getY()) );	
	    		check = 4;
	    	}
	    	
	    	for (Edge E: Curve.edges)  { // vertical edge prevention
	    		Point P1 = E.getP1();
	    		Point P2 = E.getP2();
	    		if ( P1.getX() == P2.getX() ) { 
	    			P1.setX( P1.getX() - 1 );
	    			P2.setX( P2.getX() + 1);
	    		}
	    	}
    	}
    
    	numberPoints(false);
    	ArrayList<Point> fakeIntPoints = new ArrayList<Point>();
    	for (Point I: Curve.intPoints) {
    		if ( I.getIntPointIndex() == -1) {
    			fakeIntPoints.add(I);
    		}
    	}
    	for (Point I: fakeIntPoints)
    		Curve.removeIntPoint(I);
    	numberPoints(true);
    	repaint();
    }
    
    public void closeCurve() { // finish the curve when the user clicks back on the 1st point
    	Point first = Curve.curve.get(0);
    	Point last = Curve.get(Curve.points.size() -1 );
    	built = true;
    	smoothDraw = false;
    	//Edge e = new Edge(first, last);
    	Edge e = edgeAdd(first, last);
    	findIntersections(e);
    	numberPoints(true);
    	repaint();
    }
    
    public class notBuiltException extends Exception {
        public notBuiltException(String message) {
            super(message);
        }
    }
    
    public void done() {
    	try {
	    	if ( built ) {
		    	this.complete = true;
		    	repaint();
		    	for (Point P: Curve.curve) 
		    		P.unSelect();
		    	pointsSelected.clear();
	    	}
	    	else {
	    		throw new notBuiltException("Curve not fully built yet!");
	    	}
    	}
    	catch(notBuiltException e) {
    		System.err.print(e);
    	}
    }
    
    // check
    public boolean containsAlternate(ArrayList<Point> A, Point P ) {
    	for (Point Q: A) {
    		if ((int) Q.getX() == (int)P.getX() && (int)Q.getY() == (int)P.getY() ) {
    			return true;
    		}
    	}
    	return false;
    }
    
    public boolean equalsApprox(Point P, Point Q) {
    	return P.distance(Q) < 1.5;
    }
    
    public void contractLoop(Curve C) {
    	/**
    	Point anchorPoint = null;
    	for (Point anchor: anchorPoints) {
    		if ( contains(C.curve, anchor) ) {
    			anchorPoint = anchor; // save the anchor 
    			break; 
    		}
    	} **/
    	Point anchorPoint = C.anchor;
    	if ( anchorPoint == null) {
    		try { throw new Exception("Anchor not found. Self-overlapping loop could not be contracted."); }
    		catch (Exception E) { System.err.println(E); }
    	}
    	ArrayList<Point> badIntPoints = new ArrayList<Point>();
    	ArrayList<Point> badPoints = new ArrayList<Point>();
    	
    	for (Point intPoint: Curve.intPoints) { // for every intpoint, see if it has a match on the curve, if so, delete it (excluding the basepoint)
    		for (Point P: C.curve) {
	    		if ( intPoint.equals(P) && !intPoint.equals(anchorPoint) ) {
	    			badPoints.add( P);
	    			badIntPoints.add( intPoint );
	    		} 
    		}
    	} 
    	for (Point bad: badIntPoints) { 
    		removeIntPoint(bad); // remove all intpoints globally
    	}
    	for (Point bad: badPoints) { 
			C.removePoint(bad); // remove all intpoints locally as well
    	}
    	ArrayList<Point> realPoints = new ArrayList<Point>();
    	ArrayList<Edge> realEdges = new ArrayList<Edge>();
    	
    	for (Point P: C.curve) {
    		boolean found = false;
    		for (Point Q: Curve.curve) { // get the originals of every point
    			if ( Q.equals(P)  ) {
    				realPoints.add(Q);
    				found = true;
    				break;
    			}
    		}
    		if (! found) { 
    			try {
    				throw new Exception("Match not found.");
    			}
    			catch (Exception E) {
    				System.err.println(E);
    			}
    		}
    	}
    	for (int i=0; i< realPoints.size(); i++) {
    		Point cur = realPoints.get(i);
    		Point next = realPoints.get( (i+1) % realPoints.size() );
    		Edge e = new Edge(cur, next); // 
    		realEdges.add(e);
    	}
    	selfOverlappingCurves.remove(C); // delete it from the collection of SO curves 
    	// the curve is now a "monogon" 
    	Face newMonogon = new Face(realEdges); 
    	TitusI(newMonogon);
    	return; 
    }
    
    public void TitusI(Face F) {
    	selectedFace = null;
    	highlightFaces = -1;
    	titusSeq.add("Ia on Face " + Curve.faces.indexOf(F) );
    	try {
    		if ( F.getType() != 1) {
    			throw new newException("Cannot apply Ia move. Face is not a monogon.");
        	}
    	}
    	catch ( newException EE) {
    		System.err.println(EE);
    		return;
    	}
    	
    	boolean zeroRemoved = false; //remember if the 0th point was deleted or not
    	Point intPoint = null; // the intersection point from the face 
    	 // indices of the 
    	ArrayList<Integer> indices = new ArrayList<Integer>();
    	for (int i=F.getSize()-1; i>=0 ; i--) { // look through p
    		Point P = F.getPoint(i);
    		if ( !P.isIntersection() ) { // save indices of non-int points 
    			indices.add( new Integer( Curve.points.indexOf(P) ) );
    		}
    		else {
    			intPoint = P;
    		}
    	}
    	try {
	    	if ( intPoint == null) {
	    		throw new newException("NULL INT POINT DURING TITUS Ia MOVE");
	    	}
    	}
    	catch ( newException E) {
    		System.err.println(E);
    	}
    	
    	Collections.sort(indices); // sort indices of all non-int points
    	Collections.reverse(indices); // then reverse so they are in decreasing order 
    	// points connected to the one int point of the face
    	ArrayList<Point> connections = new ArrayList<Point>();
    	// edges of the intersection point
    	ArrayList<Edge> edgeC = new ArrayList<Edge>();
    	for ( Edge e: intPoint.getShortEdges() ) {
    		connections.add(e.getP2());
    		edgeC.add(e);
    	}
    	for (Integer k: indices) {
    		int j = k.intValue();
    		if ( j==0) { zeroRemoved = true; } 
    		
    		if (connections.contains(Curve.points.get(j))) {
    			int n = connections.indexOf(Curve.points.get(j));
    			Edge E = edgeC.get(n);
    			intPoint.removeShortEdge( E );
    			connections.remove(n);
    			edgeC.remove(n);
    		}
    		
    		removePoint(Curve.points.get(j)); // remove the point 
    	}
    	
    	// local edges may be connected to int points that are deleted!
    	/** 
    	ArrayList<Edge> localEdges = intPoint.getShortEdges(); **/ 
    	ArrayList<Edge> globalEdges = intPoint.getEdges();
    	if (globalEdges.size()!= 2) {
    		try {
    			throw new Exception("Error! Titus move not working. Only "+globalEdges.size() +" edges");
    		}
    		catch(Exception E) {
    			System.err.println(E);
        		return;    		
    		}
    	}
    	else {
    		// get the points connected to the int point, make new edges,  
    		// 					add the int point to the points array 
    		intPoint.notIntersection();
    		Edge e1 = intPoint.getEdge(0);
    		Edge e2 = intPoint.getEdge(1);
    		Point P1;
    		Point P2;
    		// get the points that need to be connected to the int point of F
    		if ( Curve.points.contains(e1.getP1() ) ) { P1 =  e1.getP1(); } else { P1 = e1.getP2(); }
    		if ( Curve.points.contains(e2.getP1() ) ) { P2 =  e2.getP1(); } else { P2 = e2.getP2(); }
    		int j;
    		// get the larger index of the two non-int points P1,P2 
    		// if we have intpoint -> 0, then treat 0 as max value ( 0 ≈ points.size() (modular) )
    		if ( intPoint.getVector(1).getP2().equals(Curve.points.get(0)) ) {
    			j = 0;
    		}
    		else {
    			j = Math.max( Curve.points.indexOf(P1), Curve.points.indexOf(P2) );
    		}
    		// reset the edges of the int point
    		intPoint.getEdges().clear();
    		intPoint.getShortEdges().clear();
    		
    		// insert the int point between both points, then
    		// add edges both ways
    		if ( zeroRemoved ) { Curve.points.add(0, intPoint); }
    		else { Curve.points.add(j, intPoint); }
    		edgeAdd(intPoint, P1);
    		edgeAdd(intPoint, P2);
    	}
    	boolean g1 = faceReps.remove(F.getRep());
    	boolean g2 = Curve.faces.remove(F);
    	if ( !g1 || !g2 ) { // ensure the face F and its rep were removed 
    		try {
	    		if ( ! g1 ) {  throw new Exception("Face rep not removed during Titus Ia move!"); }
	    		else if ( ! g2 ){ throw new Exception("Face F not removed during Titus Ia move!"); }
	    		else {
	    			throw new Exception("Neither face nor face rep removed during Titus Ia move!");
	    		}
    		}
    		catch (Exception E) {
    			System.err.println(E);
    		}
    	}
    	areaSwept += F.getArea();  // keep track of the area of the face 
    	areaSwept = round(areaSwept);
    //	selfOverlappingCurves.clear();
    //	anchorPoints.clear();
    //	findCurves(root, Curve);
    	numberPoints(true);
    }
    
    /**
    public void findCurves(Node root, Curve curve) { 
    	if (root == null) {
    		selfOverlappingCurves.add( curve ) ;
    		return;
    	}
    	if ( root.left != null) {
    		findCurves( root.left, splitCurve(root.getPoint(), true, curve) );
    	}
    	else { 
    		Curve newCurve = splitCurve(root.getPoint(), true, curve);
    		findCurves( null, newCurve ); 
    	}
    	if ( root.right != null) {
    		findCurves( root.right, splitCurve(root.getPoint(), false, curve)  );
    	}
    	else { findCurves( null, splitCurve( root.getPoint() , false, curve)  ); }
    	
    }  **/
    
    public class newException extends Exception {
        public newException(String message) {
            super(message);
        }
    }
    
    // turn made from edge E to edge e
    public double turn(Edge E, Edge e) { // simple math to calculate if left turn or right turn
		/**
		 * 		|1 p_x p_y |
 		 *  det |1 q_x q_y |
		 * 		|1 r_x r_y |
		 */
		
		Point p = E.getP1();
		Point q = E.getP2();
		Point r = e.getP2();
		if ( E.getP2().equals(e.getP1())) {
			double val = ( q.getX()-p.getX() )*(r.getY()-p.getY()) - (q.getY()-p.getY())*(r.getX()-p.getX());
		if ( Math.abs(val) < 10) { // margin of error in computation, most turns are between 500-5000
				return 0;
			} 
			else if ( val > 0) {
				return 1;
			}
			else {
				return -1;
			}
		}
		return Integer.MAX_VALUE;
	}
    
    /**
    public double slope(Edge E ) {
    	if ( E.getP1().getX() != E.getP2().getX() ) {
    		return  ( E.getP2().getY() - E.getP1().getY() ) / ( E.getP2().getX() - E.getP1().getX() );
    	}
    	else {
    		return Integer.MAX_VALUE;
    	}
    }
    
    public boolean doesIntersect( Edge E, Edge e1) {
    	if ( pointShared(E, e1) || Math.abs(slope(E) - slope(e1)) < 0.001 )
    		return false;
    	Point a = E.getP1();
    	Point b = E.getP2();
    	Point c = e1.getP1();
    	Point d = e1.getP2();
    	double x1 = a.getX(); double y1 = a.getY();
    	double x2 = b.getX(); double y2 = b.getY();
    	double x3 = c.getX(); double y3 = c.getY();
    	double x4 = d.getX(); double y4 = d.getY();
    	Point Q = intersection(E, e1);
    	
    	boolean answer = !(  ( Q.getX() < x1 && Q.getX() < x2 ) || ( Q.getX() > x1 && Q.getX() > x2 )
    			|| ( Q.getY() < y1 && Q.getY() < y2 ) || ( Q.getY() > y1 && Q.getY() > y2 )
    			|| ( Q.getX() < x3 && Q.getX() < x4 ) || ( Q.getX() > x3 && Q.getX() > x4 )
    			|| ( Q.getY() < y3 && Q.getY() < y4 ) || ( Q.getY() > y3 && Q.getY() > y4 )  );
    	if ( !answer && E.intersects(e1)) {
    		System.out.println("Wrong intersects method");
    	}
    	return answer;
    } **/
    
    public Point intersection(Edge E, Edge e1) { // find the intersection of two edges
			// slope of the first line
			double m1 = (E.getP2().getY() - E.getP1().getY() )/ (E.getP2().getX() - E.getP1().getX() );
			if ( Double.isInfinite(m1) ) {
				return E.infiniteSlope(e1);
			}
			double b1 = E.p2.getY() - m1*E.p2.getX();
			//slope of the second line
			double m2 = (e1.p2.getY() - e1.p1.getY() )/ (e1.getP2().getX() - e1.p1.getX() );
			if (Double.isInfinite(m2)) {
				return e1.infiniteSlope(E);
			}
			double b2 = e1.p2.getY() - m2*e1.p2.getX();
			double intx = (b2-b1)/(m1-m2);
			if ( m1 == m2) {
				System.out.println("ERROR - ZERO SLOPE");
			}
			
			if ( Math.abs(m1*intx + b1 - m2*intx - b2) > 4) { // ensure intersec. was calculated correctly
				System.out.println("ERROR LINES DON'T INTERSECT");
			}
			if ( Double.isNaN(intx) ) {
				//System.out.println(e1.getP1().toString() + " and " + e1.getP2().toString());
				System.out.println("Slopes are: " + m1 + " and " + m2 );
				System.out.println("y intercepts are: " + b1 + " and " + b2 );
			}
			double intyALT = m2*intx + b2;
		//	System.out.println("IntPoint at ("+ intx +  "," + inty + ")");
			return new Point(intx, intyALT);

	}
    
    public Point getIntersection(Edge E, Edge e1) { // find the intersection of two edges
		if ( E.doesIntersect( e1) ) {
			// slope of the first line
			double m1 = (E.getP2().getY() - E.getP1().getY() )/ (E.getP2().getX() - E.getP1().getX() );
			if ( Double.isInfinite(m1) ) {
				return E.infiniteSlope(e1);
			}
			double b1 = E.p2.getY() - m1*E.p2.getX();
			//slope of the second line
			double m2 = (e1.p2.getY() - e1.p1.getY() )/ (e1.getP2().getX() - e1.p1.getX() );
			if (Double.isInfinite(m2)) {
				return e1.infiniteSlope(E);
			}
			double b2 = e1.p2.getY() - m2*e1.p2.getX();
			double intx = (b2-b1)/(m1-m2);
			if ( m1 == m2) {
				System.out.println("ERROR - ZERO SLOPE");
			}
			
			if ( Math.abs(m1*intx + b1 - m2*intx - b2) > 4) { // ensure intersec. was calculated correctly
				System.out.println("ERROR LINES DON'T INTERSECT");
			}
			if ( Double.isNaN(intx) ) {
				//System.out.println(e1.getP1().toString() + " and " + e1.getP2().toString());
				System.out.println("Slopes are: " + m1 + " and " + m2 );
				System.out.println("y intercepts are: " + b1 + " and " + b2 );
			}
			double intyALT = m2*intx + b2;
		//	System.out.println("IntPoint at ("+ intx +  "," + inty + ")");
			return new Point(intx, intyALT);
		}
		else {
			try {
				throw new newException("No intersection between edges!!");
			}
			catch (newException EE) {
				System.err.println(EE);
				return new Point(0,0);
			}
		}
	}
    
    //find point moved epsilon in 'upwards' direction from the edge
    // input both edges the point is on, that is, E1 = (Q, p1) and E2 = (p1, R) 
    public Point translate(Polygon Poly, Point p, Edge E1, Edge E2) { 
    	
    	double e = 6;
    	try {
    		if ( !(E1.getP2().equals(p)) || !(E2.getP1().equals(p) ) ) {
    		    throw new newException("Error! Point is not the hinge of both edges");
        	}
    		else if ( (E1.getP2().getX() - E1.getP1().getX() == 0) && (E2.getP2().getX() - E2.getP1().getX()==0 ) ) {
        		throw new newException("ERROR! BOTH LINES ARE VERTICAL" );
        	}
    		else {
    			Point Q = E1.getP1();
    			Point R = E2.getP2();
    			double m1 = (E1.getP2().getY() - E1.getP1().getY() )/ (E1.getP2().getX() - E1.getP1().getX() );
    			double m2 = (E2.getP2().getY() - E2.getP1().getY() )/ (E2.getP2().getX() - E2.getP1().getX() );
    			// THE FIRST LINE IS VERTICAL 
    			if ( Math.abs(E1.getP2().getX() - E1.getP1().getX()) == 0) { // only first edge is vertical
    	    		double x = p.getX();
    	    		m2 = (E2.getP2().getY() - E2.getP1().getY() )/ (E2.getP2().getX() - E2.getP1().getX() );
    		    	double b2 = E2.getP2().getY() - m2*E2.getP2().getX();
    		    	double b2NEW = -e*Math.sqrt(m2*m2+1) + b2; 
    		    	if ( upIsInside(Poly, E2)) {
    		    		b2NEW = e*Math.sqrt(m2*m2+1) + b2; 
    		    	}
    		    	
    		    	if ( p.getY() > E1.getP1().getY() ) { // if P is at the bottom, shift left	
    		    	//	System.out.println("Case 1.2: First line vertical, going down. Shift left");
    		    //		System.out.println("Points: (" + Q.getX() + "," + Q.getY() + ")->(" +p.getX()+
    		    		//		","+p.getY() + ")->(" + R.getX()+ ","+ R.getY() + ")");
    		    		return new Point( (x-e), (m2*(x-e)+b2NEW) );
    		    	}
    		    	else { // if P is at the top, shift right 
    		    	//	System.out.println("Case 1.1: First line vertical, going up. Shift right");
    		    	//	System.out.println("Points: (" + Q.getX() + "," + Q.getY() + ")->(" +p.getX()+
    		    	//			","+p.getY() + ")->(" + R.getX()+ ","+ R.getY() + ")");
    		    		return new Point( (x+e), (m2*(x+e)+b2NEW));
    		    	}
    		    	
    	    	}
    			// THE SECOND LINE IS VERTICAL
    	    	else if ( Math.abs(E2.getP2().getX() - E2.getP1().getX()) == 0 ) { //only second edge is vertical
    	    		double x = p.getX();
    	    		m1 = (E1.getP2().getY() - E1.getP1().getY() )/ (E1.getP2().getX() - E1.getP1().getX() );
    		    	double b1 = E1.getP2().getY() - m1*E1.getP2().getX();
    		    	double b1NEW = -e*Math.sqrt(m1*m1+1) + b1; 
    		    	if ( upIsInside(Poly, E1)) {
    		    		b1NEW = e*Math.sqrt(m1*m1+1) + b1; 
    		    	}
    		    	
    		    	if ( p.getY() > E2.getP2().getY() ) { // if P is at the bottom, shift right   		
    		    		System.out.println("Case 2.2: Second line vertical, going down. Shift right");
    		    		System.out.println("Points: (" + Q.getX() + "," + Q.getY() + ")->(" +p.getX()+
    		    				","+p.getY() + ")->(" + R.getX()+ ","+ R.getY() + ")");
    		    		return new Point(  (x+e),   (m1*(x+e)+b1NEW) );
    		    	}
    		    	else { // if P is at the top, shift left 
    		    		System.out.println("Case 2.1: Second line vertical, going down. Shift left");
    		    		System.out.println("Points: (" + Q.getX() + "," + Q.getY() + ")->(" +p.getX()+
    		    				","+p.getY() + ")->(" + R.getX()+ ","+ R.getY() + ")");
    		    		return new Point(  (x-e),   (m1*(x-e)+b1NEW) );
    		    	}
    	    	}
    	    	else { //BOTH edges are not vertical 
    		    	// convert edges into line segments
    		    	m1 = (E1.getP2().getY() - E1.getP1().getY() )/ (E1.getP2().getX() - E1.getP1().getX() );
    		    	double b1 = E1.getP2().getY() - m1*E1.getP2().getX();
    		    	m2 = (E2.getP2().getY() - E2.getP1().getY() )/ (E2.getP2().getX() - E2.getP1().getX() );
    		    	double b2 = E2.getP2().getY() - m2*E2.getP2().getX();
    		    	// compute the new y intercepts
    		    	double b1NEW = -e*Math.sqrt(m1*m1+1) + b1; 
    		    	double b2NEW = -e*Math.sqrt(m2*m2+1) + b2; 
    		    	// should have a strictly descending chain R.getX() < p.getX() < Q.getX()
    		    	if ( upIsInside(Poly, E1) ) {
		    			b1NEW = e*Math.sqrt(m1*m1+1) + b1;
		    		}
		    		if ( upIsInside(Poly, E2) ) {
		    			b2NEW = e*Math.sqrt(m2*m2+1) + b2;
		    		}
    		    	// convert lines back into edges
    		    	Edge line1 = new Edge( new Point(0, b1NEW), new Point(1300, 1300*m1+b1NEW ) );
    		    	Edge line2 = new Edge( new Point(0, b2NEW), new Point(1300, 1300*m2+b2NEW) );
    		    	// return the intersection 
    		    	if ( intersects(line1, line2) ){ 
    		    		return getIntersection(line1, line2);
    		    	}
    		    	else {
    		    		return new Point(0,0);
    		    	}
    	    	}
    		}
    	}
    	catch(newException E) {
    		System.err.println(E);
    		return null;
    	}  	
    }
    
    // determine if the point P is above the line extrapolated from the segment E 
    public boolean above(Point P, Edge E) {
    	if ( E.getP2().getX() == E.getP1().getX() ) { // 'infinite' slope
    		return P.getX() < E.getP1().getX(); // return if P is to the left of E 
    	}
    	else {
    		double m = (E.getP2().getY() - E.getP1().getY() )/ (E.getP2().getX() - E.getP1().getX() );
    		double b = E.getP2().getY() - m*E.getP2().getX();
    		return P.getY() < m*P.getX()+b;
    	}
    }
   
    // add the points from A in order, starting at index i 
    public void addSection(int i, ArrayList<Point> A) {
    	
    	for (int j=i+1; j <A.size()+i+1; j++) {
    		Curve.points.add( j , A.get(j-(i+1)));
    		repaint();
    	}
    	for (int k=i; k<=A.size()+i; k++) {
    		Point P = Curve.points.get( (k)%Curve.points.size() );
    		Point Q = Curve.points.get( (k+1)% Curve.points.size() );
    		edgeAdd(  P, Q );
    		repaint();
    	}
    }
    
    // determine if the outside of the polygon is above or below the edge 
    public static boolean upIsInside(Polygon P, Edge E1) {
    	double epsilon = 5;
    	Point midpoint = new Point( (0.5*(E1.getP1().getX() + E1.getP2().getX())), 
    			(0.5*( E1.getP1().getY() + E1.getP2().getY()) ) );
    	double x = midpoint.getX();
    	double m = (E1.getP2().getY()-E1.getP1().getY()) / (E1.getP2().getX() - E1.getP1().getX() );
    	double b = E1.getP1().getY() - m*E1.getP1().getX();
    	double bNEW = -epsilon*Math.sqrt(m*m+1) + b;
    	
    	if ( E1.getP1().getX() == E1.getP2().getX() ) { // vertical line
    		// shift the midpoint over to the left a bit
    		return P.contains( new Point(midpoint.getX()-epsilon, midpoint.getY()+1 ));
    	}
    	double oldY = x * m+ b;
    	
    	double newY = x*m;
    	newY = newY + bNEW;
    	boolean answer;
    	boolean a2;
    	if ( Math.abs(E1.getP2().getX() - E1.getP1().getX() ) < 15) {
    		Point top = ( (E1.getP1().getY() < E1.getP2().getY() ) ? E1.getP1() : E1.getP2() );
    		Point bottom = ( top.equals(E1.getP1())  ? E1.getP2() : E1.getP1() );
    		boolean shiftLeft = top.getX() > bottom.getX();
    		double diff = Math.abs((E1.getP2().getX() - E1.getP1().getX())/ 4);
    		Point test = new Point( x + (shiftLeft ? -diff : diff), midpoint.getY()-5 );
    		answer = P.contains(test);
    		a2 = P.contains( new Point(x, newY));
    	}
    	else {
    		answer = P.contains( new Point( x, newY)  );
    		a2 = true;
    	}
    	boolean alternate = a2;
    	// evaluate the midpoint x coordinate shifted up epsilon, see if it is in the polygon
    	//boolean answer = P.contains( new Point(x, midpoint.getY()-4) );
    	return answer;
    }

    public void TitusII(Face F, boolean useTop) { // upper is 1, lower is 0
    	selectedFace = null;
    	highlightFaces = -1;
    	int side = ( useTop ) ? 1: 2;
    	titusSeq.add("IIa on side " + side + " of Face " + Curve.faces.indexOf(F) );
    	try {
    		if (F.getType() == 2) {
    			ArrayList<Point> newUpper = new ArrayList<Point>();
    			ArrayList<Point> upper = F.splitFirst(); // upper half of the original face
    			ArrayList<Point> lower = F.splitSecond(); // lower half of the original face
    			boolean CCW;  // determine if the face is CCW on the curve traversal
    			Polygon Poly = F.getPolygon();
    			Point start = upper.get(0);
    			Point stop = upper.get(upper.size()-1); 
    			//  upper(0) -> upper(1) is CCW direction since all faces are traversed CCW
    			// check if the curve traversal matches this direction

    			if ( start.getVector(0).getP2().equals(upper.get(1)) ||
    					start.getVector(1).getP2().equals(upper.get(1))) {
    					CCW = true;
    				}
    			else { CCW = false; }
    			
    			if ( !useTop ) {  // swap upper and lower 
    				ArrayList<Point> temp = upper; 
    				upper = lower;
    				lower = temp;
    			}
    			
    			boolean empty = false;
    			Point R, S;
    			R = S = null;
    			
    			if ( upper.size() == 2 ) {//store the 2 int points on empty upper 1/2 face
    				R = upper.get(0);
    				S = upper.get(1);
    				empty = true;
    			}
    			if ( lower.size() == 2 ) { //swap upper + lower so "upper" 1/2 face is empty (with non int points)
    				R = lower.get(0); 
    				S = lower.get(1);
    				ArrayList<Point> A = new ArrayList<Point>();
    				A = upper;
    				upper = lower;
    				lower = A;
    				empty = true; 
    			} 
    			if ( empty ) { 
    				// switch points, if necessary, so R comes first 
	    			Point newR, newS;
	    			newR = translate(Poly, R, new Edge(lower.get(lower.size()-2), R), new Edge (R, S) );
	    			newS = translate(Poly, S, new Edge(R, S), new Edge (S, lower.get(1)) );
	    			newUpper.add(newR);
	    			newUpper.add(newS); 
    			} 
    			else {
	    			for (int i=0; i < upper.size(); i++) {
	    				Point P = upper.get(i);
	    				if ( i==0 || i==upper.size()-1) { // working with an INT POINT! 
	    					int j = F.getPoints().indexOf(P);
	    					Point p1;
	    					// p1 is the point on the face preceding P (the i^th point) 
	    					if ( j==0 ) { 
	    						p1 = F.getPoint( F.getSize()-1 ); 
	    					}
	    					else {
	    						p1 = F.getPoint( (j-1)%(F.getSize()) );
	    					}
	    					Point p2 = F.getPoint( (j+1)%(F.getSize()) );
	    					Point bottom = null;
	    					
	    					try { // find the point on the bottom 1/2 face connected to the int point
		    					if ( lower.contains(p1) ) {
		    						bottom = p1;
		    					}
		    					else if ( lower.contains(p2) ) {
		    						bottom = p2;
		    					}
		    					else {
		    						throw new newException("Points adjacent to int points "
		    								+ "not on either side of the face");
		    					}
	    					}
	    					catch (newException E) {
		    					System.err.println(E + "");
		    					return;
		    				}
	    					// working with intersection points
	    					if (i==0) { 
	    						Point next = upper.get(1); 
	    						Point newPoint = translate(Poly, P, new Edge(bottom, P), new Edge(P, next));
	    						newUpper.add(newPoint);
	    					}
	    					else { 
	    						Point prev = upper.get(upper.size()-2);
	    						Point newPoint = translate(Poly, P, new Edge(bottom, P), new Edge(P, prev));
	    						newUpper.add(newPoint);
	    					}
	    				}
	    				else { // working with a non-INT point
	    					Point prev = upper.get( (i-1)%(upper.size()));
	    					Point next = upper.get( (i+1)%(upper.size()));
	    					// not working with an intersection point
	    					Point newPoint = translate(Poly, P, new Edge(prev, P), new Edge(P, next));
	    					newUpper.add(newPoint);
	    				}
	    			} 
    			}
    			
    			Point keyInt = null; // intersection point connected to the connector
				Point connector = null;  // initial point which the new upper face will be connected to
				int j;
				// find the point AFTER the connector, the first one on the bottom 1/2 face
				if (lower.contains(start.getVector(0).getP2() ) ) {
					keyInt = start;
					Point P = start.getVector(0).getP2();
					j = Curve.points.indexOf(P);
				}
				else if (lower.contains(start.getVector(1).getP2() ) ) {
					keyInt = start;
					Point P = start.getVector(1).getP2();
					j = Curve.points.indexOf(P);
				}
				else if (lower.contains(stop.getVector(0).getP2() ) ) {
					keyInt = stop;
					Point P = stop.getVector(0).getP2();
					j = Curve.points.indexOf(P);
				}
				else if (lower.contains(stop.getVector(1).getP2() ) ) {
					keyInt = stop;
					Point P = stop.getVector(1).getP2();
					j = Curve.points.indexOf(P);
				}
				else {
					try {
						throw new newException("Connector not found!!");
					}
					catch (newException E) {
						System.err.println(E);
						return;
					}
				}
				int m; // index of the connector in the points array
				// the connector is the point PREVIOUS to P on the curve
				if ( j==0 ) { m = Curve.points.size()-1; } else { m = j-1; }
				connector = Curve.points.get(m); 
				
				for (Point p: lower) { // delete all points on the old lower face
					removePoint(p);
				}
    				
				boolean reverse = false;
				// (if upper/lower 1/2 face is empty) 
				if ( upper.size() == 2) {//reverse if the keyInt is the 2nd int point
					reverse = keyInt.equals(S);
				}
				else if ( keyInt.equals(stop) ) { // reverse if keyInt is 2nd int point
					reverse = true;
				}
				
				if ( upper.size() != 2 && !useTop) {
					reverse = !reverse; // use opposite orientation since on the botton now
				}
				
    			/**  reverse the new upper section depending on the direction 
    			 * that the bigon F was traversed!! - if F was traversed CW, 
    			 * then maintain direction, if CCW, then reverse,
    			 *  b/c face is necessarily in CCW orientation
    			 * from the face alteration during instantiation **/
    				// need to flip the order if face was traversed in counter-clockwise order
    				if ( reverse ) {
    					ArrayList<Point> A = new ArrayList<Point>();
    					for (int k=newUpper.size()-1; k>=0 ; k--) {
    						A.add( newUpper.get(k) );
    					}
    					addSection( Curve.points.indexOf(connector), A);
    				}
    				else { // if traversed clockwise, preserve the order
    					addSection( Curve.points.indexOf(connector), newUpper);
    				}
    		}
    		else { 
    			throw new newException("Not a bigon! Can't apply a IIa move!");
    		}
    		
    		boolean g1 = faceReps.remove(F.getRep());
        	boolean g2 = Curve.faces.remove(F);
        	if ( !g1 || !g2 ) { // check that the face and its rep were removed
        		if ( ! g1 ) { System.out.println("Face rep not removed!!!"); }
        		else if ( ! g2 ){ System.out.println("Face not removed!!!"); }
        		else {
        			System.out.println("Neither face nor face rep removed!");
        		}
        	}
        	
        	//distanceCheck(); 
        	
        	areaSwept += F.getArea();
        	areaSwept = round(areaSwept);
        //	selfOverlappingCurves.clear();
        //	findCurves(root, Curve);
        	numberPoints(true); // re-number the points
    	}
    	
    	catch (newException e) {
    		System.err.println(e);
    	}
    }
    
    public Edge directedGrab(Edge E) {
    	for (Edge e: directedEdges) {
    		if ( e.equals(E) ) {
    			return e;
    		}
    	}
    	try {
    		throw new newException("ERROR! This edge isn't on the curve!");
    	}
    	catch (newException EE) {
    		System.err.println(EE);
    		return null;
    	}
    }
    
    public void TitusIII(Face F, int part) {
    	selectedFace = null;
    	highlightFaces = -1;
    	try {
    		if ( F.getType() != 3) {
    			throw new newException("Cannot apply a IIIa move. Input Face is not a triangle.");
        	}
    	}
    	catch (newException EE) {
    		System.err.print(EE);
    		return;
    	}
    	titusSeq.add("III on side " + (part+1) + " of Face " + Curve.faces.indexOf(F));
    	
    	ArrayList<Point> section = new ArrayList<Point>();
    	try {  // get the subsection of the face with which to apply the III move  
	    	if ( part == 0) { section = F.split(0); }
	    	else if ( part ==1) { section = F.split(1); }
	    	else if ( part == 2) { section = F.split(2); } 
	    	else { throw new newException("Can only operate on either part 0, 1, or 2"); }
    	}
    	catch ( newException EE) { 
    		System.err.println(EE); 
    		return;
    	}
    	
		Point P = section.get(0); // get the 1st int point of the section 
		Point Q = section.get(section.size()-1); // get the 2nd int point of the section		
		//  section is ordered CCW since it is a subsection of a face 
		Edge e = directedGrab( new Edge(P, section.get(1))); // get the edge w/orientation from the curve
		boolean CCW = P.equals( e.getP1() ); // check if e has same orientation as path P -> Q
		
		ArrayList<Point> newPart = new ArrayList<Point>(); // the new section to be inserted
		ArrayList<Point> restOfFace = F.restOfCurve(part); // get the rest of the face in CCW orientation
		newPart.add( translate(F.getPolygon(), Q, new Edge(section.get(section.size()-2), Q), 
				new Edge(Q, restOfFace.get(1) ) ) );
		for (int j=1; j < restOfFace.size()-1; j++) {
			Point current = restOfFace.get(j); // grab each point, with the ones b4 and after
			Point prev = restOfFace.get(j-1);
			Point next = restOfFace.get(j+1);
			// shift the point and add it to the new section  
			newPart.add( translate(F.getPolygon(), current, new Edge(prev, current), 
					new Edge(current, next) ) );
		}
		newPart.add( translate(F.getPolygon(), P, new Edge(restOfFace.get(restOfFace.size()-2), P), 
				new Edge(P, section.get(1) ) ) );
		
		
		Point start = (CCW) ? P : Q;
		Point oppo; // point AFTER the connector on the curve;
		int j;
		boolean reverse = start.equals(P);
		Point connector; 
		
		if ( section.size() == 2 ) { // case when P and Q are connected directly 
			Edge keyEdge;
			try {
				// find the edge keyEdge which has P and Q as int points 
				if ( P.getEdge(0).getIntPoints().contains(Q) ) {
					keyEdge = P.getEdge(0);
				}
				else if ( P.getEdge(1).getIntPoints().contains(Q) ) {
					keyEdge = P.getEdge(1);
				}
				else {
					throw new newException("Both of the edges of P did not contain Q as an intersection point");
				}
			}
			catch (newException EEE) {
				System.err.println(EEE);
				return;
			}
			 // determine the orientation of the keyEdge on the curve 
			Point go,end; 
			if ( keyEdge.getP1().getNum() < keyEdge.getP2().getNum() ) {
				go = keyEdge.getP1();
				end = keyEdge.getP2();
			}
			else {
				go = keyEdge.getP2();
				end = keyEdge.getP1();
			}
			 // swap start and stop b/c 0 ≈ points.size() (mod points.size) 
			if ( end.equals(Curve.points.get(Curve.points.size()-1)) && go.equals(Curve.points.get(0)) ) {
				Point temp = end; 
				end = go;
				go = temp;
			}
			// the first point on the keyEdge is the connector 
			edgeRemove( keyEdge.getP1(), keyEdge.getP2() ); // this edge no longer exists
			connector = go;
		}
		else { 
			// find the point connected to the starting point, call that OPPO
			if (section.contains(start.getVector(0).getP2() ) ) {
				oppo = start.getVector(0).getP2();
				j = Curve.points.indexOf(oppo);
			}
			else if (section.contains(start.getVector(1).getP2() ) ) {
				oppo = start.getVector(1).getP2(); 
				j = Curve.points.indexOf(oppo);
			}
			else {
				try {
					throw new newException("Connector not found!!");
				}
				catch (newException E) {
					System.err.println(E);
					return;
				}
			}
			int n; // index of the connector in the points array
			// the connector is the point PREVIOUS to OPPO on the curve
			if ( j==0 ) { n = Curve.points.size()-1; } else { n = j-1; }
			connector = Curve.points.get(n); 
		}
		
		for (Point p: section) { // delete all points of the section 
			removePoint(p);
		}
		removeIntPoint(P);  // remove the int points from the section 
		removeIntPoint(Q);
		
		// if order is  start < P < Q, then don't reverse 
		if ( !reverse ) {
			addSection( Curve.points.indexOf(connector), newPart );
		}
		else { // otherwise we have P < Q< start, so reverse 
			
			ArrayList<Point> A = new ArrayList<Point>();
			for (int k=newPart.size()-1; k>=0 ; k--) {
				A.add( newPart.get(k) );
			}
			addSection( Curve.points.indexOf(connector), A );
		}
		 	
		boolean g1 = faceReps.remove(F.getRep());
    	boolean g2 = Curve.faces.remove(F);
    	if ( !g1 || !g2 ) { // ensure the face F and its rep were removed 
    		try {
	    		if ( ! g1 ) {  throw new newException("Face rep not removed during Titus Ia move!"); }
	    		else if ( ! g2 ){ throw new newException("Face F not removed during Titus Ia move!"); }
	    		else {
	    			throw new newException("Neither face nor face rep removed during Titus Ia move!");
	    		}
    		}
    		catch (newException EE) {
    			System.err.println(EE);
    		}
    	}
    	
    	areaSwept += F.getArea();  // keep track of the area of the face 
    	areaSwept = round(areaSwept);
    	//selfOverlappingCurves.clear();
    	//findCurves(root, Curve);
    	numberPoints(true);
    	
    }
    
    public char getLetter(int i) { // a ≈ 0, b ≈ 1, ... , z ≈ 25
    	if ( i < 0 || i > 25) {
    		try { 
    			throw new Exception("a == 0, z == 25. Input an integer between these values");
    		}
    		catch (Exception E){
    			System.err.println(E);
    		}
    	}
    	return (char)(i+97);
    }
    
   // public ArrayList<String> generateIntSeq() { 
    	
  //  }
    
    public int TNS(String intSeq) {  // compute the total nested splits (# splitting trees) of the intSeq
    	if ( intSeq.length() == 0) {
    		return 1;
    	}
    	int total = 0;
    	ArrayList<Character> intPoints = new ArrayList<Character>();
    	for (int i=0; i< intSeq.length() ; i++) {
    		char cur = intSeq.charAt(i);
    		if ( !intPoints.contains(new Character(cur)) && splitSeq(intSeq, cur, false) != null ) {
    			intPoints.add( new Character(cur) );
    		}
    	}
    	for ( Character intPoint: intPoints) {
    		char c = intPoint.charValue(); // take the character
    		// add to the sum the product of the total nested splits of the left and right
    		total += TNS( splitSeq(intSeq, c, false) ) * TNS(splitSeq(intSeq, c, true) );
    	}
    	return total; 
    }
    public String splitSeq(String intSeq, char c, boolean half) {
    	int count = 0;
    	int ind1 = -1; int ind2 = -1;
    	for (int i=0; i < intSeq.length() ; i++) {
    		char s = intSeq.charAt(i);
    		if ( s == c) {
    			count ++;
    		}
    		if ( ind1 == -1 ) { ind1 = i; } 
    		else { ind2 = i; }
    	}
    	if ( count != 2) {
    		return null;
    	}
    	String direct = "";
    	String wraparound = "";
    	// direct 
    	for (int i= ind1+1; i < ind2; i++) { // take substring strictly btwn ind1 < j < ind2
    		direct += intSeq.charAt(i);
    	}
    	for (int j= 0; j < ind1; j++) { // take substring strictly btwn ind1 < j < ind2
    		wraparound += intSeq.charAt(j);
    	}
    	//wraparound
    	for (int j= ind2 + 1; j < intSeq.length() ; j++) { // take substring strictly btwn ind1 < j < ind2
    		wraparound += intSeq.charAt(j);
    	}
    	return (half ? wraparound : direct);
    }
   
    public Curve splitCurve(Point intPoint, String half, Curve curve) {
    	if ( curve == null) {
    		return null;
    	}
    	ArrayList<Point> curveNEW = new ArrayList<Point>();
    	// copy points over so the originals are unaffected
    	ArrayList<Point> used = new ArrayList<Point>();
    	for (Point P: curve.getCurve() ) { 
    		if ( !contains(used, P) ) {
    			used.add(P);
    			Point Q = new Point(P.getX(), P.getY() );
        		Q.setNum( P.getNum()); // keep the #, for debugging purposes
        		Q.setIndex( P.getIndex(0) ); Q.setIndex( P.getIndex(0) );
        		Q.setIntPointIndex( P.getIntPointIndex() ); // in case we have an int point passed down
        		Q.name( P.toString() ); // pass down names to int points
        		
        		curveNEW.add( Q );  
    		}
    		else {
    			Point add = null;
    			for (Point Q: curveNEW) {  // P was copied already, so find the copy Q in the array and add it again
    				if (Q.equals(P)) { 
    					add = Q; 
    					break; 
    				}
    			}
    			try { if (add == null) { throw new Exception("Point was used but could not be found"); }     		}
    			catch ( Exception EE) { System.err.println(EE); return null; }
    			curveNEW.add( add );
    		}
    		
    	}
    	try { // ensure that the intpoint is on the curve twice
	    	int count = 0;
	    	for (Point P: curve.getCurve() ) {
	    		if ( P.equals(intPoint) ) { 
	    			count += 1; 
	    		} 
	    	}
	    	if ( count != 2 ) { throw new Exception("Point does not split on this curve."); }
    	}
    	catch (Exception EE) { 
    		//System.out.println(EE); 
    		return null; 
    	}
    	boolean first = false;
    	boolean go = false;
    	ArrayList<Point> firstHalf = new ArrayList<Point>(); 
    	ArrayList<Point> secondHalf = new ArrayList<Point>();

    	if ( intPoint.getIndex(0) == -1) {
    		//System.out.println("The point with index -1 was " + intPoint.toString() );
    	}
    	for (Point P: curveNEW) { // add all points up to intPoint to the second half
    		if ( ! P.equals(intPoint) ) {
    			secondHalf.add( P );
    		}
    		else { 
    			break;
    		}
    	}  
    	for (Point P: curveNEW) { 
    		if (P.equals(intPoint) ) {
    			go = true;
    			first = !first;
    		}
    		if ( go ) {
    			( (first) ? firstHalf: secondHalf).add(P);
    		}
    	}
    	int count = 0;
    	for (int i=0; i< firstHalf.size(); i++) {
    		Point now  = firstHalf.get(i);
    		Point next = firstHalf.get( (i+1) % firstHalf.size() );
    		
    		if ( !curve.adjacent(now, next) ) {
    			count ++;
    		}
    		
    	}
    	count = 0;
    	for (int i=0; i< secondHalf.size(); i++) {
    		Point now  = secondHalf.get(i);
    		Point next = secondHalf.get( (i+1) % secondHalf.size() );
    		if ( !curve.adjacent(now, next) ) {
    			count ++;
    		}
    		
    	}
    	if (count > 0 ) { System.err.println("Error. Points not connected."); } 
    	Curve C1 = new Curve(secondHalf); 
    	Curve C2 = new Curve(firstHalf);
    
    	if ( half.equals("direct") ) {
    		return C2;
    	}
    	else if ( half.equals("wrap") ) {
    		return C1;
    	}
    	else { 
    		try { throw new Exception("ERORR. Neither direct nor wrap split was asked for"); }
    		catch ( Exception E)  { System.err.println(E); }
    		return null;
    	}
    } 
    
    // return the root of the splitting tree
    
    public Homotopy minHomotopy( Curve C, int minIndex ) {
    	ArrayList<Point> curve = C.getCurve(); 
    	if ( curve.size() <= 2) { 
    		baseCases++;
    		return new Homotopy();
    	}
    	
    	boolean noIntPoints = true;
    	ArrayList<Point> reverseCurve = new ArrayList<Point>();
    	for (int j = curve.size()-1; j >= 0; j--) {
    		if ( curve.get(j).isIntersection() )  { noIntPoints = false; }
    		reverseCurve.add( curve.get(j) );
    	}
    	// SELF-OVERLAPPING METHOD ONLY WORKS FOR CCW CURVES!
    	boolean overlapping = isSelfOverlapping(C) || isSelfOverlapping(C.reverse());
    	if ( overlapping || noIntPoints ) {
    		baseCases++;
    		ArrayList<Integer> Anchor = new ArrayList<Integer>();
    		Point anchor;
    		if ( C.anchor != null && C.anchor.toString().substring(0, 1).equals("0")) { // extra check in case of 0 
    			Anchor.add( new Integer(0) );
        		return new Homotopy(C.windingArea(), Anchor);
    		}
    		if ( C.anchor != null && C.intPoints.contains(C.anchor) ) {
    			anchor = C.anchor;
    		}
    		else 
    			anchor = C.getAnchor();
			String S = anchor.toString();
    		Anchor.add(Integer.parseInt(S.substring(0, S.length()-1 ) ) );
    		return new Homotopy(C.windingArea(), Anchor);
    	}
    	
    	ArrayList<Point> candidateIntPoints = new ArrayList<Point>();
    	
    	for ( int i= C.intPoints.size()-1; i >=0; i--) {
    		Point P = C.intPoints.get(i);
    		Curve gammaP = splitCurve(P, "direct", C);
    		if ( gammaP != null) {
	    		boolean SO = isSelfOverlapping( gammaP ) || isSelfOverlapping( gammaP.reverse() );
	    		
	    		if (P.isIntersection() && gammaP != null && (gammaP.intPoints.size() == 0 || SO) ) {
	    			if ( !contains(candidateIntPoints, P) ) {
	    				candidateIntPoints.add(P);
	    			}
	    			// determine minimum homotopy on each half of the curve that splits about P 
	    		}
    		}
    	}
    	
    	Hashtable<Point, Double> windingAreas = new Hashtable<Point, Double>();
    	
    	for (Point P: candidateIntPoints){ 
    		Curve directSplit = splitCurve(P, "direct", C); // compute the splits 
    		Curve wrapSplit = splitCurve(P, "wrap", C);
	    	//directSplit.numberPoints(true); // number the points in prep to find winding Area
	    	//wrapSplit.numberPoints(true);
	    	// compute the winding areas, add the sum to the Hashtable 
	    	Double wArea = new Double( round(directSplit.windingArea() + wrapSplit.windingArea(), 2) );
	    	windingAreas.put(P, wArea);
    	}
    	
    	candidateIntPoints = sort(candidateIntPoints, windingAreas); // sort the table
    	
    	Double minArea = new Double(Integer.MAX_VALUE);
    	Homotopy bestHomotopy = new Homotopy( Integer.MAX_VALUE, new ArrayList<Integer>() );
    	
    	for (Point P: candidateIntPoints){ 
    		if ( windingAreas.get(P) < (minArea-.03) ) {
	    		int index = Integer.parseInt(P.toString().substring(0, P.toString().length()-1));
				if ( index < minIndex) {
					Curve directSplit = splitCurve(P, "direct", C); // compute the splits 
					Curve wrapSplit = splitCurve(P, "wrap", C);
					wrapSplit.anchor = C.curve.get(0);
					Homotopy direct = minHomotopy(directSplit, index);
					Homotopy wrap = minHomotopy(wrapSplit, index);
					Homotopy curHomotopy = direct.add(wrap);
					Double cur = new Double(curHomotopy.area);
			    	if ( cur < (minArea-.03) ) {
			    		minArea = cur;
			    		bestHomotopy = curHomotopy;
			    	}
				}
    		}
    	}

		return bestHomotopy;
    }
    
    /***********************************************************************************************/
    
    public ArrayList<Point> sort (ArrayList<Point> candidates, Hashtable<Point, Double> windingAreas) {
    	for (int k=0; k < candidates.size(); k++) { // index to start searching 
    		Double min = new Double(1000000);
    		Point best = null;
	    	for (int i= k; i<candidates.size(); i++) {
	    		Point P = candidates.get(i);
	    		if (windingAreas.get(P) < min) { 
	    			best = P; 					// update current minimum
	    			min = windingAreas.get(P);
	    		}
	    	}
	    	int index = candidates.indexOf(best); // index of the k^th minimum
	    	Point old = candidates.get(k);		// point being swapped with best 
	    	if ( !best.equals(old) ) {
		    	candidates.remove(best);
		    	candidates.remove(old);	// remove old
		    	candidates.add(k, best);  // replace old with best 
		    	candidates.add(index, old); // add old back in new index
	    	}
    	}
    	return candidates;
    }
    
    
   
    
    private class MyMouseAdapter extends MouseAdapter{
		
    	
		// mouse released to set down points
		public void mouseReleased(MouseEvent e) {
			
			print = true;
			moving = false;
			
			/** 
			if ( smoothDraw) {
				closeCurve();
				smoothDraw = false;
			}**/
			
			if ( complete && pointsSelected.size() > 0 ) {
				pointsSelected.clear(); // cannot add points once the curve is complete
			}
			
			if ( curPoint == null) { // if not adding a point, for whatever reason
				//System.out.println("NULL CURPOINT! - Point deleted!");
			}
			else if (Curve.points.contains(curPoint)) { // if moving an old point
				
				if ( curPoint.distanceMoved() >0 && curPoint.closestVal(allPoints) < THRESHOLD ) { // if too close to an existing point
					perturb(curPoint); // move around till not too close 
				}
				else if (curPoint.distanceMoved() < THRESHOLD/4 && !curPoint.isNew()) { // if point move was negligible, do nothing
					//System.out.println("Negligible movement: " + curPoint.distanceMoved());			
				} 
				else {
					movePoint(curPoint, e.getX(), e.getY()); 
					curPoint.updateLocation( e.getX(), e.getY() ); // set new location
				}	
			}
			
			if ( built & (!complete || allowMovement) ) {
				numberPoints(true); // re-number points
				repaint();
			} 
			
			curPoint = null; // set the current point to be moved to null
			if ( !complete) {
				repaint();   
			}
		}
		
		
		public void mouseClicked(MouseEvent e){ // do nothing during a double click
			
			ArrayList<Point> points = Curve.points;
			
			if ( justFinished ) { 
				points.get(0).unSelect(); 
				justFinished = false; // ensure statement isn't entered again
				pointsSelected.clear();  
				return; 
			}
			
			int s = points.size(); 
			int ind1=-1000; int ind2=-2000; //indices of point already selected, candidate for selection
									// in order of size 
			
			curPoint = new Point(e.getX(), e.getY() );
			
			for (Point p: points) {
				if ( curPoint.distance(p) < THRESHOLD ) {
					curPoint = p;
				}
			}
			
			if ( pointNew  ) { // new point was added; it's no longer new
				pointNew = false;
				curPoint.unSelect();
				pointsSelected.remove(curPoint);
				numberPoints(true);
				repaint();
				return;
			}
			
			int r = points.indexOf(curPoint);
			int t = (pointsSelected.size()==1) ? points.indexOf(pointsSelected.get(0)) : -1;
			ind1 = ( r < t) ? r : t;
			ind2 = ( r > t) ? r : t;
			
			// unselected non-intersection point was clicked on, first point to be selected
			if (  !curPoint.isNew() && built & !complete && points.contains(curPoint) 
					&& !curPoint.isSelected() && pointsSelected.size()==0) { // point was selected 
				System.out.println("this is non-int point # " + points.indexOf(curPoint));
				curPoint.select();
				pointsSelected.add(curPoint);
			}
			// unselected point clicked on, only one other point selected, points are adjacent
			else if ( points.contains(curPoint ) && !curPoint.isSelected() && pointsSelected.size()==1
					&& (  Math.abs( points.indexOf(curPoint) - points.indexOf(pointsSelected.get(0)) )== 1   
					|| ( (ind2+1)%s == 0 && ind1 == 0 ) ) ){ // point was selected 
				curPoint.select();
				pointsSelected.add(curPoint);
			}
			else if ( curPoint.isSelected() ) { // un-select the point
				curPoint.unSelect();
				pointsSelected.remove(curPoint);
			}
			
			if ( built & !complete || allowMovement) {
				numberPoints(true); // re-number points
				repaint();
			} 
			
			if ( complete && curPoint!=null ) {
				double d = curPoint.closestVal(faceReps);
				if ( d < THRESHOLD ) {
					// get the index of the close face
					int i = faceReps.indexOf(curPoint.closest(faceReps));
					try {
						if ( Curve.faces.size() != faceReps.size() ) {
							throw new newException("There are not the same number of faceReps as faces.");
						}
						else if ( Curve.faces.get(i).getType() == 0 ) {
							throw new newException("Cannot select this face. It is not a 1-gon, 2-gon, or a 3-gon.");
						}
					}
					catch (newException EE) {
						System.err.println(EE);
					}
					int type = Curve.faces.get(i).getType();
					
					if ( (type == 1 || type == 2 || type == 3) && !allowMovement ) {
						if ( selectedFace == null || !Curve.faces.get(i).equals(selectedFace) ) {
							selectedFace = Curve.faces.get(i);
							frame.east.buttonPanel.type = type;
							frame.east.buttonPanel.update();
						}
						else if ( selectedFace != null && Curve.faces.get(i).equals(selectedFace) ) {
							selectedFace = null;
							frame.east.buttonPanel.type = -1;
							frame.east.buttonPanel.update();
						}
						highlightFaces = -1;
						frame.east.colorPanel.highlightFaces = -1;
					}
					repaint();
				}
			}

			if (built && (!complete || allowMovement) ) {
				numberPoints(true);
				repaint();
			}
			
			curPoint = null;
	    }
		
		public void mouseDragged (MouseEvent e)  {
			
			if ( !complete || allowMovement ) {
				if(SwingUtilities.isLeftMouseButton(e)){
					// move point around till user sets the mouse back down
					if ( curPoint != null) {
						movePoint(curPoint, e.getX(), e.getY()  );
					}
				}
				print = false;
				moving = true;
				repaint();
			}
		}

		public void mouseMoved(MouseEvent e) { 

			if ( smoothDraw && go) {
				smoothPoint.setX( e.getX() );
				smoothPoint.setY( e.getY() );
				
				if ( smoothPoint.distanceMoved() > 2*THRESHOLD ) {
					addPoint( smoothPoint.copy() );
					smoothPoint.updateLocation(e.getX(), e.getY());
				}
				repaint();
				return;
			}
		}

		public void mousePressed (MouseEvent e){
			
			moving = false;
			boolean addPoint = false;
			boolean deletePoint = false;
			
			if (e.getButton() == MouseEvent.BUTTON1) { 
				if ( smoothDraw && !go) {
					smoothPoint = new Point(e.getX(), e.getY() );
					go = true;
				}
				else if ( smoothDraw && go) {
					smoothDraw = false;
					go = false;
					continuousDrawProcess();
					return;
				}
				curPoint = new Point(e.getX(),e.getY()); //save current location
				boolean add = true;
				boolean close = false;
				if ( Curve.points.size()> 0 ) {
					for (Point p: Curve.points) {
						if ( p.distance(curPoint) < THRESHOLD) {
							curPoint = p;
							// ensure point is not over-used and that edge isn't traversed and re-traversed
							 if ( Curve.points.size() >= 3 && p.equals( Curve.points.get(0) ) && !built ) {
								//p.use();
								add = false;
								close = true;
								built = true;
							}
							else { // otherwise just
								add = false;
								break;
							}
						}
					}
				}
				if ( !complete ) {
					 // close the curve, remember that a point was "technically" added, un-select the point
					if ( !complete && close ) { closeCurve(); addPoint = true; justFinished = true; }
					if ( add && pointsSelected.size() == 2 ) {
						addPointBtwn(curPoint);
					}
					else if ( add && !built ) { // if point is non-degenerate, add it!
						addPoint = true;
						addPoint(curPoint); 
					}
				}
			}
				
			if ( (!complete || allowMovement) && built) { 
				if (e.getButton()==MouseEvent.BUTTON3) {
					// Search through list and delete nearest neighbor within some threshold
					int nearestIndex = -1;
					double nearestDistance = Double.MAX_VALUE;
					for(int i=0; i<Curve.points.size(); i++){
						Point current = Curve.points.get(i);
						if(current.distance(e.getPoint())<nearestDistance){
							nearestIndex = i;
							nearestDistance=current.distance(e.getPoint());
						}
					}
					if(nearestDistance<=THRESHOLD ){
						//Delete point
						Point p = Curve.points.get(nearestIndex); // point about to be deleted
						if ( contains( pointsSelected , p ) ) {
							pointsSelected.remove(p);
						}
						deletePoint = true;
						Curve.deletePoint(p);
						if (Curve.points.size() < 2) {
							curPoint = null;
						}
					}
				}
				print = true;
				if ( built && (addPoint || deletePoint)   ) {
					numberPoints(true);
				}
				else if ( built || allowMovement  ) {
					numberPoints(false);
				} 
				repaint();
				if ( Curve.points.size() <= 2 ) { 
					built = false;
				} 
			}
		}
    }
}
