import javax.swing.*;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Scanner;
import java.awt.geom.*;
import java.io.File;
import javax.swing.JFrame;
import javax.swing.JButton;
import javax.swing.JFileChooser;


public class DrawingPanelFrame extends JFrame {
	
	public DrawingPanel dP;
	public EastPanel east;
	boolean saved; 
	
	public static void main(String[] args) {
		DrawingPanelFrame frame = new DrawingPanelFrame("Homotopy Visualizer"); 
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	}
   
	public class newException extends Exception {
        public newException(String message) {
            super(message);
        }
    }
	
	public DrawingPanelFrame(String title){
		super(title);
		setLayout(new BorderLayout() );
		this.dP = new DrawingPanel(this);
		this.east = new EastPanel(this);
		dP.setSize(new Dimension(500,500));
		add(east,  BorderLayout.EAST );
		add(dP, BorderLayout.CENTER);
		this.add(dP);
		createMenubar();
		
		setSize(new Dimension(1300,750));
	    //pack();   // sets the frame's size to accommodate its components
	    setLocation(10,100);
	    setVisible(true); // displays the frame
	}
	
	
	public class incompleteCurveException extends Exception {
        public incompleteCurveException(String message) {
            super(message);
        }
    }
    
	public void colorCurves() { 
		double size = dP.selfOverlappingCurves.size();
		
		if ( size > 0 ) {
			for ( int j=0; j< dP.selfOverlappingCurves.size(); j++ ) {
				Curve C = dP.selfOverlappingCurves.get(j);
				int k = (int)(250/(size-1) * j );
				Color c;
				if ( j % 3 ==0 ) {
					c = new Color( 250, (int)(220/(size)*j), (int)(220/(size-1)*j));
				}
				else if ( j % 3 == 1) {
					int i = j/3;
					c = new Color( (int) ((210*i)/(size/3)) ,(int)((140*i)/(size/3)) , (int)((250*i)/(size/3)) );
				}
				else {
					c = new Color((int)(100/size*j),  k , (int) 240 ) ;
				}
				
				C.setColor(c);
				C.anchor.color = c;	
			}
		}
	}
	
    public void restart() {
    	dP.titusSeq.clear(); // reset the Titus sequence
    	
    	try {
			if ( !saved ) {
				throw new newException("Cannot restart - curve was never completed");
			}
		}
		catch (newException E) {
			System.err.println(E);
			System.exit(0);
		}
		
		try{
	        Scanner scanner = new Scanner(new File("current"));  	        
	    	int x, y;
	    	
	    	ArrayList<Point> P = new ArrayList<Point>();
	    	while(scanner.hasNext()){
	    		// New point
	    		x = (int)scanner.nextDouble();
	    		y = (int)scanner.nextDouble();
	    		System.out.println("Point "+x+","+y);
	    		P.add(new Point(x, y) );
	    	} 
	    	try {
    	    	//dP.clear();  // don't want to erase titus sequence info
	    		dP.Curve.clear();
	        	dP.allPoints.clear();
	        	dP.Curve.edges.clear();
	        	dP.shortEdges.clear();
	        	dP.Curve.faces.clear();
	        	dP.intSeq = "";
	        	dP.Curve.intPoints.clear();
	        	dP.faceReps.clear();
	        	dP.built = false;
	        	dP.complete = false;
	        	dP.areaSwept = 0;
	        	dP.go = false;
	        	dP.splits = 0 ;
	        	
    	    	if ( P.size() >= 3 ) {
    	    		for (int i=0; i < P.size(); i ++) {
    	    			Point p = P.get(i);
    	    			dP.addPoint(p);
    	    		}
    	    		dP.closeCurve();
    	    		dP.startingWindingArea = dP.windingArea;
    	    		dP.complete = true;
    	    		Counter C = new Counter();
    	    		/**
    	    		if ( dP.Curve.intPoints.size() <= 9) {
    	    			dP.splits = C.TNS( C.convert(dP.intSeqIntegers) );
    	    		} **/
    	    		if ( dP.minimumHomotopyArea != -1 ) {
    	    			dP.anchorPoints.clear();
	    	    		dP.selfOverlappingCurves.clear();
	    	    		
	    	    		// get the anchor points from the indices 
	    	    		
	    	    		for ( Integer I: dP.anchorIndices) { // find the desired int point 
	    	    			for (Point intPoint: dP.Curve.intPoints) {
	    	    				String s = I.intValue() + "i";
	    						if ( intPoint.toString().equals(s) ) {
	    							dP.anchorPoints.add(intPoint);
	    							break;
	    						}
	    	    			}
	    				}
	    	    		
	    	    		dP.anchorPoints.add(dP.Curve.get(0)); // first point on the curve is necessarily the last anchor
	    	    		
	    	    		findCurves(dP.anchorPoints); // get the curves from the anchor points 
	    	    		colorCurves();
    	    		}
    	    		dP.repaint();
    	    	}
    	    	else {
    	    		scanner.close();
    	    		throw new incompleteCurveException("Insufficient points to generate closed curve!");
    	    	}
	    	}
	    	catch (incompleteCurveException e) {
	    		System.err.println(e);
	    		System.exit(0);
	    	}
	    	repaint();
	    }
	    catch(FileNotFoundException e){
	        System.err.println("Could not find file "+ "current" + "."+ e);
	        System.exit(0);
	    } 
		catch(IOException e){
	        System.err.println("Error reading integer from file " + "current" +". "+e);
	        System.exit(0);
	    }	   
	    repaint();
    }
    
    public void collapse() { 
    	try {
			if (dP.Curve.faces.size() != 1 && (!dP.overlapping) ) {
				throw new newException("Cannot collapse. There is not only one face.");
			}
			else {
				//dP.areaSwept += dP.windingArea;
				dP.areaSwept += dP.Curve.windingArea();
				dP.titusSeq.add("Ia on Face 0");
				if ( dP.areaSwept < dP.bestArea ) {
					dP.bestArea = dP.areaSwept; 
					dP.bestTitus.clear();
					for (String S: dP.titusSeq) {
						dP.bestTitus.add(S);
					}
				}
				restart();
			}		
		}
		catch (newException EE) {
			System.err.println(EE);
		}
    }
    
    public void reverseCurve() {
  
    	ArrayList<Point> reversedPoints = new ArrayList<Point>();
    	reversedPoints.add( dP.Curve.points.get(0).copy() );
    	for (int i= dP.Curve.points.size()-1; i>=1; i--) {
    		Point P = dP.Curve.points.get(i);
    		reversedPoints.add( new Point(P.getX(), P.getY() ) );
    	}
    	
    	try {
	    	dP.clear();
	    	if ( reversedPoints.size() >= 3 ) {
	    		for (int i=0; i < reversedPoints.size(); i ++) {
	    			Point p = reversedPoints.get(i);
	    			dP.addPoint(p);
	    		}
	    		dP.closeCurve();
	    	}
	    	else {
	    		throw new incompleteCurveException("Insufficient points to generate closed curve!");
	    	}
    	}
    	catch (incompleteCurveException e) {
    		System.err.println(e);
    		System.exit(0);
    	}
    	dP.repaint();
    }
    
    
    // assume anchors are in order 
    public void findCurves( ArrayList<Point> anchors) { 
    	Curve current = new Curve(dP.Curve.curve); // copy the curve over 
    	current.numberPoints(false); 
    	for (Point P: anchors) {
    		if ( P.toString().equals("0") )  { // last point is different b/c curve is simple: no split needed
    			current.anchor = P;
    			dP.selfOverlappingCurves.add(current);
    			return;
    		}
    		Curve gammaP = dP.splitCurve(P, "direct", current);
    		gammaP.anchor = P;
    		dP.selfOverlappingCurves.add(gammaP);
    		current = dP.splitCurve(P, "wrap", current); // the wraparound is the leftover curve 
    	}	
    }
    
    /** 
    public void findAnchors(Node root, int depth) {
		if ( root == null )  {
			Point P = new Point(0, 0);
			P.setDepth(depth);
			//dP.anchorPoints.add( P );
			return ;
		}
		findAnchors( root.left, depth+1);
		dP.anchorPoints.add( root.getPoint() );
		findAnchors( root.right, depth +1);
    } 
    
    public void saveAnchorsAndCurves(Node root, Curve curve, int depth) { 
    	if (root == null) {
    		Point P = new Point(0, 0);
			P.setDepth(depth);
			dP.anchorPoints.add( P );
			dP.originalAnchors.add(P);
			curve.anchor = P; // allow the curve to point at the null anchor
    		dP.selfOverlappingCurves.add( curve ) ;
    		return;
    	}
    	if ( root.left != null) {
    		saveAnchorsAndCurves( root.left, dP.splitCurve(root.getPoint(), true, curve), depth+1);
    	}
    	else { 
    		saveAnchorsAndCurves( null, dP.splitCurve(root.getPoint(), true, curve), depth+1); 
    	}
    	Point cur = root.getPoint();
    	Point copy = cur.copy();
    	copy.setDepth( cur.getDepth() );
    	dP.anchorPoints.add( cur );
    	if ( root.right != null) {
    		saveAnchorsAndCurves( root.right, dP.splitCurve(root.getPoint(), false, curve), depth+1 );
    	}
    	else { saveAnchorsAndCurves( null, dP.splitCurve( root.getPoint() , false, curve), depth+1); }
    	
    } 
    
    public void findCurves(Node root, Curve curve) { 
    	if (root == null) {
    		dP.selfOverlappingCurves.add( curve ) ;
    		return;
    	}
    	if ( root.left != null) {
    		findCurves( root.left, dP.splitCurve(root.getPoint(), true, curve) );
    	}
    	else { findCurves( null, dP.splitCurve(root.getPoint(), true, curve) ); }
    	if ( root.right != null) {
    		findCurves( root.right, dP.splitCurve(root.getPoint(), false, curve)  );
    	}
    	else { findCurves( null, dP.splitCurve( root.getPoint() , false, curve)  ); }
    	
    }  **/
    
	public void createMenubar(){
	    JMenuBar menubar = new JMenuBar();

	    JMenu fileM = new JMenu("File");
	    fileM.setMnemonic(KeyEvent.VK_F);

	    JMenuItem doneMI = new JMenuItem("Done");
	    doneMI.setMnemonic(KeyEvent.VK_D);
	    doneMI.setToolTipText("Curve fully built");
	    doneMI.addActionListener(new ActionListener(){
	    	public void actionPerformed(ActionEvent event){
	            // clear input
	    		dP.done();
	    		dP.pointsSelected.clear();
	    		saved = true; // remember that the curve was saved
	    		dP.startingWindingArea = dP.windingArea;
	    		dP.titusSeq.clear();
	    		dP.selfOverlappingCurves.clear();
	    		dP.anchorPoints.clear();
	    		
	    		System.out.println("New curve");
	    		colorCurves();
	    		
	    		try{
	    			PrintWriter writer = new PrintWriter("current");
	    	    	for(int i=0; i< dP.Curve.points.size(); i++){
	    	    		Point point = dP.Curve.points.get(i);
	    	    		writer.print(point.getX());
	    	    		writer.print(" ");
	    	    		writer.println(point.getY());
	    	    	}
		    	    	writer.close();
		    	    } 
	    		catch(FileNotFoundException e){
		    	        System.err.println("Could not find file " + "current" + "." + e);
		    	        System.exit(0);
	    	    } 
	    		catch(IOException e){
		    	        System.err.println("Error reading integer from file "+"current"+". "+e);
		    	        System.exit(0); 
	    		}
	    		dP.repaint();
	    		//dP.overlapping = dP.IsSelfOverlapping(dP.allPoints);
	    	}
	    });
	    fileM.add(doneMI);
	    
	    
	    JMenuItem compute = new JMenuItem("Compute Min Homotopy");
	    compute.setMnemonic(KeyEvent.VK_D);
	    compute.setToolTipText("Compute the minimum homotopy for the current curve.");
	    compute.addActionListener(new ActionListener(){
	    	public void actionPerformed(ActionEvent event){
	    		if ( dP.complete) {
		    		dP.baseCases = 0;
		    		dP.extraSplits = 0;
		    		
		    		dP.anchorPoints.clear();
		    		dP.selfOverlappingCurves.clear();
		    		dP.Curve.anchor = dP.Curve.get(0);
		    		Homotopy minHomotopy = dP.minHomotopy(dP.Curve, 10000);
		    		
		    		ArrayList<Integer> anchorIndices = minHomotopy.anchorIndices;
		    		dP.anchorIndices = anchorIndices;
		    		dP.minimumHomotopyArea = minHomotopy.area;
		    		    		
		    		for ( Integer I: anchorIndices) { // find the desired int point 
		    			for (Point intPoint: dP.Curve.intPoints) {
		    				String s = I.intValue() + "i";
							if ( intPoint.toString().equals(s) ) {
								dP.anchorPoints.add(intPoint);
								break;
							}
		    			}
					}
		    		
		    		dP.anchorPoints.add(dP.Curve.get(0)); // first point on the curve is necessarily the last anchor
		    		findCurves(dP.anchorPoints); // compute the curves from the anchor indices
		    		
		    		for (Point anchor: dP.anchorPoints) {
	    				Point P = new Point(anchor.getX(), anchor.getY() );
	    				P.setDepth(anchor.getDepth());
	    				dP.originalAnchors.add(P); 
		    		}
		    		
		    		
		    		dP.backupOverlappingCurves = dP.selfOverlappingCurves; // save the curves
		    		///findCurves(root, dP.Curve  ); // recompute dP.selfOverlapping curves
		    		// so that the curves in the backups are different
		    		colorCurves(); 
		    		
		    		dP.numberPoints(true);
		    		dP.repaint();
	    		}
	    	}
	    });
	    fileM.add(compute);
	    
	    JMenuItem reverseMI = new JMenuItem("Reverse");
	    reverseMI.setMnemonic(KeyEvent.VK_KP_UP);
	    reverseMI.setToolTipText("Reverse the orientation of the curve");
	    reverseMI.addActionListener(new ActionListener(){
	    	public void actionPerformed(ActionEvent event){
	            reverseCurve();
	    	}
	    });
	    fileM.add(reverseMI);
	    
	 // Restart menu item
	    JMenuItem contractMI = new JMenuItem("Contract Next SO Loop");
	    contractMI.setMnemonic(KeyEvent.VK_R);
	    contractMI.setToolTipText("Contract the next self-overlapping loop in the queue");
	    contractMI.addActionListener(new ActionListener(){
	    	public void actionPerformed(ActionEvent event){
	    		if ( dP.selfOverlappingCurves.size() > 0 ) { 
	    			if  ( dP.selfOverlappingCurves.size() == 1)  {
	    				collapse();
	    				return;
	    			}
		    		Curve cur =  dP.selfOverlappingCurves.get(0);
		    		dP.contractLoop( cur ); // contract loop 
		    		dP.anchorPoints.remove( cur.anchor);
		  	/**( anchor indices are saved to redraw K-decomp during restart without additional computation) **/
		    		dP.repaint();
	    		}
	        }
	    });
	    fileM.add(contractMI);
	    
	    
	    /** 
	    JMenuItem process = new JMenuItem("Pre-Process");
	    process.setMnemonic(KeyEvent.VK_KP_UP);
	    process.setToolTipText("Pre-process the curve by deleting all exterior monogons");
	    process.addActionListener(new ActionListener(){
	    	public void actionPerformed(ActionEvent event){
	            dP.Curve.process();
	            dP.numberPoints(true);
	            dP.repaint();
	    	}
	    });
	    fileM.add(process);
	    **/
	    
	    JMenuItem smoothDraw = new JMenuItem("Continuous Draw");
	    smoothDraw.setMnemonic(KeyEvent.VK_KP_UP);
	    smoothDraw.setToolTipText("Press to start and hold while drawing curve. Must return to start.");
	    smoothDraw.addActionListener(new ActionListener(){
	    	public void actionPerformed(ActionEvent event){
	    		dP.clear();
	    		dP.complete = false;
	    		dP.smoothDraw = true;
	    		
	    	}
	    });
	    fileM.add(smoothDraw);
	    
	    // Clear menu item
	    JMenuItem clearMI = new JMenuItem("Clear");
	    clearMI.setMnemonic(KeyEvent.VK_C);
	    clearMI.setToolTipText("Remove all points");
	    clearMI.addActionListener(new ActionListener(){
	    	public void actionPerformed(ActionEvent event){
	            // clear input
	    		dP.complete = false;
	    		dP.clear();
	    	}
	    });
	    fileM.add(clearMI);
	    
	 // Restart menu item
	    JMenuItem restartMI = new JMenuItem("Restart");
	    restartMI.setMnemonic(KeyEvent.VK_R);
	    restartMI.setToolTipText("Revert the curve to its initial state");
	    restartMI.addActionListener(new ActionListener(){
	    	public void actionPerformed(ActionEvent event){
	    		restart();
	        }
	    });
	    fileM.add(restartMI);
	    
	 // Restart menu item
	    JMenuItem collapseMI = new JMenuItem("Collapse");
	    collapseMI.setMnemonic(KeyEvent.VK_KP_DOWN);
	    collapseMI.setToolTipText("Collapse the final monogon and restart");
	    collapseMI.addActionListener(new ActionListener(){
	    	public void actionPerformed(ActionEvent event){
	    		collapse();
	        }
	    });
	    fileM.add(collapseMI);
	    
	    // Load menu item
	    JMenuItem loadMI = new JMenuItem("Load");
	    loadMI.setMnemonic(KeyEvent.VK_L);
	    loadMI.setToolTipText("Load an existing curve");
	    loadMI.addActionListener(new ActionListener(){
	    	public void actionPerformed(ActionEvent event){
	            // load input
	    		
	    		String filename;
	    		File selectedFile ;
	    		JFileChooser fc = new JFileChooser();
	    		final String dir = System.getProperty("user.dir");
	    		fc.setCurrentDirectory(new File(dir));
	    		int result = fc.showOpenDialog(new JFrame() );
	    		
	    		try {
	    			if ( result == JFileChooser.APPROVE_OPTION ) {
		    			selectedFile = fc.getSelectedFile();
		    			filename = selectedFile.toString();
		    			System.out.println("Selected file: " + selectedFile.getAbsolutePath());
		    		}
	    			else {
	    				throw new newException("File was not approved.");
	    			}
	    		}
	    		catch (newException EE) {
	    			System.err.println(EE);
	    			System.exit(0);
	    			return;
	    		}
	    		
	    		Scanner scanner; 
	    		
	    	    try{
	    	        scanner = new Scanner(selectedFile);  	        
	    	    	int x, y, n=0;
	    	    	
	    	    	ArrayList<Point> P = new ArrayList<Point>();
	    	    	while(scanner.hasNext()){
	    	    		// New point
	    	    		x = (int)scanner.nextDouble();
	    	    		y = (int)scanner.nextDouble();
	    	    		n++;
	    	    		System.out.println("Point "+x+","+y);
	    	    		P.add(new Point(x, y) );
	    	    	} 
	    	    	try {
		    	    	dP.clear();
		    	    	if ( P.size() >= 3 ) {
		    	    		for (int i=0; i < P.size(); i ++) {
		    	    			Point p = P.get(i);
		    	    			dP.addPoint(p);
		    	    		}
		    	    		dP.closeCurve();
		    	    	}
		    	    	else {
		    	    		scanner.close();
		    	    		throw new incompleteCurveException("Insufficient points to generate closed curve!");
		    	    	}
	    	    	}
	    	    	catch (incompleteCurveException e) {
	    	    		System.err.println(e);
	    	    		System.exit(0);
	    	    	}
	    	    	repaint();
	    	    }
	    	    catch(FileNotFoundException e){
	    	        System.err.println("Could not find file "+ selectedFile.toString() +". "+e);
	    	        System.exit(0);
	    	    } catch(IOException e){
	    	        System.err.println("Error reading integer from file " + selectedFile.toString() +". "+e);
	    	        System.exit(0);
	    	    }	   
	    	    repaint();
	        }
	    });
	    fileM.add(loadMI);
	    
	    // Save menu item
	    JMenuItem saveMI = new JMenuItem("Save");
	    saveMI.setMnemonic(KeyEvent.VK_S);
	    saveMI.setToolTipText("Save input");
	    saveMI.addActionListener(new ActionListener(){
	    	public void actionPerformed(ActionEvent event){
	            // save input
	    	 
	    	    JFileChooser fileChooser = new JFileChooser();
	    	    fileChooser.setDialogTitle("Specify a file to save");    // open the JFileChooser
	    	    final String dir = System.getProperty("user.dir"); // open the current directory
	    		fileChooser.setCurrentDirectory(new File(dir));
	    		
	    	    int userSelection = fileChooser.showSaveDialog(dP.frame); 
	    	    File fileToSave;
	    	    
	    	    try {
		    	    if (userSelection == JFileChooser.APPROVE_OPTION) {
		    	        fileToSave = fileChooser.getSelectedFile();
		    	        System.out.println("Save as file: " + fileToSave.getAbsolutePath());
		    	    }
		    	    else {
		    	    	throw new Exception("Not approved. File could not be saved." );
		    	    }
	    	    }
	    	    catch ( Exception EE) { 
	    	    	System.err.print(EE);
	    	    	return;
	    	    }
	    	   
	    	    try{
	    	    	PrintWriter writer = new PrintWriter(fileToSave);
	    	    	for(int i=0; i< dP.Curve.points.size(); i++){
	    	    		Point point = dP.Curve.points.get(i);
	    	    		writer.print(point.getX());
	    	    		writer.print(" ");
	    	    		writer.println(point.getY());
	    	    	}
	    	    	writer.close();
	    	    } catch(FileNotFoundException e){
	    	        System.err.println("Could not find file "+ fileToSave.getName() +". "+e);
	    	        System.exit(0);
	    	    } catch(IOException e){
	    	        System.err.println("Error reading integer from file "+ fileToSave.getName() +". "+e);
	    	        System.exit(0);
	    	    }	   
	    	    
	        }
	    });
	    fileM.add(saveMI);
	    
	    // Exit menu item
	    JMenuItem exitMI = new JMenuItem("Exit");
	    exitMI.setMnemonic(KeyEvent.VK_E);
	    exitMI.setToolTipText("Exit application");
	    exitMI.addActionListener(new ActionListener(){
	    	public void actionPerformed(ActionEvent event){
	                System.exit(0);
	        }
	    });
	    fileM.add(exitMI);
   
	    menubar.add(fileM);
	    setJMenuBar(menubar);
		
	}
}