import java.util.ArrayList;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;


public class Edge {
	
	static final int THRESHOLD = 15;
	// Attributes ( the 2 points that make up the edge )
	Point p1 = new Point();
    Point p2 = new Point();
	//private ArrayList<Triangle> triangles;
	private static int count=0; // One common count for ALL Edge objects
	private ArrayList<Point> intPoints = new ArrayList<Point>();
	boolean used; // whether the edge has been traversed when finding faces
	
	// constructor
	public Edge(Point newP1, Point newP2){
		p1 = newP1;
		p2 = newP2;
		count++;
		used = false;
	}
	
	public boolean isUsed() {
		return used;
	}
	
	public void unUse() {
		used = false;
	}
	
	public void use() {
		used = true;
	}
	public double[] vector() { // return the edge as a vector
		double [] a = { p2.getX()-p1.getX(), p2.getY()-p1.getY() };
		return a;
	}
	
	public void reverse() { // return a copy of the edge with reversed orientation
		Point P = p2;
		this.p2=  p1;
		this.p1 = P;
		ArrayList<Point> A = new ArrayList<Point>();
		if (intPoints.size() > 0 ) {
			for (int i=intPoints.size()-1; i>=0; i--) {
				Point u = intPoints.get(i);
				A.add(u);
			}
		}
		intPoints = A;
	}
	
	public double turn(Edge e) { // simple math to calculate if left turn or right turn
		/**
		 * 		|1 p_x p_y |
 		 *  det |1 q_x q_y |
		 * 		|1 r_x r_y |
		 */
		
		Point p = this.getP1();
		Point q = this.getP2();
		Point r = e.getP2();
		if ( this.p2.equals(e.getP1())) {
			double val = ( q.getX()-p.getX() )*(r.getY()-p.getY()) - (q.getY()-p.getY())*(r.getX()-p.getX());
		if ( Math.abs(val) < 10) { // margin of error in computation, most turns are between 500-5000
				return 0;
			} 
			else if ( val > 0) {
				return 1;
			}
			else {
				return -1;
			}
		}
		return Integer.MAX_VALUE;
	}
	
	// compute the turn and print info
	/**
	public double turnPrint(Edge e) {
		Point p = this.getP1();
		Point q = this.getP2();
		Point r = e.getP2();
		if ( this.p2.equals(e.getP1())) {
			double val = ( q.getX()-p.getX() )*(r.getY()-p.getY()) - (q.getY()-p.getY())*(r.getX()-p.getX());
			System.out.println("Edges" + this.toString() + "->" + e.toString() + " has value: " + val);
			if ( val == 0) {
				return 0;
			}
			else if ( val > 0) {
				return -1;
			}
			else {
				return 1;
			}
		}
		System.out.println("ERROR: EDGES DON'T LEAD INTO EACH OTHER");
		return Integer.MAX_VALUE;
	} **/
	
	public double angle() {  // get angle between edge and horizontal line through base-point (P1)
		double angle1 = Math.atan2( -this.vector()[1] , this.vector()[0] );
		angle1  = Math.round(  ((angle1 * 180)/Math.PI) *1000)/1000;
		if ( angle1 < 0) {
			angle1 = 360-Math.abs(angle1);
		}
	  //  System.out.println("Angle is: " + angle1 + " degrees on edge between " + p1.getNum() + "," + p2.getNum());
		return angle1;
	}
	
	public static int getCount(){
		return count;
	}
	
	public double distance(Point P) {
	    Point p2 = new Point( this.p2.getX() - p1.getX(), this.p2.getY() - p1.getY() );
	    double squares = p2.getX() *p2.getX() + p2.getY() *p2.getY();
	    double u = ( (P.getX() - p1.getX()) * p2.getX() + (P.getY() - p1.getY() ) * p2.getY() ) / squares;

	    if (u > 1)
	        u = 1;
	    else if (u < 0)
	        u = 0;

	    double x = p1.getX() + u * p2.getX();
	    double y = p1.getY() + u * p2.getY();

	    double dx = x - P.getX();
	    double dy = y - P.getY();

	    double dist = Math.sqrt(dx*dx + dy*dy);

	    return dist;
	}
	
	/** 
	// distance P to any point on the edge E
	public double distance(Point P) {
    	double x0 = P.getX(); double y0 = P.getY();
    	double x1 = p1.getX(); double y1 = p1.getY();
    	double x2 = p2.getX(); double y2 = p2.getY();
    	return Math.abs( (x2-x1)*(y1-y0) - (x1-x0)*(y2-y1) )/ (Math.sqrt( (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) ) );
	}**/
	
	public String toString(){
		// returns a String representation of the Edge
		return "["+p1.toString()+","+p2.toString()+"]";
	}
	
	public boolean intersects(Edge e1) {
		return Line2D.linesIntersect(this.p1.getX(), this.p1.getY(), this.p2.getX(), this.p2.getY(), e1.p1.getX(), e1.p1.getY(), e1.p2.getX(), e1.p2.getY());
	}
	//determine if the edge intersections e1 
	public boolean doesIntersect(Edge e1) {
		if ( Line2D.linesIntersect(this.p1.getX(), this.p1.getY(), this.p2.getX(), this.p2.getY(), e1.p1.getX(), e1.p1.getY(), e1.p2.getX(), e1.p2.getY()) ) {
			double m1 = (this.p2.getY() - this.p1.getY() )/ (this.getP2().getX() - this.p1.getX() );
			double b1 = this.p2.getY() - m1*this.p2.getX();
			//slope of the second line
			double m2 = (e1.p2.getY() - e1.p1.getY() )/ (e1.getP2().getX() - e1.p1.getX() );
			double b2 = e1.p2.getY() - m2*e1.p2.getX();
			double intx = (b2-b1)/(m1-m2);
			double inty = m1*intx + b1;
			Point intPoint = new Point(intx, inty);
			if ( intPoint.distance(e1.getP1()) < 6 || intPoint.distance(e1.getP2())< 6 ) {
				return false;
			}
			else {
				return true;
			}
		}
		return false;
	}
	
	public ArrayList<Point> getIntPoints(){  // return all intersection points
		return intPoints;
	}
	
	public void addIntPoint(Point p) { // add an intersection point
		intPoints.add(p);
	}
	
	public Point getIntPoint(int i) { // get an intersection point by index
		return intPoints.get(i);
	}
	
	public void removeIntPoint(int i) { // delete an intersection point by reference
		intPoints.remove(i);
	}
	public void removeIntPoint(Point p) { // delete intersection point by index
		intPoints.remove(p);
	}
	
	public Point infiniteSlope( Edge e1) {
		double starty = this.getP1().getY();
		double endy = this.getP2().getY();
		double x = this.getP1().getX();
		if ( endy < starty ) { // switch is endy< starty
			double v = starty;
			starty = endy;
			endy = v;
		}
		// slope of the second non-vertical line
		double m2 = (e1.p2.getY() - e1.p1.getY() )/ (e1.getP2().getX() - e1.p1.getX() );
		// y intersection point of the second, non-vertical line
		double b2 = e1.p2.getY() - m2*e1.p2.getX();
		double inty = m2*x + b2; // check where L2 intersects the vertical line L1
		if ( Math.abs(starty - inty) < 3 ) {
			//System.out.println(" THIS SHOULDN'T HAPPEN");
			return new Point(1000,1000);
		}
		else if (Math.abs(endy-inty) < 3 ) {
			//System.out.println(" THIS SHOULDN'T HAPPEN");
			return new Point(1000,1000);
		}
		else if ( starty < inty && inty < endy) {
			//System.out.println("infinite slope working correctly");
			return new Point(x, inty);
		}
		else {
			System.out.println("Starty: " + starty + " and endy: " + endy + " but int at " +  inty);
			System.out.println("ERROR W/INFINITE SLOPE METHOD: (0,0) RETURNED");
			return new Point(0,0);
		}
	}
	public Point getIntersection(Edge e1) {
		if ( doesIntersect(e1) ) {
			// slope of the first line
			double m1 = (this.p2.getY() - this.p1.getY() )/ (this.getP2().getX() - this.p1.getX() );
			if ( Double.isInfinite(m1) ) {
				return this.infiniteSlope(e1);
			}
			double b1 = this.p2.getY() - m1*this.p2.getX();
			//slope of the second line
			double m2 = (e1.p2.getY() - e1.p1.getY() )/ (e1.getP2().getX() - e1.p1.getX() );
			if (Double.isInfinite(m2)) {
				return e1.infiniteSlope(this);
			}
			double b2 = e1.p2.getY() - m2*e1.p2.getX();
			double intx = (b2-b1)/(m1-m2);
			if ( m1 == m2) {
				System.out.println("ERROR - ZERO SLOPE");
			}
			
			if ( Math.abs(m1*intx + b1 - m2*intx - b2) > 4) { // ensure intersec. was calculated correctly
				System.out.println("ERROR LINES DON'T INTERSECT");
			}
			if ( Double.isNaN(intx) ) {
				//System.out.println(e1.getP1().toString() + " and " + e1.getP2().toString());
				System.out.println("Slopes are: " + m1 + " and " + m2 );
				System.out.println("y intercepts are: " + b1 + " and " + b2 );
			}
			double inty = m1*intx + b1;
			double intyALT = m2*intx + b2;
		//	System.out.println("IntPoint at ("+ intx +  "," + inty + ")");
			return new Point(intx, intyALT);
		}
		else {
			System.out.println("ERROR!!! - NO INTERSECTION BETWEEN EDGES");
			return new Point(50, 50);
		}
	}
	// Getter and setter
	public Point getP1(){
		return p1;
	}
	
	public Point getP2(){
		return p2;
	}
	
	public void setP1(Point newP1){
		p1 = newP1;
	}
	
	public void setP2(Point newP2){
		p2 = newP2;
	}
	
	public boolean equals(Edge e2){
		if(p1.equals(e2.getP1()) && p2.equals(e2.getP2())){
			return true;
		}
		else if (p1.equals(e2.getP2()) && p2.equals(e2.getP1()) ) {
			return true;
		}
		else{
			return false;
		}
	}
		
}
