import javax.swing.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;


public class Face {
	private ArrayList<Edge> edges = new ArrayList<Edge>();
	private ArrayList<Point> points = new ArrayList<Point>();
	private double area;
	private int wNum;
	private Point rep; // representative point, inside the area of the face
	private int size; // number of points/edges on the face
	private int[] pointNums; // integer array of the # of each point, determined by the traversal
	//needed for face comparison
	private int[] pointNumsNoRep; // integer array of #'s of points on the face WITHOUT REPETITION 
	private int num; // # of face, in order the faces were computed
	private Color c; 
	private int type; // 1 == monogon, 2 == bigon, 3 == triangle, 0 == none 
	
	private boolean good;
	private Polygon Poly; // for drawing the face
	private int depth;
	
	public Face(ArrayList <Edge> edges ) { // constructor w/ input of edges in CCW order
		this.edges = edges;
		for (int i=0; i<edges.size(); i++) {
			Edge e = edges.get(i);
			points.add(e.getP1()); // get the first point of each face to get points in CCW order
		}
		size = edges.size();
		double xtotal = 0; // sum the x and y coordinates 
		double ytotal = 0;
		pointNums = new int[size]; // get the #'s of the points on the face 
		ArrayList<Integer> pointNumsNoRep1 = new ArrayList<Integer>();
		Poly = new Polygon();
		for (int i=0; i< points.size(); i++) {
			Point p = points.get(i); 
			pointNums[i] = p.getNum();
			if ( !pointNumsNoRep1.contains(new Integer(p.getNum() )) ) {pointNumsNoRep1.add(new Integer(p.getNum())); }
			xtotal += p.getX(); 
			ytotal += p.getY();
			Poly.addPoint( (int)p.getX(), (int)p.getY());
		}
		
		pointNumsNoRep = new int[pointNumsNoRep1.size() ];
		for (int i=0; i< pointNumsNoRep1.size(); i++ ) {
			int k = pointNumsNoRep1.get(i).intValue();
			pointNumsNoRep[i] = k;
		}
		
		this.area = faceArea(points);
		
		c = new Color(100, 100, 100);
				
		
		//c = new Color( 20 + (int) (Math.random()*100), 70 + (int) (Math.random()*185), 100+(int) (Math.random()*155) );
		
		
		
		//int k = (int)(Math.random()*90);
		//c = new Color( 25+(int)(Math.random()*40)+ k, 75 + k + (int)(Math.random()*90), 100 + k + (int)(Math.random()*65) );
		rep = new Point( xtotal/size, ytotal/size); //coord of rep are avg.s of x,y coord. of other points
		//getNewRep();
		this.setType();
	}
	
	public void setGood(boolean b) {
		good = b;
	}
	public boolean isGood() {
		return good;
	}
	
	public void setColor(Color C) {
		this.c = C;
	}
	
	public Color getColor() {
		return this.c;
	}
	
	public void setDepth(int i ){ 
		this.depth = i;
	}
	
	public int getDepth() {
		return this.depth;
	}
	
	public void setWindingNum(int j) {
		wNum = j;
	}
	
	public int getWindingNum() { 
		return wNum;
	}
	
	// determine if the outside of the polygon is above or below the edge 
    public static boolean upIsInside(Polygon P, Edge E1) {
    	double epsilon = 2;
    	Point midpoint = new Point( (int)(0.5*(E1.getP1().getX() + E1.getP2().getX())), 
    			(int)(0.5*( E1.getP1().getY() + E1.getP2().getY()) ) );
    	double x = midpoint.getX();
    	double m = (E1.getP2().getY()-E1.getP1().getY()) / (E1.getP2().getX() - E1.getP1().getX() );
    	double b = E1.getP1().getY() - m*E1.getP1().getX();
    	double bNEW = -epsilon*Math.sqrt(m*m+1) + b;
    	
    	if ( E1.getP1().getX() == E1.getP2().getX() ) { // vertical line
    		// shift the midpoint over to the left a bit
    		return P.contains( new Point(midpoint.getX()-epsilon, midpoint.getY()));
    	}
    	boolean answer = P.contains( new Point( x, x*m+bNEW)  );

    	// evaluate the midpoint x coordinate shifted up epsilon, see if it is in the polygon
    	//boolean answer = P.contains( new Point(x, midpoint.getY()-4) );
    	return answer;
    }
    
    public void getNewRep() {
    	ArrayList<Point> candidates = new ArrayList<Point>();
    	candidates.add(rep);
    	for (Edge E: edges) {
    		candidates.add( getNewRep(E) ); // get rep from each edge
    	}
    	int best = 0;
    	Point newRep = rep;
    	for (Point P: candidates) {
    		for (int i=0; i<1000; i++) {
    			if ( !this.Poly.contains(P.getX()-2*i, P.getY()-2*i, 1+4*i, 1+4*i) ) {
    				newRep = ( i > best ? P : newRep); // leave it alone unless P has more space
    				best = ( i > best) ? i : best;
    				break;
    			}
    		}
    	}
    	rep = newRep;
    }
    
	public Point getNewRep(Edge e) {
		//test that the rep isn't right on the border of the face 
		//if ( ! Poly.contains(rep.getX()-5, rep.getY()-5, 10, 10) ) {
			// get the midpoint of the edge
			Point mid = new Point( (0.5*(e.getP1().getX()+e.getP2().getX() ) ),
					(0.5* ( e.getP1().getY()+e.getP2().getY() ) ) );
			int signY;
			if ( upIsInside(Poly, e) ) { signY = -1;  } // test movement upwards
			else { signY = 1; } // test movement downwards
			
			int maxY = 0;
			for (int i=1; i<200; i++) { // test containment, moving the point vertically 10 up/down
				Point P = new Point(mid.getX(), mid.getY()+ signY*i*5);
				if ( !Poly.contains(P) ) { // if we moved too far, stop
					maxY = i; // save the max parameter for up/down movement 
					break;
				}
			}
			
			int bestY = (int)(0.5*maxY); // take half of the best y parameter,
			//shift the midpoint up that amount * 5*sign
			
			int maxL = 0; int maxR = 0;
			for (int i=1; i<200; i++) { // test containment, moving the point left 5 units incrementally
				Point P = new Point(mid.getX() - i*5, mid.getY() + bestY*signY*5);
				if ( !Poly.contains(P) ) { // if we moved too far, stop
					maxL = i; // save the max parameter for up/down movement 
					break;
				}
			}
			for (int i=1; i<200; i++) { // test containment, moving the point right 5 units incrementally
				Point P = new Point(mid.getX() + i*5, mid.getY() + bestY*signY*5);
				if ( !Poly.contains(P) ) { // if we moved too far, stop
					maxR = i; // save the max parameter for up/down movement 
					break;
				}
			}
			int bestX = (int)(0.5*(-maxL+maxR)); // take the middle of the x values 
			// inside of the polygon at this Y value
			
			try { 
				if (maxL == 0 && maxR == 0) {
					throw new newException("ERROR!! Moving left and right from newY point was outside face!");
				}
			}
			catch ( newException EE) {
				//rep = new Point( mid.getX(), mid.getY() + bestY*signY*5);
				System.err.println(EE);
				return rep;
			}
			
			// bestX is ALREADY a correctly signed quantity
			return new Point( mid.getX() + bestX*5, mid.getY() + bestY*signY*5);
		
	}
	
	public Polygon getPolygon() {
		return this.Poly;
	}
	
	public Point getPoint(int i) {
		return this.points.get(i);
	}
	
	public void setType() { // determine if face is monogon, bigon, triangle, or none
		int count = 0;
		ArrayList<Point> intPoints = new ArrayList<Point>();
		for (Point P: this.points) {
			if (P.isIntersection() && !intPoints.contains(P) ) {
				count++;
				intPoints.add(P);
			}
			else if ( intPoints.contains(P) ) { // ignore this case for now
				type = 0; 
				return;
			}
		}
		if ( count == 3) { // if all 3 points are intPoints, we have triangle 
			type = 3;
		}
		else if ( count == 1) { // if there is only 1 intPoint, then we have a monogon
			type = 1;
		}
		else if ( count == 2) { // if there are two intPoints that aren't connected, we have a bigon
			type = 2;
		}
		else {
			type = 0; // doesn't fit any of the three cases 
		}
	}
	
	public int getType() { 
		return this.type;
	}
	
	public class newException extends Exception {
        public newException(String message) {
            super(message);
        }
    }
	
	public boolean above(Point P, Edge E) {
    	if ( E.getP2().getX() == E.getP1().getX() ) { // 'infinite' slope
    		return P.getX() < E.getP1().getX(); // return if P is to the left of E 
    	}
    	else {
    		double m = (E.getP2().getY() - E.getP1().getY() )/ (E.getP2().getX() - E.getP1().getX() );
    		double b = E.getP2().getY() - m*E.getP2().getX();
    		return P.getY() < m*P.getX()+b;
    	}
    }
	
	public ArrayList<Point> triFirst() {
		return split(0);
	}
	public ArrayList<Point> triSecond() {
		return split(1);
	}
	public ArrayList<Point> triThird() {
		return split(2);
	}
	
	public ArrayList<Point> restOfCurve(int part) { // return the part you have, give the rest of the curve
		try { 
			if (this.type != 3) {
				throw new newException("Cannot break into 3 parts. Not a 3-gon.");
			}
			ArrayList<Point> restOfTri = new ArrayList<Point>();
			/** the int point at the end of split(i) is also at beginning of split(i+1 (mod 3))   **/
			if ( part == 0) {
				restOfTri.addAll( split(1) ); restOfTri.remove(restOfTri.size()-1);
				restOfTri.addAll( split(2) ); return restOfTri;
			}
			else if ( part == 1 ) {
				restOfTri.addAll( split(2) ); restOfTri.remove(restOfTri.size()-1);
				restOfTri.addAll(split(0) ); return restOfTri;
			}
			else if ( part == 2) {
				restOfTri.addAll( split(0) ); restOfTri.remove(restOfTri.size()-1);
				restOfTri.addAll( split(1) ); return restOfTri;
			}
			else {
				throw new newException("Error. Must input 0, 1, or 2 to determine which portion to return");
			}
		}
		catch(newException EE) {
			System.err.println(EE);
			return null;
		}
	}
	
	public ArrayList<Point> split(int part) {
		try { 
			if (this.type != 3) {
				throw new newException("Cannot break into 3 parts. Not a 3-gon.");
			}
			int j, k, l;
			j = k = l = -1;
			for (int i=0; i<points.size(); i++) {
				Point P = points.get(i);
				if ( P.isIntersection() ) {
					if ( j == -1 ) { j = points.indexOf(P); }
					else if ( k == -1) { k = points.indexOf(P); }
					else { 
						l = points.indexOf(P);
						break;
					}
				}
			}
			ArrayList<Point> first, second, third;
			first = new ArrayList<Point>();
			second = new ArrayList<Point>();
			third = new ArrayList<Point>();
			for (int i=l; i<points.size()+l; i++) {
				int n = i % points.size();
				if ( n <= j || n >= l) {
					first.add( points.get(n) );
				}
				if ( n >= j && n <= k) {
					second.add( points.get(n) );
				}
				if ( n >= k && n < l) {
					third.add( points.get(n) );
				}
			}
			third.add( points.get(l) );
			if ( part == 0) { return first; }
			else if ( part == 1) { return second; }
			else if ( part == 2 ) { return third; }
			else {
				throw new newException("Error. Must input 0, 1, or 2 to determine which portion to return");
			}
		}
		catch (newException EE) {
			System.err.println(EE);
			return null;
		}
	}
	
	public ArrayList<Point> split(boolean half) {  // return half of the face
		try {
			if (this.type == 2) { // ensure we have a bigon to split
				int j = -1; 
				int k = -1; 
				for (Point P: points) {
					if ( P.isIntersection() ) {
						if ( j != -1) { // if the first intersection point has been found
							k = points.indexOf(P); // label the second one 
							break;
						}
						else { // label the first intersection point
							j = points.indexOf(P);
						}
					}
				}
				if ( j == -1 || k == -1 ) { 
					System.out.println("ERROR! Intersection points weren't found!");
				}
				int m = Math.min(j, k); 
				int n = Math.max(j, k);
				j = m;
				k = n;  // let j be the index of the first int point, k index of the second;
				
				ArrayList<Point> firstHalf = new ArrayList<Point>();
				for (int l=j; l<=k ; l++) { 
					Point current = points.get(l);
					firstHalf.add(current);
				}
				ArrayList<Point> secondHalf = new ArrayList<Point>();
				for (int l=k; l<points.size(); l++) {
					Point current = points.get(l);
					secondHalf.add(current);
				}
				for (int r=0; r<=j; r++) {
					Point current = points.get(r);
					secondHalf.add(current);
				}
				boolean firstAbove;
				// determine if the halves were named correctly
				if ( firstHalf.size() == 2) {
					boolean b = !(above(secondHalf.get(1), new Edge(points.get(j),points.get(k))));
					firstAbove = b;
				}
				else {
					boolean b = above(firstHalf.get(1),new Edge(points.get(j), points.get(k) ) );
					firstAbove = b;
				}
				// check if the 1st non int point on the firstHalf is above the line connecting
				// the two intersection points of the face 
				if ( half ) {
					if ( firstAbove ) {
						return firstHalf;
					}
					else { return secondHalf; }
				}
				else { // return the bottom half
					if ( firstAbove) {
						return secondHalf;
					}
					else {
						return firstHalf;
					}
				}
				
			}
			else {
				throw new newException("Cannot split - the face is not a bigon!");
			}
		}
		catch (newException e) {
			System.err.println(e);
			return new ArrayList<Point>(); // return null array
		}
	}

	// split a bigon into the upper face 
	public ArrayList<Point> splitFirst() {  // return arrayList of points, in numbered order
		return split(true);
	}
	
	// split bigon into the lower face 
	public ArrayList<Point> splitSecond() {  // return arrayList of points, in numbered order
		return split(false);
	}
	public boolean isCW() { // check if the face is oriented clockwise
		double det = 0.0;
		for (int i = 0; i < this.points.size(); i++) {
			det = det + (this.points.get(i).getX() * this.points.get(((i+1) % this.points.size())).getY());
		}
		
		for (int j = 0; j < this.points.size(); j++) {
			det = det - (this.points.get(j).getY() * this.points.get(((j+1) % this.points.size())).getX());
		}
		
		return det > 0;
	}
	
	public static double faceArea(ArrayList<Point> facePoints) {
		double det = 0.0;
		for (int i = 0; i < facePoints.size(); i++) {
			det = det + (facePoints.get(i).getX() * facePoints.get(((i+1) % facePoints.size())).getY());
		}
		
		for (int j = 0; j < facePoints.size(); j++) {
			det = det - (facePoints.get(j).getY() * facePoints.get(((j+1) % facePoints.size())).getX());
		}
		
		det = Math.abs(det);
		double answer = Math.round((det/2));
		return answer/ 1000;
	}
	 
	public void reverse() { // reverse the orientation of the face 
		ArrayList<Edge> newEdges = new ArrayList<Edge>();
		ArrayList<Point> newPoints = new ArrayList<Point>();  
		newPoints.add(points.get(0));
		newEdges.add( new Edge(points.get(0), points.get(points.size()-1)) );
		for (int i = points.size()-1; i >=1; i--) { // run through points in opposite order 
			Point cur = points.get(i);  
			Point next = points.get( (i-1)% points.size() ) ;
			Edge e = new Edge(cur, next); // create edge between each point and its predecessor in the array
			newEdges.add(e);  
			newPoints.add(cur);
		}
		this.edges = newEdges; // update face's ArrayList of points and edges 
		this.points = newPoints;
	}
	
	public void setNum(int i) {
		this.num = i;
	}
	
	public int getNum() { 
		return num;
	}
	/**  // can't have two constructors which take ArrayLists of different types of objects
	 * because java sucks and can't differentiate the two 
	public Face(ArrayList <Point> points ) { // constructor w/ input of points in CCW order
		this.points = points;
		for (i=0; i<points.size(); i++) {
			Point cur = points.get(i);
			Point next = points.get( (i+1)%points.size() );
			Edge e = new Edge(cur, next);
			Edges.add(e);
		}
	} **/
	
	public void incrementPoints() { // count this face towards each component points' face count
		ArrayList<Point> used = new ArrayList<Point>();
		for (Point p: points) {
			if ( ! used.contains(p) ) {
				used.add(p);
				p.foundFace();
				p.addFace(this.num);
			}
		}
	}
	
	public boolean isExterior() { // check if face is exterior
		for (Point p: points) {
			if ( !p.isIntersection() && p.faceMembership() == 1 ) { // check if there is a point only connected to this face
				return true;
			}
		}
		return false;
	}
	public int[] getPointNums() {
		return this.pointNums;
	}
	
	public int[] getPointNumsNoRep() {
		return this.pointNumsNoRep;
	}
	
	public ArrayList<Edge> getEdges() {
		return edges;
	}
	
	public ArrayList<Point> getPoints() {
		return points;
	}
	public int getSize() {
		return points.size();
	}
	
	public Point getRep() {
		return rep;
	}
	
	public double getArea() {
		return area;
	}
	
	public String toString() {
		String rep = "";
		for (Point p: points) {
			if ( p.toString().contains("i") ) {
				rep += p.toString() + "->";
			}
		}
		if ( rep == "")
			return "0";
		rep = rep.substring(0, rep.length()-2);
		return rep;
	}
	
	public boolean contains(Face f) {
		int[] points1 = f.getPointNums();
		for (int k: points1) { // look through all points of second face 
			boolean contained = false;
			for (int j: this.pointNums ) { 
				if ( j == k) { contained = true; } // determine if a match is found
			}
			if ( !contained) { return false; } // if point in 2nd face isn't in the 1st face, return false
		}
		return true;
	}
	
	public boolean equalsNoRep(Face f) { // determine if faces are made of the same points
		int[] points1 = f.getPointNums();
		if ( f.getSize() != this.size) {  // if faces don't have same # of points, they are different 
			return false; 
		}
		Arrays.sort(this.pointNums);
		Arrays.sort(points1);
		
		return Arrays.equals(this.pointNums, points1);
	}
	
	public boolean equals(Face f) { // determine if faces are made of the same points
		int[] points1 = f.getPointNumsNoRep();
		if ( f.getSize() != this.size) {  // if faces don't have same # of points, they are different 
			return false; 
		}
		Arrays.sort(this.pointNumsNoRep);
		Arrays.sort(points1);
		
		return Arrays.equals(this.pointNumsNoRep, points1);
	}

}


