import javax.swing.*;


import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;

public class Point extends Point2D {
	// Attributes
	private double x; // (x,y)-coordinates of point
	private double y; 
	private static int count=0; // One common count for ALL Point objects
	private boolean intersection;
	private double startX;
	private double startY;
	private ArrayList<Edge> fullEdges = new ArrayList<Edge>();
	private ArrayList<Edge> shortEdges = new ArrayList<Edge>();
	private Edge[] vectors = new Edge[2];
	private int num; // number of the point (order in intersection sequence)
	private int faceConnections; // how many faces the point is connected to 
	private ArrayList<Integer> faces; // indices of faces point is connected to 
	private boolean justMade;
	private boolean selected; // if point is selected, to add new 
	private int uses; // # times the point traversed in the curve (check double traversal of start point)
	private int visits;
	private boolean sign;
	private String originalName;
	private int[] indices = {-1, -1};
	private int intPointIndex = -1;
	private int depth = -1;
	Color color;
	
	public Point(double newX, double newY){ // constructor
		x = newX;
		y = newY;
		startX = newX;
		startY = newY;
		count++;
		intersection = false;
		num = -1;
		faceConnections = 0;
		faces = new ArrayList<Integer>();
		uses = 1; // point has been used
		justMade = true;
		selected = false;
		visits = 0;
	}
	
	public Point copy() {
		return new Point( this.getX(), this.getY() );
	}
	public void setDepth(int j) {
		depth = j;
	}
	
	public int getDepth() {
		return depth;
	}
	public void setIntPointIndex(int k) {
		intPointIndex = k;
	}
	
	public int getIntPointIndex() {
		return intPointIndex;
	}
	
	public class newException extends Exception {
        public newException(String message) {
            super(message);
        }
    }
	
	public void resetIndices() {
		indices[0] = -1; indices[1] = -1;
	}
	
	public void setIndex(int k) {
		if ( indices[0] == -1) { indices[0] = k; }
		else if ( indices[1] == -1) { indices[1] = k; }
		else { 
			try { throw new newException("Indices of the int point are already set."); }
			catch ( newException EE) { System.err.print(EE); }
		}
	}
	
	public int getIndex(int j) {
		if ( j == 0 || j == 1 ) {
			return indices[j];
		}
		else { 
			try { throw new newException("Can only get indices 0 or 1."); }
			catch (newException EE) { System.err.println(EE); return -10; }
		}
	}
	
	
	public Edge getVector(int i) {
		try {
			if ( 0 <= i && i < 2) {
				return vectors[i];
			}
			else {
				throw new newException("There is no vector " + i);
			}
		}
		catch (newException a) {
			System.err.println(a); 
			return null;
		}
	} 
	
	public void addVector(Edge e) {
		try {
			if ( vectors[0]==null) {
				vectors[0] = e;
			}
			else if ( vectors[1] == null) {
				vectors[1] = e;
			}
			else {
				throw new newException("Error! vectors are already loaded!");
			}
		}
		catch(newException E) {
			System.err.println(E);
		}
	}
	
	
	public boolean calcSign() { // calculate sign of determinant of 2 tangent vectors at point
		Point P1 = vectors[0].getP2();
		Point P2 = vectors[1].getP2();
		sign =  ( (P1.getX()- this.getX())*(P2.getY()-this.getY()) -
				(P1.getY()-this.getY())*(P2.getX()-this.getX()) ) > 0;
		return sign;
	}
	
	public boolean getSign() {
		return sign;
	}
	
	public int getVisits() { 
		return visits;
	}
	
	public void visit() {
		visits ++;
	}
	
	public void resetVisits() {
		visits = 0;
	}
	
	public boolean det() {
		if (!intersection) {
			System.out.println("Error not an int point! shouldn't calc +/- sign!");
			return true;
		}
		Edge e1 = this.fullEdges.get(0);
		Edge e2 = this.fullEdges.get(1);
		//if ( points.indexOf(e1.getP1()) 
		return false;
		
	}
	public boolean isSelected() {
		return this.selected;
	}
	
	public void select() {
		this.selected = true;
	}
	
	public void unSelect() { 
		this.selected = false;
	}
	
	public boolean isNew() { // determine if point is brand new
		return justMade;
	}
	
	public void made() { // point is not new anymore
		justMade = false;
	}
	
	public Point(){ // constructor
		this(0,0);
	}
	
	public void resetUses() {
		uses = 0;
	}
	
	public int getUses() {
		return uses;
	}
	public void use() {
		uses++;
	}
	public void addFace(int i) {
		faces.add(new Integer(i));
	}
	
	public ArrayList<Integer> getFaces() {
		return faces;
	}
	public void foundFace() {
		faceConnections++;
	}
	
	public void removeConnections() {
		faceConnections = 0;
	}
	
	public int faceMembership() {
		return faceConnections;
	}
	
	public double closestVal (ArrayList<Point> A ){
		Point p = this.closest(A);
		return this.distance(p);
	}
	
	public double closestValEQ(ArrayList<Point> A) {
		return this.distance( closestEqAllowed(A) );
	}
	
	public Point closestEqAllowed(ArrayList<Point> A) { // return closest point in the array 
		double minDistance = 1000000;
		Point best = new Point(0,0);
		for (Point p: A) {
			double curDistance = this.distance(p);
			if ( curDistance < minDistance) {
				best = p;
				minDistance = curDistance;
			}
		}
		return best;
	}
	
	public Point closest(ArrayList<Point> A) { // return closest point in the array 
		double minDistance = 1000000;
		Point best = new Point(0,0);
		for (Point p: A) {
			double curDistance = this.distance(p);
			if ( !this.equals(p) && curDistance < minDistance) {
				best = p;
				minDistance = curDistance;
			}
		}
		return best;
	}
	
	public Edge getShortEdge(int i) { 
		return shortEdges.get(i);
	}
	
	public void sortShortEdges() {  // sort the edges based on CCW angle
		ArrayList<Edge> sorted = new ArrayList<Edge>();
		ArrayList<Edge> edges = new ArrayList<Edge>();
		int startSize = shortEdges.size();
		for (Edge e: shortEdges) { //copy over shortEdges to new arrayList edges to not affect the former
			edges.add(e); 
		}
		for (int i=0; i<startSize; i++) { // run edges.size() times, getting i^th smallest elem. each time
			double min = 100000; // angle size that will be larger than all other angle sizes
			Edge best = null; // edge w/ currently smallest angle
			for (int j=0; j<edges.size(); j++) {
				Edge cur = edges.get(j);
				if ( cur.angle() < min ) {// update smallest angle and edge w/ smallest angle
					min = cur.angle(); 
					best = cur;
				}
			}
			sorted.add(best); // add it to the sorted array
			edges.remove(best); // remove best one so it isn't used again 
		}
		//System.out.println("New check on edges for point " + this.num);
		for (Edge f: sorted ) {
	//		System.out.println("Angle of edge to point "+ f.getP2().getNum() + " is " + f.angle());
		}
		shortEdges = sorted; // update order of shortEdges now
	}
	
	public Point closest(Edge E){   // look through intPoints and endpoints, take closest points
		
		//System.out.print("On the edge: " + E.toString());
		Point best = new Point();
		double minDistance = 1000000;
		if ( E.getIntPoints().size() > 0 ) {//ensure some int points exist
			for (Point intPoint: E.getIntPoints() ) {
				double curDistance = this.distance(intPoint);
				if ( !this.equals(intPoint) && curDistance < minDistance) {
					best = intPoint;
					minDistance = curDistance;
				}
			}
		}
		double dis1 = this.distance(E.getP1()); // check distance to 1st endpoint of edge
		if ( !this.equals(E.getP1()) && dis1 < minDistance) {
			best = E.getP1();
			minDistance = dis1;
		}
		double dis2 = this.distance(E.getP2()); // check distance to 2nd endpoint of edge
		if ( !this.equals(E.getP2()) && dis2 < minDistance) {
			best = E.getP2();
		}
		return best;
    }
	
	public void setNum(int k) { // set the point # in the intersection sequence
		this.num = k;
	}
	
	public int getNum() { // get the point # in the intersection sequence
		return this.num;
	}
	
	public void setIntersection() {  // set that the point is an intersection point
		intersection = true;
	}
	public boolean isIntersection() { // return whether the point is an intersection point
		return intersection;
	}
	
	public void notIntersection() { // point is no longer an int point
		intersection = false;
	}
	
	public ArrayList<Edge> getShortEdges() { // edges of the point, connecting to adjacent points
		return this.shortEdges;
	}
	
	public void addShortEdge(Edge e) { // add new edge connecting to an adjacent point
		shortEdges.add(e);
	}
	
	public void removeShortEdge(Edge e) { // remove a short edge
		shortEdges.remove(e);
	}
	
	public ArrayList<Edge> getEdges() { // return all edges
		return this.fullEdges;
	}
	
	public Edge getEdge(int i) {
		return fullEdges.get(i);
	}
	/**
	public Edge getEdge(Edge e) {
		fullEdges.contains(o)
	} **/
	
	public void addEdge(Edge e) { // add an edge
		this.fullEdges.add(e);
	}
	
	public void removeEdge(int i) { // remove edge at index i
		this.fullEdges.remove(i);
	}
	
	public void name(String s) {
		originalName = s;
	}
	
	public String toString() {
		return originalName;
	}
	/**
	public String toString(){
		// returns a String representation of the point
		return " " + (( isIntersection() ) ? intPointIndex +"i" : num);
	}**/
	
	public void setLocation(int x, int y) { // change location of the point
		this.x = x;
		this.y = y;
	}
	
	public void setLocation(double x, double y) { // change location of the point
		this.x = (int) x;
		this.y = (int) y;
	}
	
	public void updateLocation(double x, double y) { // move point, update "starting point" coordinates
		this.setLocation(x, y);
		startX = (int) x;
		startY = (int) y;
	}
	
	public double distanceMoved() { // determine how far point was moved from its "starting point"
		return this.distance(new Point(startX, startY));
	}
	
	// Getter and setter
	public double getX(){
		return x;
	}
	
	public double getY(){
		return y;
	}
	
	public void setX(double newX){
		x = newX;
	}
	
	public void setY(double newY){
		y = newY;
	} 
	
	public double getStartX() {
		return this.startX;
	}
	
	public double getStartY() {
		return this.startY;
	}
	
	public boolean equals(Point p2){
		if( Math.abs(x-p2.getX()) < 1 && Math.abs(y - p2.getY()) < 1 ){
			return true;
		} 
		else{
			return false;
		}
	}
}
